#ifndef ALGOIM_STENCILPOLY_HPP
#define ALGOIM_STENCILPOLY_HPP

/* algoim::StencilPoly<N,Degree> implements stencil-based polynomial interpolation algorithms
   for 10 different classes of polynomial in 2D and 3D on uniform Cartesian grids. The templated
   typedef allows the (N,Degree) parameter pair to be translated to the specific polynomial
   class implementing the interpolation algorithm:
      N = 2 or 3.
      Degree = -1: choose the bicubic/tricubic interpolant.
      Degree \in {2,3,4,5}: choose the corresponding "Taylor" polynomial stencil
   If (N,Degree) is undefined, a compiler error relating to "InvalidPoly" will be generated.

   For more information, refer to the paper
      R. I. Saye, High-order methods for computing distances to implicitly defined surfaces,
      Communications in Applied Mathematics and Computational Science, 9(1), 107-141 (2014),
      http://dx.doi.org/10.2140/camcos.2014.9.107
*/

#include "uvector.hpp"

namespace algoim
{
    namespace detail
    {
        // Stencils and precomputed pseudoinverses of Vandermonde matrices for certain kinds of least-squares
        // polynomial interpolation. The following constants have approximately 40 digits of accuracy, and
        // could therefore be used with IEEE754 quadruple precision (binary128) if the compiler supported it.
        struct StencilPolyData
        {
            // A 12-point stencil in two dimensions, suitable for obtaining a least-squares polynomial of degree two or three
            static const uvector<int,2>* N2_stencil12points()
            {
                static const uvector<int,2> data[12] =
                {
                    // Vertices of square
                    uvector<int,2>(0, 0),
                    uvector<int,2>(0, 1),
                    uvector<int,2>(1, 0),
                    uvector<int,2>(1, 1),
                    // plus vertices of neighbouring faces
                    uvector<int,2>(-1, 0),
                    uvector<int,2>(-1, 1),
                    uvector<int,2>(2, 0),
                    uvector<int,2>(2, 1),
                    uvector<int,2>(0, -1),
                    uvector<int,2>(1, -1),
                    uvector<int,2>(0, 2),
                    uvector<int,2>(1, 2)
                };
                return data;
            }

            // A 24-point stencil in two dimensions, suitable for obtaining a least-squares polynomial of degree four and five
            static const uvector<int,2>* N2_stencil24points()
            {
                static const uvector<int,2> data[24] =
                {
                    // Vertices of square
                    uvector<int,2>(0, 0),
                    uvector<int,2>(0, 1),
                    uvector<int,2>(1, 0),
                    uvector<int,2>(1, 1),
                    // plus vertices of neighbouring faces
                    uvector<int,2>(-1, 0),
                    uvector<int,2>(-1, 1),
                    uvector<int,2>(2, 0),
                    uvector<int,2>(2, 1),
                    uvector<int,2>(0, -1),
                    uvector<int,2>(1, -1),
                    uvector<int,2>(0, 2),
                    uvector<int,2>(1, 2),
                    // plus neighbouring faces of those faces
                    uvector<int,2>(-2, 0),
                    uvector<int,2>(-2, 1),
                    uvector<int,2>(3, 0),
                    uvector<int,2>(3, 1),
                    uvector<int,2>(0, -2),
                    uvector<int,2>(1, -2),
                    uvector<int,2>(0, 3),
                    uvector<int,2>(1, 3),
                    // plus corners of 4 x 4 patch
                    uvector<int,2>(-1,-1),
                    uvector<int,2>(2,-1),
                    uvector<int,2>(-1,2),
                    uvector<int,2>(2,2)
                };
                return data;
            }

            static const double* N2_polyDegree2Inverse()
            {
                static const double data[6*12] = {0.3086124401913875598086124401913875598086,0.2368421052631578947368421052631578947368,0.2368421052631578947368421052631578947368,0.2177033492822966507177033492822966507177,0.13038277511961722488038277511961722488,
                   0.005980861244019138755980861244019138756,-0.0849282296650717703349282296650717703349,-0.0514354066985645933014354066985645933014,0.13038277511961722488038277511961722488,0.005980861244019138755980861244019138756,
                   -0.0849282296650717703349282296650717703349,-0.0514354066985645933014354066985645933014,0.0532296650717703349282296650717703349282,0.1058612440191387559808612440191387559809,0.1967703349282296650717703349282296650718,
                   0.1441387559808612440191387559808612440191,-0.340311004784688995215311004784688995215,-0.1824162679425837320574162679425837320574,0.0903110047846889952153110047846889952153,-0.0675837320574162679425837320574162679426,
                   -0.124401913875598086124401913875598086124,0.1244019138755980861244019138755980861244,0.0334928229665071770334928229665071770335,-0.0334928229665071770334928229665071770335,0.0532296650717703349282296650717703349282,
                   0.1967703349282296650717703349282296650718,0.1058612440191387559808612440191387559809,0.1441387559808612440191387559808612440191,-0.124401913875598086124401913875598086124,0.1244019138755980861244019138755980861244,
                   0.0334928229665071770334928229665071770335,-0.0334928229665071770334928229665071770335,-0.340311004784688995215311004784688995215,-0.1824162679425837320574162679425837320574,0.0903110047846889952153110047846889952153,
                   -0.0675837320574162679425837320574162679426,-0.125,-0.125,-0.125,-0.125,0.125,0.125,0.125,0.125,0,0,0,0,0.0526315789473684210526315789473684210526,-0.0526315789473684210526315789473684210526,-0.0526315789473684210526315789473684210526,
                   0.0526315789473684210526315789473684210526,0.1578947368421052631578947368421052631579,-0.1578947368421052631578947368421052631579,-0.1578947368421052631578947368421052631579,0.1578947368421052631578947368421052631579,
                   0.1578947368421052631578947368421052631579,-0.1578947368421052631578947368421052631579,-0.1578947368421052631578947368421052631579,0.1578947368421052631578947368421052631579,-0.125,-0.125,-0.125,-0.125,0,0,0,0,0.125,0.125,0.125,0.125};
                return data;
            }

            static const double* N2_polyDegree3Inverse()
            {
                static const double data[10*12] = {0.763157894736842105263157894736842105263,0.2368421052631578947368421052631578947368,0.2368421052631578947368421052631578947368,-0.236842105263157894736842105263157894737,0.039473684210526315789473684210526315789,
                   -0.039473684210526315789473684210526315789,-0.0394736842105263157894736842105263157895,0.0394736842105263157894736842105263157895,0.039473684210526315789473684210526315789,-0.039473684210526315789473684210526315789,
                   -0.0394736842105263157894736842105263157895,0.0394736842105263157894736842105263157895,-0.151315789473684210526315789473684210526,-0.348684210526315789473684210526315789474,0.651315789473684210526315789473684210526,
                   0.348684210526315789473684210526315789474,-0.370614035087719298245614035087719298246,0.037280701754385964912280701754385964912,-0.1293859649122807017543859649122807017544,-0.0372807017543859649122807017543859649123,
                   -0.078947368421052631578947368421052631579,0.078947368421052631578947368421052631579,0.0789473684210526315789473684210526315789,-0.0789473684210526315789473684210526315789,-0.151315789473684210526315789473684210526,
                   0.651315789473684210526315789473684210526,-0.348684210526315789473684210526315789474,0.348684210526315789473684210526315789474,-0.078947368421052631578947368421052631579,0.078947368421052631578947368421052631579,
                   0.0789473684210526315789473684210526315789,-0.0789473684210526315789473684210526315789,-0.370614035087719298245614035087719298246,0.037280701754385964912280701754385964912,-0.1293859649122807017543859649122807017544,
                   -0.0372807017543859649122807017543859649123,-0.625,-0.375,0.125,0.375,0.375,0.125,0.125,-0.125,0,0,0,0,-0.447368421052631578947368421052631578947,-0.0526315789473684210526315789473684210526,-0.0526315789473684210526315789473684210526,
                   0.552631578947368421052631578947368421053,0.407894736842105263157894736842105263158,-0.407894736842105263157894736842105263158,0.0921052631578947368421052631578947368421,-0.0921052631578947368421052631578947368421,
                   0.407894736842105263157894736842105263158,-0.407894736842105263157894736842105263158,0.0921052631578947368421052631578947368421,-0.0921052631578947368421052631578947368421,-0.625,0.125,-0.375,0.375,0,0,0,0,0.375,0.125,0.125,-0.125,0.25,0.25,-0.25,-0.25,
                   -0.0833333333333333333333333333333333333333,-0.0833333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,0,0,0,0,0.25,-0.25,0.25,-0.25,-0.25,0.25,-0.25,0.25,0,0,0,0,0.25,0.25,-0.25,
                   -0.25,0,0,0,0,-0.25,0.25,-0.25,0.25,0.25,-0.25,0.25,-0.25,0,0,0,0,-0.0833333333333333333333333333333333333333,-0.0833333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,0.0833333333333333333333333333333333333333};
                return data;
            }

            static const double* N2_polyDegree4Inverse()
            {
                static const double data[15*24] = {0.360647482014388489208633093525179856115,0.23,0.23,0.1793525179856115107913669064748201438849,0.143633093525179856115107913669064748201,0.0293884892086330935251798561151079136691,-0.0993884892086330935251798561151079136691,
                   -0.0736330935251798561151079136690647482014,0.143633093525179856115107913669064748201,0.0293884892086330935251798561151079136691,-0.0993884892086330935251798561151079136691,-0.0736330935251798561151079136690647482014,
                   -0.0699640287769784172661870503597122302158,0.0285971223021582733812949640287769784173,0.0214028776978417266187050359712230215827,0.0199640287769784172661870503597122302158,-0.0699640287769784172661870503597122302158,
                   0.0285971223021582733812949640287769784173,0.0214028776978417266187050359712230215827,0.0199640287769784172661870503597122302158,0.0638129496402877697841726618705035971223,-0.03,-0.03,-0.0038129496402877697841726618705035971223,
                   0.0600299760191846522781774580335731414868,0.1084952038369304556354916067146282973621,0.2965047961630695443645083932853717026379,0.2016366906474820143884892086330935251799,-0.411375186564035484898794251312236923748,
                   -0.2813406407740940115040834465294896949573,0.0343068412402944777045496469956901611578,-0.0915910139021649813016719491539635424528,-0.079466867065787929097281615267226778018,0.1252402483607519578742600325334138283779,
                   -0.0004791760996796968019989602723415673056,-0.0452942051952843319749794569938454830541,0.0626984077074005131559088393620767721487,0.0384706570408009256930120239472757458369,-0.0736396547097985946906810216162734148346,
                   0.0558039232949304891750934916402542301823,0.0897460004024752226191075111938421290939,-0.092893482416863711827740604719021985209,0.0000321770555583505223792993577166239037,0.0031153049588301386862537941674632322114,
                   -0.1031736429038587311968606932635709614127,0.0731376716808371484630477436232831916285,0.0400441465009810333551340745585349901897,-0.0100081752779594506213211249182472204055,0.0600299760191846522781774580335731414868,
                   0.2965047961630695443645083932853717026379,0.1084952038369304556354916067146282973621,0.2016366906474820143884892086330935251799,-0.079466867065787929097281615267226778018,0.1252402483607519578742600325334138283779,
                   -0.0004791760996796968019989602723415673056,-0.0452942051952843319749794569938454830541,-0.411375186564035484898794251312236923748,-0.2813406407740940115040834465294896949573,0.0343068412402944777045496469956901611578,
                   -0.0915910139021649813016719491539635424528,0.0897460004024752226191075111938421290939,-0.092893482416863711827740604719021985209,0.0000321770555583505223792993577166239037,0.0031153049588301386862537941674632322114,
                   0.0626984077074005131559088393620767721487,0.0384706570408009256930120239472757458369,-0.0736396547097985946906810216162734148346,0.0558039232949304891750934916402542301823,-0.1031736429038587311968606932635709614127,
                   0.0400441465009810333551340745585349901897,0.0731376716808371484630477436232831916285,-0.0100081752779594506213211249182472204055,-0.1946642685851318944844124700239808153477,-0.1464628297362110311750599520383693045564,
                   -0.1202038369304556354916067146282973621103,-0.1220023980815347721822541966426858513189,0.0967424661669265985812748402676460230417,0.1307755194445841927856316345524978618504,0.2178258791568144086129697640489007395482,
                   0.1796561352316748000201237611309553755597,-0.058559830457312471700960909594003119183,0.0127864491623484429239824923278160688233,0.03293033405443477385923429088896714796,0.0128430472405292549177441263772199023998,
                   0.0854809159665274773188442253190454629304,-0.0366200046955442638895876305948248394292,-0.0537063356307960624507387097315154869112,-0.0368212423068538176451845516593718032567,0.0232979700155959148764904160587613825024,
                   -0.0201504880012074256678573225335815263873,-0.0012655959148764904160587613825023896966,-0.0018818860995119987925743321426774664185,-0.001953891432308698495748855461085676913,0.0319898626553302812295618051013734466972,
                   -0.0115353172007848266841072596468279921517,-0.0185006540222367560497056899934597776324,0.0794304556354916067146282973621103117506,-0.0008333333333333333333333333333333333333,-0.0008333333333333333333333333333333333333,
                   0.1722362110311750599520383693045563549161,0.0065222367560497056899934597776324395029,-0.2230689993459777632439502943100065402224,-0.1273855461085676913015042511445389143231,0.0939323086984957488554610856769130150425,
                   0.0065222367560497056899934597776324395029,-0.2230689993459777632439502943100065402224,-0.1273855461085676913015042511445389143231,0.0939323086984957488554610856769130150425,-0.0139415740135164595596250272509265315021,
                   0.0202365380422934379768912143012862437323,0.1191574013516459559625027250926531502071,-0.1254523653804229343797689121430128624373,-0.0139415740135164595596250272509265315021,0.0202365380422934379768912143012862437323,
                   0.1191574013516459559625027250926531502071,-0.1254523653804229343797689121430128624373,0.2289355788096795291039895356442119032047,-0.0438636363636363636363636363636363636364,-0.0438636363636363636363636363636363636364,
                   0.108791693917593198168737737083060824068,-0.1946642685851318944844124700239808153477,-0.1202038369304556354916067146282973621103,-0.1464628297362110311750599520383693045564,-0.1220023980815347721822541966426858513189,
                   -0.058559830457312471700960909594003119183,0.0127864491623484429239824923278160688233,0.03293033405443477385923429088896714796,0.0128430472405292549177441263772199023998,0.0967424661669265985812748402676460230417,
                   0.1307755194445841927856316345524978618504,0.2178258791568144086129697640489007395482,0.1796561352316748000201237611309553755597,0.0232979700155959148764904160587613825024,-0.0201504880012074256678573225335815263873,
                   -0.0012655959148764904160587613825023896966,-0.0018818860995119987925743321426774664185,0.0854809159665274773188442253190454629304,-0.0366200046955442638895876305948248394292,-0.0537063356307960624507387097315154869112,
                   -0.0368212423068538176451845516593718032567,-0.001953891432308698495748855461085676913,-0.0115353172007848266841072596468279921517,0.0319898626553302812295618051013734466972,-0.0185006540222367560497056899934597776324,
                   -0.0168465227817745803357314148681055155875,-0.0335131894484412470023980815347721822542,-0.0664868105515587529976019184652278177458,-0.0498201438848920863309352517985611510791,0.1028611376632959366772316412604182388355,
                   0.0787935385956968690781640421928191712364,0.0221388623367040633227683587395817611645,0.0462064614043031309218359578071808287636,0.0237820932065536382083144673072730626687,-0.0237820932065536382083144673072730626687,
                   -0.0066957622713018396471633881705824151867,0.0066957622713018396471633881705824151867,-0.0672290838657745132565276450168536499472,-0.0209004125371031845851989736881823212758,0.0255624171991078465898609783501869832805,
                   -0.0207662541295634820814676929784843453908,-0.0144828193389344468481159128641143029632,0.0144828193389344468481159128641143029632,-0.0002054300615451694588385235867250255739,0.0002054300615451694588385235867250255739,
                   0.0113145846958796599084368868541530412034,-0.0113145846958796599084368868541530412034,-0.0023217789404839764551994767822105951602,0.00232177894048397645519947678221059516024,-0.0142985611510791366906474820143884892086,
                   -0.1107014388489208633093525179856115107914,-0.0642985611510791366906474820143884892086,-0.0607014388489208633093525179856115107914,0.0965330532776575942043567942848518388087,0.0284669467223424057956432057151481611913,
                   0.0243302560748603914071539970820546360115,0.1006697439251396085928460029179453639885,0.1539901645117472455601952004829702671429,0.0625565980781808119937616340494038335765,-0.0289901645117472455601952004829702671429,
                   0.0624434019218191880062383659505961664235,-0.1221009206620717412084318559138703023595,0.1221009206620717412084318559138703023595,0.0168850933239422448055541580721436836545,-0.0168850933239422448055541580721436836545,
                   -0.024563565930472405292549177441263772199,0.0182686019016954268752829903909040599688,0.024563565930472405292549177441263772199,-0.0182686019016954268752829903909040599688,-0.0720814257684761281883584041857423152387,
                   -0.1129905166775670372792674950948332243296,-0.0529185742315238718116415958142576847613,-0.0120094833224329627207325049051667756704,-0.0142985611510791366906474820143884892086,-0.0642985611510791366906474820143884892086,
                   -0.1107014388489208633093525179856115107914,-0.0607014388489208633093525179856115107914,0.1539901645117472455601952004829702671429,0.0625565980781808119937616340494038335765,-0.0289901645117472455601952004829702671429,
                   0.0624434019218191880062383659505961664235,0.0965330532776575942043567942848518388087,0.0284669467223424057956432057151481611913,0.0243302560748603914071539970820546360115,0.1006697439251396085928460029179453639885,
                   -0.024563565930472405292549177441263772199,0.0182686019016954268752829903909040599688,0.024563565930472405292549177441263772199,-0.0182686019016954268752829903909040599688,-0.1221009206620717412084318559138703023595,
                   0.1221009206620717412084318559138703023595,0.0168850933239422448055541580721436836545,-0.0168850933239422448055541580721436836545,-0.0720814257684761281883584041857423152387,-0.0529185742315238718116415958142576847613,
                   -0.1129905166775670372792674950948332243296,-0.0120094833224329627207325049051667756704,-0.0168465227817745803357314148681055155875,-0.0664868105515587529976019184652278177458,-0.0335131894484412470023980815347721822542,
                   -0.0498201438848920863309352517985611510791,0.0237820932065536382083144673072730626687,-0.0237820932065536382083144673072730626687,-0.0066957622713018396471633881705824151867,0.0066957622713018396471633881705824151867,
                   0.1028611376632959366772316412604182388355,0.0787935385956968690781640421928191712364,0.0221388623367040633227683587395817611645,0.0462064614043031309218359578071808287636,-0.0144828193389344468481159128641143029632,
                   0.0144828193389344468481159128641143029632,-0.0002054300615451694588385235867250255739,0.0002054300615451694588385235867250255739,-0.0672290838657745132565276450168536499472,-0.0209004125371031845851989736881823212758,
                   0.0255624171991078465898609783501869832805,-0.0207662541295634820814676929784843453908,0.0113145846958796599084368868541530412034,-0.0023217789404839764551994767822105951602,-0.0113145846958796599084368868541530412034,
                   0.00232177894048397645519947678221059516024,0.02083333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,0.02083333333333333333333333333333333333333,-0.03125,-0.03125,-0.03125,
                   -0.03125,0,0,0,0,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,-0.0166666666666666666666666666666666666667,
                   0.01666666666666666666666666666666666666667,0.01666666666666666666666666666666666666667,-0.0166666666666666666666666666666666666667,-0.0240675990675990675990675990675990675991,0.0240675990675990675990675990675990675991,
                   0.0240675990675990675990675990675990675991,-0.0240675990675990675990675990675990675991,-0.0304778554778554778554778554778554778555,0.0304778554778554778554778554778554778555,0.0304778554778554778554778554778554778555,
                   -0.0304778554778554778554778554778554778555,0.0463286713286713286713286713286713286713,-0.0463286713286713286713286713286713286713,-0.0463286713286713286713286713286713286713,0.0463286713286713286713286713286713286713,
                   0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,0.0142773892773892773892773892773892773893,-0.0136363636363636363636363636363636363636,
                   0.0136363636363636363636363636363636363636,0.0136363636363636363636363636363636363636,-0.0136363636363636363636363636363636363636,0.0625,0.0625,0.0625,0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,0,0,0,0,0,0,0,0,0.0625,0.0625,
                   0.0625,0.0625,-0.0166666666666666666666666666666666666667,0.01666666666666666666666666666666666666667,0.01666666666666666666666666666666666666667,-0.0166666666666666666666666666666666666667,-0.0304778554778554778554778554778554778555,
                   0.0304778554778554778554778554778554778555,0.0304778554778554778554778554778554778555,-0.0304778554778554778554778554778554778555,-0.0240675990675990675990675990675990675991,0.0240675990675990675990675990675990675991,
                   0.0240675990675990675990675990675990675991,-0.0240675990675990675990675990675990675991,0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,
                   0.0142773892773892773892773892773892773893,0.0463286713286713286713286713286713286713,-0.0463286713286713286713286713286713286713,-0.0463286713286713286713286713286713286713,0.0463286713286713286713286713286713286713,
                   -0.0136363636363636363636363636363636363636,0.0136363636363636363636363636363636363636,0.0136363636363636363636363636363636363636,-0.0136363636363636363636363636363636363636,0.02083333333333333333333333333333333333333,
                   0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,0.02083333333333333333333333333333333333333,0,0,0,0,-0.03125,-0.03125,-0.03125,-0.03125,0,0,0,0,0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0};
                return data;
            }

            static const double* N2_polyDegree5Inverse()
            {
                static const double data[21*24] = {0.77,0.23,0.23,-0.23,0.035,-0.035,-0.035,0.035,0.035,-0.035,-0.035,0.035,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.025,-0.025,0.03,-0.03,-0.03,0.03,-0.0358333333333333333333333333333333333333,-0.2975,0.7025,0.2975,
                   -0.5014831002331002331002331002331002331,0.0014831002331002331002331002331002331,-0.2485168997668997668997668997668997669,-0.0014831002331002331002331002331002331002,-0.062380536130536130536130536130536130536,0.062380536130536130536130536130536130536,
                   0.062380536130536130536130536130536130536,-0.0623805361305361305361305361305361305361,0.0842511655011655011655011655011655011655,-0.0342511655011655011655011655011655011655,-0.0009178321678321678321678321678321678322,
                   0.0342511655011655011655011655011655011655,0.0464306526806526806526806526806526806527,-0.0464306526806526806526806526806526806527,-0.0464306526806526806526806526806526806527,0.0464306526806526806526806526806526806527,
                   -0.0565909090909090909090909090909090909091,0.0565909090909090909090909090909090909091,0.0565909090909090909090909090909090909091,-0.0565909090909090909090909090909090909091,-0.0358333333333333333333333333333333333333,0.7025,-0.2975,0.2975,
                   -0.062380536130536130536130536130536130536,0.062380536130536130536130536130536130536,0.062380536130536130536130536130536130536,-0.0623805361305361305361305361305361305361,-0.5014831002331002331002331002331002331,0.0014831002331002331002331002331002331,
                   -0.2485168997668997668997668997668997669,-0.0014831002331002331002331002331002331002,0.0464306526806526806526806526806526806527,-0.0464306526806526806526806526806526806527,-0.0464306526806526806526806526806526806527,
                   0.0464306526806526806526806526806526806527,0.0842511655011655011655011655011655011655,-0.0342511655011655011655011655011655011655,-0.0009178321678321678321678321678321678322,0.0342511655011655011655011655011655011655,
                   -0.0565909090909090909090909090909090909091,0.0565909090909090909090909090909090909091,0.0565909090909090909090909090909090909091,-0.0565909090909090909090909090909090909091,-0.783333333333333333333333333333333333333,
                   -0.466666666666666666666666666666666666667,0.2,0.466666666666666666666666666666666666667,0.471532634032634032634032634032634032634,0.195134032634032634032634032634032634033,0.153467365967365967365967365967365967366,
                   -0.195134032634032634032634032634032634033,-0.022858391608391608391608391608391608392,0.022858391608391608391608391608391608392,0.022858391608391608391608391608391608392,-0.022858391608391608391608391608391608392,
                   0.0034965034965034965034965034965034965035,-0.0451631701631701631701631701631701631702,-0.0451631701631701631701631701631701631702,0.0451631701631701631701631701631701631702,0.010708041958041958041958041958041958042,
                   -0.010708041958041958041958041958041958042,-0.010708041958041958041958041958041958042,0.010708041958041958041958041958041958042,-0.0102272727272727272727272727272727272727,0.0102272727272727272727272727272727272727,
                   0.0102272727272727272727272727272727272727,-0.0102272727272727272727272727272727272727,-0.415833333333333333333333333333333333333,-0.0008333333333333333333333333333333333333,-0.0008333333333333333333333333333333333333,0.6675,
                   0.341893939393939393939393939393939393939,-0.508560606060606060606060606060606060606,0.1581060606060606060606060606060606060606,-0.241439393939393939393939393939393939394,0.341893939393939393939393939393939393939,
                   -0.508560606060606060606060606060606060606,0.1581060606060606060606060606060606060606,-0.241439393939393939393939393939393939394,-0.1113636363636363636363636363636363636364,0.1113636363636363636363636363636363636364,
                   0.028030303030303030303030303030303030303,-0.028030303030303030303030303030303030303,-0.1113636363636363636363636363636363636364,0.1113636363636363636363636363636363636364,0.028030303030303030303030303030303030303,
                   -0.028030303030303030303030303030303030303,0.210530303030303030303030303030303030303,-0.0438636363636363636363636363636363636364,-0.0438636363636363636363636363636363636364,0.1271969696969696969696969696969696969697,
                   -0.783333333333333333333333333333333333333,0.2,-0.466666666666666666666666666666666666667,0.466666666666666666666666666666666666667,-0.022858391608391608391608391608391608392,0.022858391608391608391608391608391608392,
                   0.022858391608391608391608391608391608392,-0.022858391608391608391608391608391608392,0.471532634032634032634032634032634032634,0.195134032634032634032634032634032634033,0.153467365967365967365967365967365967366,
                   -0.195134032634032634032634032634032634033,0.010708041958041958041958041958041958042,-0.010708041958041958041958041958041958042,-0.010708041958041958041958041958041958042,0.010708041958041958041958041958041958042,
                   0.0034965034965034965034965034965034965035,-0.0451631701631701631701631701631701631702,-0.0451631701631701631701631701631701631702,0.0451631701631701631701631701631701631702,-0.0102272727272727272727272727272727272727,
                   0.0102272727272727272727272727272727272727,0.0102272727272727272727272727272727272727,-0.0102272727272727272727272727272727272727,0.175,0.241666666666666666666666666666666666667,-0.341666666666666666666666666666666666667,
                   -0.2416666666666666666666666666666666666667,0.053700466200466200466200466200466200466,-0.0953671328671328671328671328671328671329,0.1962995337995337995337995337995337995338,0.0953671328671328671328671328671328671329,
                   0.015238927738927738927738927738927738928,-0.0152389277389277389277389277389277389277,-0.0152389277389277389277389277389277389277,0.0152389277389277389277389277389277389277,-0.0648310023310023310023310023310023310023,
                   0.0231643356643356643356643356643356643357,-0.018502331002331002331002331002331002331,-0.0231643356643356643356643356643356643357,-0.0071386946386946386946386946386946386946,0.0071386946386946386946386946386946386946,
                   0.0071386946386946386946386946386946386946,-0.0071386946386946386946386946386946386946,0.0068181818181818181818181818181818181818,-0.0068181818181818181818181818181818181818,-0.0068181818181818181818181818181818181818,
                   0.0068181818181818181818181818181818181818,0.0041666666666666666666666666666666666667,-0.504166666666666666666666666666666666667,0.329166666666666666666666666666666666667,-0.0791666666666666666666666666666666666667,
                   -0.088898601398601398601398601398601398601,0.338898601398601398601398601398601398601,-0.286101398601398601398601398601398601399,0.2861013986013986013986013986013986013986,0.316550116550116550116550116550116550117,
                   -0.14988344988344988344988344988344988345,0.183449883449883449883449883449883449883,-0.1001165501165501165501165501165501165501,-0.0486596736596736596736596736596736596737,0.0486596736596736596736596736596736596737,
                   0.0903263403263403263403263403263403263403,-0.0903263403263403263403263403263403263403,-0.0214160839160839160839160839160839160839,0.0214160839160839160839160839160839160839,0.0214160839160839160839160839160839160839,
                   -0.0214160839160839160839160839160839160839,-0.1253787878787878787878787878787878787879,-0.0412878787878787878787878787878787878788,-0.1246212121212121212121212121212121212121,0.0412878787878787878787878787878787878788,
                   0.0041666666666666666666666666666666666667,0.329166666666666666666666666666666666667,-0.504166666666666666666666666666666666667,-0.0791666666666666666666666666666666666667,0.316550116550116550116550116550116550117,
                   -0.14988344988344988344988344988344988345,0.183449883449883449883449883449883449883,-0.1001165501165501165501165501165501165501,-0.088898601398601398601398601398601398601,0.338898601398601398601398601398601398601,
                   -0.286101398601398601398601398601398601399,0.2861013986013986013986013986013986013986,-0.0214160839160839160839160839160839160839,0.0214160839160839160839160839160839160839,0.0214160839160839160839160839160839160839,
                   -0.0214160839160839160839160839160839160839,-0.0486596736596736596736596736596736596737,0.0486596736596736596736596736596736596737,0.0903263403263403263403263403263403263403,-0.0903263403263403263403263403263403263403,
                   -0.1253787878787878787878787878787878787879,-0.1246212121212121212121212121212121212121,-0.0412878787878787878787878787878787878788,0.0412878787878787878787878787878787878788,0.175,-0.341666666666666666666666666666666666667,
                   0.241666666666666666666666666666666666667,-0.2416666666666666666666666666666666666667,0.015238927738927738927738927738927738928,-0.0152389277389277389277389277389277389277,-0.0152389277389277389277389277389277389277,
                   0.0152389277389277389277389277389277389277,0.053700466200466200466200466200466200466,-0.0953671328671328671328671328671328671329,0.1962995337995337995337995337995337995338,0.0953671328671328671328671328671328671329,
                   -0.0071386946386946386946386946386946386946,0.0071386946386946386946386946386946386946,0.0071386946386946386946386946386946386946,-0.0071386946386946386946386946386946386946,-0.0648310023310023310023310023310023310023,
                   0.0231643356643356643356643356643356643357,-0.018502331002331002331002331002331002331,-0.0231643356643356643356643356643356643357,0.0068181818181818181818181818181818181818,-0.0068181818181818181818181818181818181818,
                   -0.0068181818181818181818181818181818181818,0.0068181818181818181818181818181818181818,0.1458333333333333333333333333333333333333,0.1041666666666666666666666666666666666667,-0.0625,-0.1041666666666666666666666666666666666667,
                   -0.1145833333333333333333333333333333333333,-0.0520833333333333333333333333333333333333,-0.0104166666666666666666666666666666666667,0.0520833333333333333333333333333333333333,0,0,0,0,0.03125,0.0104166666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0.191666666666666666666666666666666666667,0.0583333333333333333333333333333333333333,-0.025,-0.225,-0.1907342657342657342657342657342657342657,
                   0.1074009324009324009324009324009324009324,-0.0592657342657342657342657342657342657343,0.1425990675990675990675990675990675990676,-0.1554778554778554778554778554778554778555,0.1554778554778554778554778554778554778555,
                   -0.0945221445221445221445221445221445221445,0.0945221445221445221445221445221445221445,0.087995337995337995337995337995337995338,-0.087995337995337995337995337995337995338,-0.0046620046620046620046620046620046620047,
                   0.0046620046620046620046620046620046620047,0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,0.0142773892773892773892773892773892773893,
                   0.028030303030303030303030303030303030303,-0.028030303030303030303030303030303030303,0.0553030303030303030303030303030303030303,-0.0553030303030303030303030303030303030303,0.4375,0.0625,0.0625,-0.3125,-0.3125,0.0625,-0.1875,0.1875,-0.3125,0.0625,
                   -0.1875,0.1875,0,0,0,0,0,0,0,0,0.1875,0.0625,0.0625,-0.0625,0.191666666666666666666666666666666666667,-0.025,0.0583333333333333333333333333333333333333,-0.225,-0.1554778554778554778554778554778554778555,0.1554778554778554778554778554778554778555,
                   -0.0945221445221445221445221445221445221445,0.0945221445221445221445221445221445221445,-0.1907342657342657342657342657342657342657,0.1074009324009324009324009324009324009324,-0.0592657342657342657342657342657342657343,
                   0.1425990675990675990675990675990675990676,0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,-0.0142773892773892773892773892773892773893,0.0142773892773892773892773892773892773893,
                   0.087995337995337995337995337995337995338,-0.087995337995337995337995337995337995338,-0.0046620046620046620046620046620046620047,0.0046620046620046620046620046620046620047,0.028030303030303030303030303030303030303,
                   0.0553030303030303030303030303030303030303,-0.028030303030303030303030303030303030303,-0.0553030303030303030303030303030303030303,0.1458333333333333333333333333333333333333,-0.0625,0.1041666666666666666666666666666666666667,
                   -0.1041666666666666666666666666666666666667,0,0,0,0,-0.1145833333333333333333333333333333333333,-0.0520833333333333333333333333333333333333,-0.0104166666666666666666666666666666666667,0.0520833333333333333333333333333333333333,0,0,0,0,0.03125,
                   0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0,0,0,0,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,
                   0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,-0.0208333333333333333333333333333333333333,
                   -0.0208333333333333333333333333333333333333,0,0,0,0,-0.00416666666666666666666666666666666666667,-0.00416666666666666666666666666666666666667,0.00416666666666666666666666666666666666667,0.00416666666666666666666666666666666666667,0,0,0,0,0,0,0,0,
                   -0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0625,-0.0625,0.0625,-0.0625,0,0,0,0,
                   -0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,-0.02083333333333333333333333333333333333333,0.02083333333333333333333333333333333333333,0,0,0,0,0,0,0,0,-0.125,-0.125,0.125,0.125,
                   0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,0.125,-0.125,0.125,-0.125,0,0,0,0,0,0,0,0,
                   -0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.125,0.125,-0.125,0.125,0.125,-0.125,0.125,-0.125,
                   0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,0,0,0,0,0,0,0,0,-0.0416666666666666666666666666666666666667,
                   -0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,
                   0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0,0,0,0,0.0625,-0.0625,0.0625,-0.0625,0,0,0,0,-0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,
                   -0.02083333333333333333333333333333333333333,0.02083333333333333333333333333333333333333,0,0,0,0,-0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,
                   0.0416666666666666666666666666666666666667,0,0,0,0,0.0208333333333333333333333333333333333333,0.0208333333333333333333333333333333333333,-0.0208333333333333333333333333333333333333,-0.0208333333333333333333333333333333333333,0,0,0,0,
                   -0.00416666666666666666666666666666666666667,-0.00416666666666666666666666666666666666667,0.00416666666666666666666666666666666666667,0.00416666666666666666666666666666666666667,0,0,0,0};
                return data;
            }

            static const double* N2_bicubicInverse()
            {
                static const double data[16*16] = 
                {
                    0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 1, -2.5, 2, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 1.5, -1.5, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 
                    0.25, 0, -0.25, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, 
                    -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 
                    0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, 
                    0, 1, 0, 0, 0, -2.5, 0, 0, 0, 2, 0, 0, 0, -0.5, 0, 0, 
                    -0.5, 0, 0.5, 0, 1.25, 0, -1.25, 0, -1, 0, 1, 0, 0.25, 0, -0.25, 0, 
                    1, -2.5, 2, -0.5, -2.5, 6.25, -5, 1.25, 2, -5, 4, -1, -0.5, 1.25, -1, 0.25, 
                    -0.5, 1.5, -1.5, 0.5, 1.25, -3.75, 3.75, -1.25, -1, 3, -3, 1, 0.25, -0.75, 0.75, -0.25, 
                    0, -0.5, 0, 0, 0, 1.5, 0, 0, 0, -1.5, 0, 0, 0, 0.5, 0, 0, 
                    0.25, 0, -0.25, 0, -0.75, 0, 0.75, 0, 0.75, 0, -0.75, 0, -0.25, 0, 0.25, 0, 
                    -0.5, 1.25, -1, 0.25, 1.5, -3.75, 3, -0.75, -1.5, 3.75, -3, 0.75, 0.5, -1.25, 1, -0.25, 
                    0.25, -0.75, 0.75, -0.25, -0.75, 2.25, -2.25, 0.75, 0.75, -2.25, 2.25, -0.75, -0.25, 0.75, -0.75, 0.25
                };
                return data;
            }

            // A 32-point stencil in three dimensions, suitable for obtaining a least-squares polynomial of degree two and three
            static const uvector<int,3>* N3_stencil32points()
            {
                static const uvector<int,3> data[32] =
                {
                    // Vertices of cube
                    uvector<int,3>(0, 0, 0),
                    uvector<int,3>(0, 0, 1),
                    uvector<int,3>(0, 1, 0),
                    uvector<int,3>(0, 1, 1),
                    uvector<int,3>(1, 0, 0),
                    uvector<int,3>(1, 0, 1),
                    uvector<int,3>(1, 1, 0),
                    uvector<int,3>(1, 1, 1),
                    // plus vertices of neighbouring faces
                    uvector<int,3>(-1, 0, 0),
                    uvector<int,3>(-1, 0, 1),
                    uvector<int,3>(-1, 1, 0),
                    uvector<int,3>(-1, 1, 1),
                    uvector<int,3>(2, 0, 0),
                    uvector<int,3>(2, 0, 1),
                    uvector<int,3>(2, 1, 0),
                    uvector<int,3>(2, 1, 1),
                    uvector<int,3>(0, -1, 0),
                    uvector<int,3>(0, -1, 1),
                    uvector<int,3>(1, -1, 0),
                    uvector<int,3>(1, -1, 1),
                    uvector<int,3>(0, 2, 0),
                    uvector<int,3>(0, 2, 1),
                    uvector<int,3>(1, 2, 0),
                    uvector<int,3>(1, 2, 1),
                    uvector<int,3>(0, 0, -1),
                    uvector<int,3>(0, 1, -1),
                    uvector<int,3>(1, 0, -1),
                    uvector<int,3>(1, 1, -1),
                    uvector<int,3>(0, 0, 2),
                    uvector<int,3>(0, 1, 2),
                    uvector<int,3>(1, 0, 2),
                    uvector<int,3>(1, 1, 2)
                };
                return data;
            }

            // An 88-point stencil in three dimensions, suitable for obtaining a least-squares polynomial of degree four and five
            static const uvector<int,3>* N3_stencil88points()
            {
                static const uvector<int,3> data[88] =
                {
                    uvector<int,3>(-1, -1, -1), uvector<int,3>(-1, -1, 0), uvector<int,3>(-1, -1, 1), uvector<int,3>(-1, -1, 2),
                    uvector<int,3>(-1, 0, -1), uvector<int,3>(-1, 0, 0), uvector<int,3>(-1, 0, 1), uvector<int,3>(-1, 0, 2),
                    uvector<int,3>(-1, 1, -1), uvector<int,3>(-1, 1, 0), uvector<int,3>(-1, 1, 1), uvector<int,3>(-1, 1, 2),
                    uvector<int,3>(-1, 2, -1), uvector<int,3>(-1, 2, 0), uvector<int,3>(-1, 2, 1), uvector<int,3>(-1, 2, 2),
                    uvector<int,3>(0, -1, -1), uvector<int,3>(0, -1, 0), uvector<int,3>(0, -1, 1), uvector<int,3>(0, -1, 2),
                    uvector<int,3>(0, 0, -1), uvector<int,3>(0, 0, 0), uvector<int,3>(0, 0, 1), uvector<int,3>(0, 0, 2),
                    uvector<int,3>(0, 1, -1), uvector<int,3>(0, 1, 0), uvector<int,3>(0, 1, 1), uvector<int,3>(0, 1, 2),
                    uvector<int,3>(0, 2, -1), uvector<int,3>(0, 2, 0), uvector<int,3>(0, 2, 1), uvector<int,3>(0, 2, 2),
                    uvector<int,3>(1, -1, -1), uvector<int,3>(1, -1, 0), uvector<int,3>(1, -1, 1), uvector<int,3>(1, -1, 2),
                    uvector<int,3>(1, 0, -1), uvector<int,3>(1, 0, 0), uvector<int,3>(1, 0, 1), uvector<int,3>(1, 0, 2),
                    uvector<int,3>(1, 1, -1), uvector<int,3>(1, 1, 0), uvector<int,3>(1, 1, 1), uvector<int,3>(1, 1, 2),
                    uvector<int,3>(1, 2, -1), uvector<int,3>(1, 2, 0), uvector<int,3>(1, 2, 1), uvector<int,3>(1, 2, 2),
                    uvector<int,3>(2, -1, -1), uvector<int,3>(2, -1, 0), uvector<int,3>(2, -1, 1), uvector<int,3>(2, -1, 2),
                    uvector<int,3>(2, 0, -1), uvector<int,3>(2, 0, 0), uvector<int,3>(2, 0, 1), uvector<int,3>(2, 0, 2),
                    uvector<int,3>(2, 1, -1), uvector<int,3>(2, 1, 0), uvector<int,3>(2, 1, 1), uvector<int,3>(2, 1, 2),
                    uvector<int,3>(2, 2, -1), uvector<int,3>(2, 2, 0), uvector<int,3>(2, 2, 1), uvector<int,3>(2, 2, 2),
                    uvector<int,3>(-2, 0, 0), uvector<int,3>(-2, 0, 1), uvector<int,3>(-2, 1, 0), uvector<int,3>(-2, 1, 1),
                    uvector<int,3>(3, 0, 0), uvector<int,3>(3, 0, 1), uvector<int,3>(3, 1, 0), uvector<int,3>(3, 1, 1),
                    uvector<int,3>(0, -2, 0), uvector<int,3>(0, -2, 1), uvector<int,3>(1, -2, 0), uvector<int,3>(1, -2, 1),
                    uvector<int,3>(0, 3, 0), uvector<int,3>(0, 3, 1), uvector<int,3>(1, 3, 0), uvector<int,3>(1, 3, 1),
                    uvector<int,3>(0, 0, -2), uvector<int,3>(0, 1, -2), uvector<int,3>(1, 0, -2), uvector<int,3>(1, 1, -2),
                    uvector<int,3>(0, 0, 3), uvector<int,3>(0, 1, 3), uvector<int,3>(1, 0, 3), uvector<int,3>(1, 1, 3)
                };
                return data;
            }

            static const double* N3_polyDegree2Inverse()
            {
                static const double data[10*32] = {0.175,0.1291666666666666666666666666666666666667,0.1291666666666666666666666666666666666667,0.1083333333333333333333333333333333333333,0.1291666666666666666666666666666666666667,0.1083333333333333333333333333333333333333,
                   0.1083333333333333333333333333333333333333,0.1125,0.0958333333333333333333333333333333333333,0.025,0.025,-0.0208333333333333333333333333333333333333,-0.0416666666666666666666666666666666666667,-0.0375,-0.0375,-0.0083333333333333333333333333333333333333,
                   0.0958333333333333333333333333333333333333,0.025,0.025,-0.0208333333333333333333333333333333333333,-0.0416666666666666666666666666666666666667,-0.0375,-0.0375,-0.0083333333333333333333333333333333333333,0.0958333333333333333333333333333333333333,0.025,
                   0.025,-0.0208333333333333333333333333333333333333,-0.0416666666666666666666666666666666666667,-0.0375,-0.0375,-0.0083333333333333333333333333333333333333,0.0166666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,
                   0.0416666666666666666666666666666666666667,0.0666666666666666666666666666666666666667,0.1083333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,
                   0.0583333333333333333333333333333333333333,-0.2,-0.125,-0.125,-0.05,0.075,0,0,-0.075,-0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0708333333333333333333333333333333333333,
                   0.0458333333333333333333333333333333333333,0.0041666666666666666666666666666666666667,0.0291666666666666666666666666666666666667,-0.0041666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,
                   -0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0708333333333333333333333333333333333333,0.0458333333333333333333333333333333333333,0.0041666666666666666666666666666666666667,
                   0.0291666666666666666666666666666666666667,-0.0041666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,0.0166666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,
                   0.1083333333333333333333333333333333333333,0.0833333333333333333333333333333333333333,0.0416666666666666666666666666666666666667,0.0666666666666666666666666666666666666667,0.0833333333333333333333333333333333333333,
                   0.0583333333333333333333333333333333333333,-0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0708333333333333333333333333333333333333,0.0458333333333333333333333333333333333333,
                   0.0041666666666666666666666666666666666667,0.0291666666666666666666666666666666666667,-0.0041666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,-0.2,-0.125,-0.125,-0.05,0.075,0,0,-0.075,
                   -0.0708333333333333333333333333333333333333,0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0458333333333333333333333333333333333333,0.0041666666666666666666666666666666666667,
                   -0.0041666666666666666666666666666666666667,0.0291666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,0.0166666666666666666666666666666666666667,0.1083333333333333333333333333333333333333,
                   0.0416666666666666666666666666666666666667,0.0833333333333333333333333333333333333333,0.0416666666666666666666666666666666666667,0.0833333333333333333333333333333333333333,0.0666666666666666666666666666666666666667,
                   0.0583333333333333333333333333333333333333,-0.0708333333333333333333333333333333333333,0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0458333333333333333333333333333333333333,
                   0.0041666666666666666666666666666666666667,-0.0041666666666666666666666666666666666667,0.0291666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,-0.0708333333333333333333333333333333333333,
                   0.0708333333333333333333333333333333333333,-0.0458333333333333333333333333333333333333,0.0458333333333333333333333333333333333333,0.0041666666666666666666666666666666666667,-0.0041666666666666666666666666666666666667,
                   0.0291666666666666666666666666666666666667,-0.0291666666666666666666666666666666666667,-0.2,-0.125,-0.125,-0.05,0.075,0,0,-0.075,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0,0,
                   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.025,0.025,-0.025,-0.025,-0.025,-0.025,0.025,0.025,0.075,0.075,-0.075,-0.075,-0.075,-0.075,0.075,0.075,0.075,0.075,-0.075,-0.075,-0.075,-0.075,0.075,0.075,0.025,-0.025,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.025,-0.025,
                   0.025,-0.025,-0.025,0.025,-0.025,0.025,0.075,-0.075,0.075,-0.075,-0.075,0.075,-0.075,0.075,0.025,-0.025,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.075,0.075,-0.075,-0.075,-0.075,-0.075,0.075,0.075,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,
                   -0.0625,0,0,0,0,0,0,0,0,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0,0,0,0,0,0,0,0,0.025,-0.025,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.025,-0.025,-0.025,0.025,0.075,-0.075,0.075,-0.075,-0.075,0.075,-0.075,0.075,
                   0.075,-0.075,0.075,-0.075,-0.075,0.075,-0.075,0.075,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,-0.0625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625,0.0625};
                return data;
            }

            static const double* N3_polyDegree3Inverse()
            {
                static const double data[20*32] = {0.523214285714285714285714285714285714286,0.2392857142857142857142857142857142857143,0.2392857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,0.2392857142857142857142857142857142857143,
                   -0.0017857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,-0.235714285714285714285714285714285714286,0.0571428571428571428571428571428571428571,-0.0196428571428571428571428571428571428571,
                   -0.0196428571428571428571428571428571428571,-0.0178571428571428571428571428571428571429,-0.0446428571428571428571428571428571428571,0.0071428571428571428571428571428571428571,0.0071428571428571428571428571428571428571,
                   0.0303571428571428571428571428571428571429,0.0571428571428571428571428571428571428571,-0.0196428571428571428571428571428571428571,-0.0196428571428571428571428571428571428571,-0.0178571428571428571428571428571428571429,
                   -0.0446428571428571428571428571428571428571,0.0071428571428571428571428571428571428571,0.0071428571428571428571428571428571428571,0.0303571428571428571428571428571428571429,0.0571428571428571428571428571428571428571,
                   -0.0196428571428571428571428571428571428571,-0.0196428571428571428571428571428571428571,-0.0178571428571428571428571428571428571429,-0.0446428571428571428571428571428571428571,0.0071428571428571428571428571428571428571,
                   0.0071428571428571428571428571428571428571,0.0303571428571428571428571428571428571429,-0.0339285714285714285714285714285714285714,-0.116071428571428571428571428571428571429,-0.1160714285714285714285714285714285714286,
                   -0.233928571428571428571428571428571428571,0.408928571428571428571428571428571428571,0.2410714285714285714285714285714285714286,0.241071428571428571428571428571428571429,0.1089285714285714285714285714285714285714,
                   -0.3101190476190476190476190476190476190476,-0.056547619047619047619047619047619047619,-0.056547619047619047619047619047619047619,0.0898809523809523809523809523809523809524,-0.0648809523809523809523809523809523809524,
                   -0.068452380952380952380952380952380952381,-0.068452380952380952380952380952380952381,0.0351190476190476190476190476190476190476,-0.0767857142857142857142857142857142857143,0.0017857142857142857142857142857142857143,
                   0.0767857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,0.0517857142857142857142857142857142857143,0.0232142857142857142857142857142857142857,-0.0517857142857142857142857142857142857143,
                   -0.0232142857142857142857142857142857142857,-0.0767857142857142857142857142857142857143,0.0017857142857142857142857142857142857143,0.0767857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,
                   0.0517857142857142857142857142857142857143,0.0232142857142857142857142857142857142857,-0.0517857142857142857142857142857142857143,-0.0232142857142857142857142857142857142857,-0.0339285714285714285714285714285714285714,
                   -0.116071428571428571428571428571428571429,0.408928571428571428571428571428571428571,0.2410714285714285714285714285714285714286,-0.1160714285714285714285714285714285714286,-0.233928571428571428571428571428571428571,
                   0.241071428571428571428571428571428571429,0.1089285714285714285714285714285714285714,-0.0767857142857142857142857142857142857143,0.0017857142857142857142857142857142857143,0.0767857142857142857142857142857142857143,
                   -0.0017857142857142857142857142857142857143,0.0517857142857142857142857142857142857143,0.0232142857142857142857142857142857142857,-0.0517857142857142857142857142857142857143,-0.0232142857142857142857142857142857142857,
                   -0.3101190476190476190476190476190476190476,-0.056547619047619047619047619047619047619,-0.056547619047619047619047619047619047619,0.0898809523809523809523809523809523809524,-0.0648809523809523809523809523809523809524,
                   -0.068452380952380952380952380952380952381,-0.068452380952380952380952380952380952381,0.0351190476190476190476190476190476190476,-0.0767857142857142857142857142857142857143,0.0767857142857142857142857142857142857143,
                   0.0017857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,0.0517857142857142857142857142857142857143,-0.0517857142857142857142857142857142857143,0.0232142857142857142857142857142857142857,
                   -0.0232142857142857142857142857142857142857,-0.0339285714285714285714285714285714285714,0.408928571428571428571428571428571428571,-0.116071428571428571428571428571428571429,0.241071428571428571428571428571428571429,
                   -0.116071428571428571428571428571428571429,0.241071428571428571428571428571428571429,-0.233928571428571428571428571428571428571,0.1089285714285714285714285714285714285714,-0.0767857142857142857142857142857142857143,
                   0.0767857142857142857142857142857142857143,0.0017857142857142857142857142857142857143,-0.0017857142857142857142857142857142857143,0.0517857142857142857142857142857142857143,-0.0517857142857142857142857142857142857143,
                   0.0232142857142857142857142857142857142857,-0.0232142857142857142857142857142857142857,-0.0767857142857142857142857142857142857143,0.0767857142857142857142857142857142857143,0.0017857142857142857142857142857142857143,
                   -0.0017857142857142857142857142857142857143,0.0517857142857142857142857142857142857143,-0.0517857142857142857142857142857142857143,0.0232142857142857142857142857142857142857,-0.0232142857142857142857142857142857142857,
                   -0.3101190476190476190476190476190476190476,-0.056547619047619047619047619047619047619,-0.056547619047619047619047619047619047619,0.0898809523809523809523809523809523809524,-0.0648809523809523809523809523809523809524,
                   -0.068452380952380952380952380952380952381,-0.068452380952380952380952380952380952381,0.0351190476190476190476190476190476190476,-0.375,-0.25,-0.25,-0.125,0,0.125,0.125,0.25,0.25,0.125,0.125,0,0.125,0,0,-0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   -0.207142857142857142857142857142857142857,-0.242857142857142857142857142857142857143,-0.042857142857142857142857142857142857143,-0.0071428571428571428571428571428571428571,-0.042857142857142857142857142857142857143,
                   -0.0071428571428571428571428571428571428571,0.292857142857142857142857142857142857143,0.257142857142857142857142857142857142857,0.253571428571428571428571428571428571429,0.1464285714285714285714285714285714285714,
                   -0.2535714285714285714285714285714285714286,-0.1464285714285714285714285714285714285714,-0.0035714285714285714285714285714285714286,0.1035714285714285714285714285714285714286,0.0035714285714285714285714285714285714286,
                   -0.1035714285714285714285714285714285714286,0.253571428571428571428571428571428571429,0.1464285714285714285714285714285714285714,-0.2535714285714285714285714285714285714286,-0.1464285714285714285714285714285714285714,
                   -0.0035714285714285714285714285714285714286,0.1035714285714285714285714285714285714286,0.0035714285714285714285714285714285714286,-0.1035714285714285714285714285714285714286,0.078571428571428571428571428571428571429,
                   -0.0785714285714285714285714285714285714286,-0.0785714285714285714285714285714285714286,0.0785714285714285714285714285714285714286,-0.0285714285714285714285714285714285714286,0.0285714285714285714285714285714285714286,
                   0.0285714285714285714285714285714285714286,-0.0285714285714285714285714285714285714286,-0.207142857142857142857142857142857142857,-0.042857142857142857142857142857142857143,-0.242857142857142857142857142857142857143,
                   -0.0071428571428571428571428571428571428571,-0.042857142857142857142857142857142857143,0.292857142857142857142857142857142857143,-0.0071428571428571428571428571428571428571,0.257142857142857142857142857142857142857,
                   0.253571428571428571428571428571428571429,-0.2535714285714285714285714285714285714286,0.1464285714285714285714285714285714285714,-0.1464285714285714285714285714285714285714,-0.0035714285714285714285714285714285714286,
                   0.0035714285714285714285714285714285714286,0.1035714285714285714285714285714285714286,-0.1035714285714285714285714285714285714286,0.078571428571428571428571428571428571429,-0.0785714285714285714285714285714285714286,
                   -0.0785714285714285714285714285714285714286,0.0785714285714285714285714285714285714286,-0.0285714285714285714285714285714285714286,0.0285714285714285714285714285714285714286,0.0285714285714285714285714285714285714286,
                   -0.0285714285714285714285714285714285714286,0.253571428571428571428571428571428571429,0.1464285714285714285714285714285714285714,-0.2535714285714285714285714285714285714286,-0.1464285714285714285714285714285714285714,
                   -0.0035714285714285714285714285714285714286,0.1035714285714285714285714285714285714286,0.0035714285714285714285714285714285714286,-0.1035714285714285714285714285714285714286,-0.375,-0.25,0,0.125,-0.25,-0.125,0.125,0.25,0,0,0,0,0,0,0,0,0.25,0.125,0.125,
                   0,0.125,0,0,-0.125,0,0,0,0,0,0,0,0,-0.207142857142857142857142857142857142857,-0.042857142857142857142857142857142857143,-0.042857142857142857142857142857142857143,0.292857142857142857142857142857142857143,-0.242857142857142857142857142857142857143,
                   -0.0071428571428571428571428571428571428571,-0.0071428571428571428571428571428571428571,0.257142857142857142857142857142857142857,0.078571428571428571428571428571428571429,-0.0785714285714285714285714285714285714286,
                   -0.0785714285714285714285714285714285714286,0.0785714285714285714285714285714285714286,-0.0285714285714285714285714285714285714286,0.0285714285714285714285714285714285714286,0.0285714285714285714285714285714285714286,
                   -0.0285714285714285714285714285714285714286,0.253571428571428571428571428571428571429,-0.2535714285714285714285714285714285714286,0.1464285714285714285714285714285714285714,-0.1464285714285714285714285714285714285714,
                   -0.0035714285714285714285714285714285714286,0.0035714285714285714285714285714285714286,0.1035714285714285714285714285714285714286,-0.1035714285714285714285714285714285714286,0.253571428571428571428571428571428571429,
                   -0.2535714285714285714285714285714285714286,0.1464285714285714285714285714285714285714,-0.1464285714285714285714285714285714285714,-0.0035714285714285714285714285714285714286,0.0035714285714285714285714285714285714286,
                   0.1035714285714285714285714285714285714286,-0.1035714285714285714285714285714285714286,-0.375,0,-0.25,0.125,-0.25,0.125,-0.125,0.25,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.25,0.125,0.125,0,0.125,0,0,-0.125,0.125,0.125,0.125,0.125,-0.125,-0.125,-0.125,-0.125,
                   -0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,
                   0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.125,0.125,-0.125,-0.125,0.125,0.125,-0.125,-0.125,-0.125,-0.125,0.125,0.125,-0.125,-0.125,
                   0.125,0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,-0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.125,0.125,0.125,0.125,-0.125,-0.125,-0.125,-0.125,0,0,0,0,0,0,0,0,
                   -0.125,-0.125,0.125,0.125,-0.125,-0.125,0.125,0.125,0,0,0,0,0,0,0,0,-0.0357142857142857142857142857142857142857,0.0357142857142857142857142857142857142857,0.0357142857142857142857142857142857142857,-0.0357142857142857142857142857142857142857,
                   0.0357142857142857142857142857142857142857,-0.0357142857142857142857142857142857142857,-0.0357142857142857142857142857142857142857,0.0357142857142857142857142857142857142857,-0.1071428571428571428571428571428571428571,
                   0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,
                   -0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,
                   -0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,
                   -0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,
                   -0.1071428571428571428571428571428571428571,-0.1071428571428571428571428571428571428571,0.1071428571428571428571428571428571428571,0.125,0.125,0.125,0.125,-0.125,-0.125,-0.125,-0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.125,-0.125,0.125,0.125,-0.125,
                   -0.125,0.125,0.125,0.125,0.125,-0.125,-0.125,0.125,0.125,-0.125,-0.125,0,0,0,0,0,0,0,0,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,
                   -0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0.125,-0.125,0.125,
                   -0.125,0.125,-0.125,0.125,-0.125,0,0,0,0,0,0,0,0,-0.125,0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,0,0,0,0,0,0,0,0,0.125,0.125,-0.125,-0.125,0.125,0.125,-0.125,-0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.125,0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,
                   0.125,-0.125,0.125,-0.125,0.125,-0.125,0.125,-0.125,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,-0.0416666666666666666666666666666666666667,
                   -0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667,0.0416666666666666666666666666666666666667};
                return data;
            }

            static const double* N3_polyDegree4Inverse()
            {
                static const double data[35*88] = {-0.0338615136140676926132554947490876074211,0.0403273513273260569776631078062887136095,0.0016966618716147002976984072990116593438,0.0074371198364254781859259232987292118134,0.0403273513273260569776631078062887136095,
                   0.1057715897846208609426331654662109244643,0.0438555020117937965365355230970178248251,-0.0273124886545405536294024009974653880865,0.0016966618716147002976984072990116593438,0.0438555020117937965365355230970178248251,
                   0.0066132440888342623041205418994467446149,-0.0310039670416619780161672192556991688756,0.0074371198364254781859259232987292118134,-0.0273124886545405536294024009974653880865,-0.0310039670416619780161672192556991688756,
                   0.0363065510496504711811626842962074970474,0.0403273513273260569776631078062887136095,0.1057715897846208609426331654662109244643,0.0438555020117937965365355230970178248251,-0.0273124886545405536294024009974653880865,
                   0.1057715897846208609426331654662109244643,0.1879143869211961717397058998968550628236,0.1193649467541265229516024866990105630862,-0.0208505858609861610382977570893201623364,0.0438555020117937965365355230970178248251,
                   0.1193649467541265229516024866990105630862,0.0833502431192911985673848550731413356479,-0.0242447425181229104605861560084102298884,-0.0273124886545405536294024009974653880865,-0.0208505858609861610382977570893201623364,
                   -0.0242447425181229104605861560084102298884,-0.0366333707323741939685844832483787479514,0.0016966618716147002976984072990116593438,0.0438555020117937965365355230970178248251,0.0066132440888342623041205418994467446149,
                   -0.0310039670416619780161672192556991688756,0.0438555020117937965365355230970178248251,0.1193649467541265229516024866990105630862,0.0833502431192911985673848550731413356479,-0.0242447425181229104605861560084102298884,
                   0.0066132440888342623041205418994467446149,0.0833502431192911985673848550731413356479,0.0789400434585506637033320747866892409739,-0.0057557669998107343603546844535526966162,-0.0310039670416619780161672192556991688756,
                   -0.0242447425181229104605861560084102298884,-0.0057557669998107343603546844535526966162,-0.0137577310741615000156377913505925110778,0.0074371198364254781859259232987292118134,-0.0273124886545405536294024009974653880865,
                   -0.0310039670416619780161672192556991688756,0.0363065510496504711811626842962074970474,-0.0273124886545405536294024009974653880865,-0.0208505858609861610382977570893201623364,-0.0242447425181229104605861560084102298884,
                   -0.0366333707323741939685844832483787479514,-0.0310039670416619780161672192556991688756,-0.0242447425181229104605861560084102298884,-0.0057557669998107343603546844535526966162,-0.0137577310741615000156377913505925110778,
                   0.0363065510496504711811626842962074970474,-0.0366333707323741939685844832483787479514,-0.0137577310741615000156377913505925110778,0.0276305009558398445119896719642774808389,-0.0342145895112013338229950336309242644031,
                   -0.01124695748229624683334725150501650673666,-0.01124695748229624683334725150501650673666,0.0197424584736032469858560207592486250835,0.0181689339314600441533844855698652989672,0.00839885621647346202322066922653549407844,
                   0.00839885621647346202322066922653549407844,0.001999399637783612304007691858772365669,-0.0342145895112013338229950336309242644031,-0.01124695748229624683334725150501650673666,-0.01124695748229624683334725150501650673666,
                   0.0197424584736032469858560207592486250835,0.0181689339314600441533844855698652989672,0.00839885621647346202322066922653549407844,0.00839885621647346202322066922653549407844,0.001999399637783612304007691858772365669,
                   -0.0342145895112013338229950336309242644031,-0.01124695748229624683334725150501650673666,-0.01124695748229624683334725150501650673666,0.0197424584736032469858560207592486250835,0.0181689339314600441533844855698652989672,
                   0.00839885621647346202322066922653549407844,0.00839885621647346202322066922653549407844,0.001999399637783612304007691858772365669,0.0383572111155839135168876861582331681746,-0.09207227663628086563028537206332315312,
                   -0.0557539771858637481953764296465750065346,0.054501060578788512897631148674769729492,-0.09207227663628086563028537206332315312,-0.18866928310726634531747613795527184982,-0.134184785946445884697566727422979041168,
                   -0.0253064163203472493527683533014184283874,-0.0557539771858637481953764296465750065346,-0.134184785946445884697566727422979041168,-0.0824087079790313705009720063225378663105,-0.0009899567286289838460313272814910059704,
                   0.054501060578788512897631148674769729492,-0.0253064163203472493527683533014184283874,-0.0009899567286289838460313272814910059704,0.0230096436304535185191753176970466499498,-0.0414407161207714967460579597863374993493,
                   -0.0204591069359899436725651458620799380333,-0.0090445713645625625634345576786738797781,-0.0291347796416215884492287129624823053823,-0.0204591069359899436725651458620799380333,0.0079485710088191769405881077767327505757,
                   0.0274129761579838233228361300260318174778,0.0121198559978907477853885550581884590898,-0.0090445713645625625634345576786738797781,0.0274129761579838233228361300260318174778,0.0503427183692620993783252473592284410476,
                   0.0300538204771780052560145803920213645631,-0.0291347796416215884492287129624823053823,0.0121198559978907477853885550581884590898,0.0300538204771780052560145803920213645631,-0.0089003032743350890425966989911440381153,
                   0.0259127368577074909803994806757822628379,0.0905413994790651577382032661465002794437,0.0578297943380992506518672661638774887145,-0.023286370147407947415750189991104192508,0.0905413994790651577382032661465002794437,
                   0.1318764272003581537862317146660880146486,0.0927850723652229277320161906490736677542,0.0183264611129607497801871752751733328177,0.0578297943380992506518672661638774887145,0.0927850723652229277320161906490736677542,
                   0.0590195197084804011184836221810731565973,-0.0022843197713080716423278061616737734841,-0.023286370147407947415750189991104192508,0.0183264611129607497801871752751733328177,-0.0022843197713080716423278061616737734841,
                   -0.047812751217875166795120349324461062926,-0.0339603292844619573377407185409285064899,0.034338302173229702708353565131155015972,0.0060651383862950536971971703040750934913,0.0010292664254308963850692732661585405502,
                   0.034338302173229702708353565131155015972,0.0443103039315239474661228111357899369466,-0.0090855199607271110830239898765990179037,-0.0099166647113076848410355097189508568816,0.0060651383862950536971971703040750934913,
                   -0.0090855199607271110830239898765990179037,-0.0596083676975469067531598989254907693392,-0.0334474823104295578733870767568049919027,0.0010292664254308963850692732661585405502,-0.0099166647113076848410355097189508568816,
                   -0.0334474823104295578733870767568049919027,0.0386161538633190400696102041371254816144,0.0284830963879282247653708490470782189133,0.0266650586879323432838050626247866193814,0.0266650586879323432838050626247866193814,
                   0.0116515543997704329532548782042612842005,-0.0458975303335100797730704743911654879823,-0.0015595734558648327352818558737317670607,-0.0015595734558648327352818558737317670607,0.0388852424150097342907316669710496135605,
                   0.0239937781705030425055000682705558885674,0.0266504939270505353998647315073393089811,-0.0202895536046239972833663027560722985839,-0.0310392749003775543350324510196323184607,-0.0077554586517399640272116496210927948382,
                   -0.0071549689770293860577807514711507050104,0.0095493029162062234687084830574364372293,0.0060456811200111003293178720326164821155,0.0239937781705030425055000682705558885674,0.0266504939270505353998647315073393089811,
                   -0.0202895536046239972833663027560722985839,-0.0310392749003775543350324510196323184607,-0.0077554586517399640272116496210927948382,-0.0071549689770293860577807514711507050104,0.00954930291620622346870848305743643722926,
                   0.0060456811200111003293178720326164821155,0.0383572111155839135168876861582331681746,-0.09207227663628086563028537206332315312,-0.0557539771858637481953764296465750065346,0.054501060578788512897631148674769729492,
                   -0.0414407161207714967460579597863374993493,-0.0204591069359899436725651458620799380333,-0.0090445713645625625634345576786738797781,-0.0291347796416215884492287129624823053823,0.0259127368577074909803994806757822628379,
                   0.0905413994790651577382032661465002794437,0.0578297943380992506518672661638774887145,-0.023286370147407947415750189991104192508,-0.0339603292844619573377407185409285064899,0.034338302173229702708353565131155015972,
                   0.0060651383862950536971971703040750934913,0.0010292664254308963850692732661585405502,-0.09207227663628086563028537206332315312,-0.18866928310726634531747613795527184982,-0.134184785946445884697566727422979041168,
                   -0.0253064163203472493527683533014184283874,-0.0204591069359899436725651458620799380333,0.0079485710088191769405881077767327505757,0.0274129761579838233228361300260318174778,0.0121198559978907477853885550581884590898,
                   0.0905413994790651577382032661465002794437,0.1318764272003581537862317146660880146486,0.0927850723652229277320161906490736677542,0.0183264611129607497801871752751733328177,0.034338302173229702708353565131155015972,
                   0.0443103039315239474661228111357899369466,-0.0090855199607271110830239898765990179037,-0.0099166647113076848410355097189508568816,-0.0557539771858637481953764296465750065346,-0.134184785946445884697566727422979041168,
                   -0.0824087079790313705009720063225378663105,-0.0009899567286289838460313272814910059704,-0.0090445713645625625634345576786738797781,0.0274129761579838233228361300260318174778,0.0503427183692620993783252473592284410476,
                   0.0300538204771780052560145803920213645631,0.0578297943380992506518672661638774887145,0.0927850723652229277320161906490736677542,0.0590195197084804011184836221810731565973,-0.0022843197713080716423278061616737734841,
                   0.0060651383862950536971971703040750934913,-0.0090855199607271110830239898765990179037,-0.0596083676975469067531598989254907693392,-0.0334474823104295578733870767568049919027,0.054501060578788512897631148674769729492,
                   -0.0253064163203472493527683533014184283874,-0.0009899567286289838460313272814910059704,0.0230096436304535185191753176970466499498,-0.0291347796416215884492287129624823053823,0.0121198559978907477853885550581884590898,
                   0.0300538204771780052560145803920213645631,-0.0089003032743350890425966989911440381153,-0.023286370147407947415750189991104192508,0.0183264611129607497801871752751733328177,-0.0022843197713080716423278061616737734841,
                   -0.047812751217875166795120349324461062926,0.0010292664254308963850692732661585405502,-0.0099166647113076848410355097189508568816,-0.0334474823104295578733870767568049919027,0.0386161538633190400696102041371254816144,
                   0.0239937781705030425055000682705558885674,0.0266504939270505353998647315073393089811,-0.0202895536046239972833663027560722985839,-0.0310392749003775543350324510196323184607,-0.0077554586517399640272116496210927948382,
                   -0.0071549689770293860577807514711507050104,0.0095493029162062234687084830574364372293,0.0060456811200111003293178720326164821155,0.0284830963879282247653708490470782189133,0.0266650586879323432838050626247866193814,
                   0.0266650586879323432838050626247866193814,0.0116515543997704329532548782042612842005,-0.0458975303335100797730704743911654879823,-0.00155957345586483273528185587373176706075,-0.00155957345586483273528185587373176706075,
                   0.0388852424150097342907316669710496135605,0.0239937781705030425055000682705558885674,-0.0202895536046239972833663027560722985839,0.0266504939270505353998647315073393089811,-0.0310392749003775543350324510196323184607,
                   -0.0077554586517399640272116496210927948382,0.0095493029162062234687084830574364372293,-0.0071549689770293860577807514711507050104,0.0060456811200111003293178720326164821155,0.0383572111155839135168876861582331681746,
                   -0.0414407161207714967460579597863374993493,0.0259127368577074909803994806757822628379,-0.0339603292844619573377407185409285064899,-0.09207227663628086563028537206332315312,-0.0204591069359899436725651458620799380333,
                   0.0905413994790651577382032661465002794437,0.034338302173229702708353565131155015972,-0.0557539771858637481953764296465750065346,-0.0090445713645625625634345576786738797781,0.0578297943380992506518672661638774887145,
                   0.0060651383862950536971971703040750934913,0.054501060578788512897631148674769729492,-0.0291347796416215884492287129624823053823,-0.023286370147407947415750189991104192508,0.0010292664254308963850692732661585405502,
                   -0.09207227663628086563028537206332315312,-0.0204591069359899436725651458620799380333,0.0905413994790651577382032661465002794437,0.034338302173229702708353565131155015972,-0.18866928310726634531747613795527184982,
                   0.0079485710088191769405881077767327505757,0.1318764272003581537862317146660880146486,0.0443103039315239474661228111357899369466,-0.134184785946445884697566727422979041168,0.0274129761579838233228361300260318174778,
                   0.0927850723652229277320161906490736677542,-0.0090855199607271110830239898765990179037,-0.0253064163203472493527683533014184283874,0.0121198559978907477853885550581884590898,0.0183264611129607497801871752751733328177,
                   -0.0099166647113076848410355097189508568816,-0.0557539771858637481953764296465750065346,-0.0090445713645625625634345576786738797781,0.0578297943380992506518672661638774887145,0.0060651383862950536971971703040750934913,
                   -0.134184785946445884697566727422979041168,0.0274129761579838233228361300260318174778,0.0927850723652229277320161906490736677542,-0.0090855199607271110830239898765990179037,-0.0824087079790313705009720063225378663105,
                   0.0503427183692620993783252473592284410476,0.0590195197084804011184836221810731565973,-0.0596083676975469067531598989254907693392,-0.0009899567286289838460313272814910059704,0.0300538204771780052560145803920213645631,
                   -0.0022843197713080716423278061616737734841,-0.0334474823104295578733870767568049919027,0.054501060578788512897631148674769729492,-0.0291347796416215884492287129624823053823,-0.023286370147407947415750189991104192508,
                   0.0010292664254308963850692732661585405502,-0.0253064163203472493527683533014184283874,0.0121198559978907477853885550581884590898,0.0183264611129607497801871752751733328177,-0.0099166647113076848410355097189508568816,
                   -0.0009899567286289838460313272814910059704,0.0300538204771780052560145803920213645631,-0.0022843197713080716423278061616737734841,-0.0334474823104295578733870767568049919027,0.0230096436304535185191753176970466499498,
                   -0.0089003032743350890425966989911440381153,-0.047812751217875166795120349324461062926,0.0386161538633190400696102041371254816144,0.0239937781705030425055000682705558885674,-0.0202895536046239972833663027560722985839,
                   0.0266504939270505353998647315073393089811,-0.0310392749003775543350324510196323184607,-0.0077554586517399640272116496210927948382,0.0095493029162062234687084830574364372293,-0.0071549689770293860577807514711507050104,
                   0.0060456811200111003293178720326164821155,0.0239937781705030425055000682705558885674,-0.0202895536046239972833663027560722985839,0.0266504939270505353998647315073393089811,-0.0310392749003775543350324510196323184607,
                   -0.0077554586517399640272116496210927948382,0.0095493029162062234687084830574364372293,-0.0071549689770293860577807514711507050104,0.0060456811200111003293178720326164821155,0.0284830963879282247653708490470782189133,
                   0.0266650586879323432838050626247866193814,0.0266650586879323432838050626247866193814,0.0116515543997704329532548782042612842005,-0.0458975303335100797730704743911654879823,-0.00155957345586483273528185587373176706075,
                   -0.00155957345586483273528185587373176706075,0.0388852424150097342907316669710496135605,0.0164763571691830613593614966193889481979,0.0123081271668582688143944495132056837908,0.0162159976848591818130198165997592885692,
                   -0.0181093527591378170911494077576127293986,0.0123081271668582688143944495132056837908,0.0294125515425689054916156527130737095178,0.0426135266757478706540777609864082336249,0.0094783133625525595136216167978125869652,
                   0.0162159976848591818130198165997592885692,0.0426135266757478706540777609864082336249,0.0557322233279141775886917679939199729219,0.017015930715996510486930528188163660098,-0.0181093527591378170911494077576127293986,
                   0.0094783133625525595136216167978125869652,0.017015930715996510486930528188163660098,-0.0301760753456865436429261349194245335773,-0.0103531076012169189007120346067072345939,-0.0510334416341028148016260359035858100739,
                   -0.0325526283683208935851752287381546590267,0.020653545641137622989079447825825742556,-0.0510334416341028148016260359035858100739,-0.0825075769851558131259214071475776523817,-0.062923532521917358732975630947264316183,
                   -0.0128405125208976607241217982651404546848,-0.0325526283683208935851752287381546590267,-0.062923532521917358732975630947264316183,-0.0477350576789320688969918800634066761869,-0.0036698258373942205203292189478105694604,
                   0.020653545641137622989079447825825742556,-0.0128405125208976607241217982651404546848,-0.0036698258373942205203292189478105694604,0.0353595659720997598155797910178523905918,0.0193706701976142579997038470505958044386,
                   -0.0125384342423057325973454177141678646699,-0.009722177938549127836590813080382283243,0.0252571874812252462092327884610940886677,-0.0125384342423057325973454177141678646699,-0.042994504557354850934231748628576446176,
                   -0.0429515993346227256552100230611745023824,-0.0110953879232871701747872654015546705561,-0.009722177938549127836590813080382283243,-0.0429515993346227256552100230611745023824,-0.0473042637321437649331503228102282547913,
                   -0.0175892582018090464266908886158703549519,0.0252571874812252462092327884610940886677,-0.0110953879232871701747872654015546705561,-0.0175892582018090464266908886158703549519,0.0148430718534438293554705906310860437828,
                   -0.0143628223336383508718417975700269432159,0.0389154306295262274408706907522957866905,0.0269624244480428460184927760760739578074,-0.0309105575784149255248843475166488739769,0.0389154306295262274408706907522957866905,
                   0.0902068442998401590264043407730748700223,0.0759171958980917917931796229798364921134,0.0192343510024357080135155795558900316371,0.0269624244480428460184927760760739578074,0.0759171958980917917931796229798364921134,
                   0.0615452690153307663321068039207753293945,0.0109110916563953645658212091834656711084,-0.0309105575784149255248843475166488739769,0.0192343510024357080135155795558900316371,0.0109110916563953645658212091834656711084,
                   -0.0249393054814193482791927202480809313202,0.0594951805519509379586597184503800976353,0.009256083597912940600145114404684329955,0.009256083597912940600145114404684329955,-0.0324387095586567023279897428055683997506,
                   -0.0316640799397024162842934264396261618997,-0.023944902163313784482001654489072515609,-0.023944902163313784482001654489072515609,-0.0076814205894567982493301357030758313437,0.0006257861410851786745769070805391972806,
                   0.0086279851132198563228479027584844590375,-0.0043300107069642238967106725950227872641,-0.0042392041398928373876801832461914495579,-0.0042500129252735948673596667303673877605,0.0011572493380003232872404428716285069078,
                   0.0024561686608073354258628332940237453694,-0.0000479614809820375587775634330942840129,0.0006257861410851786745769070805391972806,0.0086279851132198563228479027584844590375,-0.0043300107069642238967106725950227872641,
                   -0.0042392041398928373876801832461914495579,-0.0042500129252735948673596667303673877605,0.0011572493380003232872404428716285069078,0.0024561686608073354258628332940237453694,-0.0000479614809820375587775634330942840129,
                   0.0603419011636773765654315820834116566354,0.090837762727834262287150610970968707634,0.0752061760336923723842951502703505151112,-0.0098123525896343690925018886260378576419,-0.0019399751390229065614616930084399329242,
                   0.012916528979010837094053868002458173688,0.0195182363678833716833974064187078057163,0.0023588179136706465736575298352456720213,-0.0662031953869426235378587308345913284547,-0.083802504411234553594116257953516723191,
                   -0.0717839586355501261581433406735638235737,-0.0379007226168513665463956751972642751724,0.0220415440094200765381218722402119517221,-0.0370768692964380335590226670591698557551,-0.0287047762731822196055342949661465999411,
                   0.047157823079187518398586988519281719164,-0.0019399751390229065614616930084399329242,0.012916528979010837094053868002458173688,0.0195182363678833716833974064187078057163,0.0023588179136706465736575298352456720213,
                   0.0137802412929872320445785151951900970814,0.0220190737660949683606102629266228615052,0.0298799804484617625908133833269761143753,0.0296097967831255894187321801937182101223,-0.0038273108625257580217839269944068295555,
                   -0.0103960234520655480325777646943381414974,-0.0094657908939260131488568344617800019625,-0.0010366131881071533706211362967324109508,-0.0212648891974946250147857594018918429711,-0.0430778557104414350232912584825543198863,
                   -0.0495150061372886531569385943671583822943,-0.0328231759210742540992720708531723846255,-0.0662031953869426235378587308345913284547,-0.083802504411234553594116257953516723191,-0.0717839586355501261581433406735638235737,
                   -0.0379007226168513665463956751972642751724,-0.0038273108625257580217839269944068295555,-0.0103960234520655480325777646943381414974,-0.0094657908939260131488568344617800019625,-0.0010366131881071533706211362967324109508,
                   0.0563743804320642839989893387936235620296,0.0612729731380361277045452664620534214896,0.0515516013393902637069002855965838895497,0.0349634295930887173225100923997466117793,0.02690805972346015511410618324492310435,
                   0.0514638311426651515233536484336128693892,0.0392807284049554076316845286220144001517,0.0058650806242549740720102162151909877765,0.0220415440094200765381218722402119517221,-0.0370768692964380335590226670591698557551,
                   -0.0287047762731822196055342949661465999411,0.047157823079187518398586988519281719164,-0.0212648891974946250147857594018918429711,-0.0430778557104414350232912584825543198863,-0.0495150061372886531569385943671583822943,
                   -0.0328231759210742540992720708531723846255,0.02690805972346015511410618324492310435,0.0514638311426651515233536484336128693892,0.0392807284049554076316845286220144001517,0.0058650806242549740720102162151909877765,
                   -0.0419249891825175296416753265638355600796,0.0458159758650418048308947231473710038761,0.044703376512672066826773439661942684771,-0.022003293568740667704672088412525580686,-0.0008295647783472188944659058447027203823,
                   -0.01143009465765248824977941776698708894281,-0.00108081552306556688617102623343722721013,0.0147095877739612214564842578407458755276,0.043349483955996584450688737939844841772,0.0368614062403651449452427561912431343831,
                   -0.0414391036545837986700518058617048941796,-0.0401408993566738781519475962650019209679,-0.0008295647783472188944659058447027203823,-0.01143009465765248824977941776698708894281,-0.00108081552306556688617102623343722721013,
                   0.0147095877739612214564842578407458755276,0.043349483955996584450688737939844841772,0.0368614062403651449452427561912431343831,-0.0414391036545837986700518058617048941796,-0.0401408993566738781519475962650019209679,
                   -0.0027083524138994210408261636647905333263,-0.00538465312530664311647532136198606613679,-0.00538465312530664311647532136198606613679,0.0134776586645127072737768063887626655999,-0.0021696417868755764720890426756824968235,
                   -0.00073349033460896869787067019919536846237,-0.00073349033460896869787067019919536846237,0.0036366224560935138678303830740732337482,0.0603419011636773765654315820834116566354,-0.0019399751390229065614616930084399329242,
                   -0.0662031953869426235378587308345913284547,0.0220415440094200765381218722402119517221,0.090837762727834262287150610970968707634,0.012916528979010837094053868002458173688,-0.083802504411234553594116257953516723191,
                   -0.0370768692964380335590226670591698557551,0.0752061760336923723842951502703505151112,0.0195182363678833716833974064187078057163,-0.0717839586355501261581433406735638235737,-0.0287047762731822196055342949661465999411,
                   -0.0098123525896343690925018886260378576419,0.0023588179136706465736575298352456720213,-0.0379007226168513665463956751972642751724,0.047157823079187518398586988519281719164,-0.0019399751390229065614616930084399329242,
                   0.0137802412929872320445785151951900970814,-0.0038273108625257580217839269944068295555,-0.0212648891974946250147857594018918429711,0.012916528979010837094053868002458173688,0.0220190737660949683606102629266228615052,
                   -0.0103960234520655480325777646943381414974,-0.0430778557104414350232912584825543198863,0.0195182363678833716833974064187078057163,0.0298799804484617625908133833269761143753,-0.0094657908939260131488568344617800019625,
                   -0.0495150061372886531569385943671583822943,0.0023588179136706465736575298352456720213,0.0296097967831255894187321801937182101223,-0.0010366131881071533706211362967324109508,-0.0328231759210742540992720708531723846255,
                   -0.0662031953869426235378587308345913284547,-0.0038273108625257580217839269944068295555,0.0563743804320642839989893387936235620296,0.02690805972346015511410618324492310435,-0.083802504411234553594116257953516723191,
                   -0.0103960234520655480325777646943381414974,0.0612729731380361277045452664620534214896,0.0514638311426651515233536484336128693892,-0.0717839586355501261581433406735638235737,-0.0094657908939260131488568344617800019625,
                   0.0515516013393902637069002855965838895497,0.0392807284049554076316845286220144001517,-0.0379007226168513665463956751972642751724,-0.0010366131881071533706211362967324109508,0.0349634295930887173225100923997466117793,
                   0.0058650806242549740720102162151909877765,0.0220415440094200765381218722402119517221,-0.0212648891974946250147857594018918429711,0.02690805972346015511410618324492310435,-0.0419249891825175296416753265638355600796,
                   -0.0370768692964380335590226670591698557551,-0.0430778557104414350232912584825543198863,0.0514638311426651515233536484336128693892,0.0458159758650418048308947231473710038761,-0.0287047762731822196055342949661465999411,
                   -0.0495150061372886531569385943671583822943,0.0392807284049554076316845286220144001517,0.044703376512672066826773439661942684771,0.047157823079187518398586988519281719164,-0.0328231759210742540992720708531723846255,
                   0.0058650806242549740720102162151909877765,-0.022003293568740667704672088412525580686,-0.0008295647783472188944659058447027203823,-0.00108081552306556688617102623343722721013,-0.01143009465765248824977941776698708894281,
                   0.0147095877739612214564842578407458755276,0.043349483955996584450688737939844841772,-0.0414391036545837986700518058617048941796,0.0368614062403651449452427561912431343831,-0.0401408993566738781519475962650019209679,
                   -0.0027083524138994210408261636647905333263,-0.00538465312530664311647532136198606613679,-0.00538465312530664311647532136198606613679,0.0134776586645127072737768063887626655999,-0.0021696417868755764720890426756824968235,
                   -0.0007334903346089686978706701991953684624,-0.0007334903346089686978706701991953684624,0.0036366224560935138678303830740732337482,-0.0008295647783472188944659058447027203823,-0.01143009465765248824977941776698708894281,
                   -0.00108081552306556688617102623343722721013,0.0147095877739612214564842578407458755276,0.043349483955996584450688737939844841772,0.0368614062403651449452427561912431343831,-0.0414391036545837986700518058617048941796,
                   -0.0401408993566738781519475962650019209679,0.0164763571691830613593614966193889481979,0.0123081271668582688143944495132056837908,0.0162159976848591818130198165997592885692,-0.0181093527591378170911494077576127293986,
                   -0.0103531076012169189007120346067072345939,-0.0510334416341028148016260359035858100739,-0.0325526283683208935851752287381546590267,0.020653545641137622989079447825825742556,0.0193706701976142579997038470505958044386,
                   -0.0125384342423057325973454177141678646699,-0.009722177938549127836590813080382283243,0.0252571874812252462092327884610940886677,-0.0143628223336383508718417975700269432159,0.0389154306295262274408706907522957866905,
                   0.0269624244480428460184927760760739578074,-0.0309105575784149255248843475166488739769,0.0123081271668582688143944495132056837908,0.0294125515425689054916156527130737095178,0.0426135266757478706540777609864082336249,
                   0.0094783133625525595136216167978125869652,-0.0510334416341028148016260359035858100739,-0.0825075769851558131259214071475776523817,-0.062923532521917358732975630947264316183,-0.0128405125208976607241217982651404546848,
                   -0.0125384342423057325973454177141678646699,-0.042994504557354850934231748628576446176,-0.0429515993346227256552100230611745023824,-0.0110953879232871701747872654015546705561,0.0389154306295262274408706907522957866905,
                   0.0902068442998401590264043407730748700223,0.0759171958980917917931796229798364921134,0.0192343510024357080135155795558900316371,0.0162159976848591818130198165997592885692,0.0426135266757478706540777609864082336249,
                   0.0557322233279141775886917679939199729219,0.017015930715996510486930528188163660098,-0.0325526283683208935851752287381546590267,-0.062923532521917358732975630947264316183,-0.0477350576789320688969918800634066761869,
                   -0.0036698258373942205203292189478105694604,-0.009722177938549127836590813080382283243,-0.0429515993346227256552100230611745023824,-0.0473042637321437649331503228102282547913,-0.0175892582018090464266908886158703549519,
                   0.0269624244480428460184927760760739578074,0.0759171958980917917931796229798364921134,0.0615452690153307663321068039207753293945,0.0109110916563953645658212091834656711084,-0.0181093527591378170911494077576127293986,
                   0.0094783133625525595136216167978125869652,0.017015930715996510486930528188163660098,-0.0301760753456865436429261349194245335773,0.020653545641137622989079447825825742556,-0.0128405125208976607241217982651404546848,
                   -0.0036698258373942205203292189478105694604,0.0353595659720997598155797910178523905918,0.0252571874812252462092327884610940886677,-0.0110953879232871701747872654015546705561,-0.0175892582018090464266908886158703549519,
                   0.0148430718534438293554705906310860437828,-0.0309105575784149255248843475166488739769,0.0192343510024357080135155795558900316371,0.0109110916563953645658212091834656711084,-0.0249393054814193482791927202480809313202,
                   0.0006257861410851786745769070805391972806,0.0086279851132198563228479027584844590375,-0.0043300107069642238967106725950227872641,-0.0042392041398928373876801832461914495579,-0.0042500129252735948673596667303673877605,
                   0.0011572493380003232872404428716285069078,0.0024561686608073354258628332940237453694,-0.0000479614809820375587775634330942840129,0.0594951805519509379586597184503800976353,0.009256083597912940600145114404684329955,
                   0.009256083597912940600145114404684329955,-0.0324387095586567023279897428055683997506,-0.0316640799397024162842934264396261618997,-0.023944902163313784482001654489072515609,-0.023944902163313784482001654489072515609,
                   -0.0076814205894567982493301357030758313437,0.0006257861410851786745769070805391972806,-0.0043300107069642238967106725950227872641,0.0086279851132198563228479027584844590375,-0.0042392041398928373876801832461914495579,
                   -0.0042500129252735948673596667303673877605,0.0024561686608073354258628332940237453694,0.0011572493380003232872404428716285069078,-0.0000479614809820375587775634330942840129,0.0603419011636773765654315820834116566354,
                   -0.0019399751390229065614616930084399329242,-0.0662031953869426235378587308345913284547,0.0220415440094200765381218722402119517221,-0.0019399751390229065614616930084399329242,0.0137802412929872320445785151951900970814,
                   -0.0038273108625257580217839269944068295555,-0.0212648891974946250147857594018918429711,-0.0662031953869426235378587308345913284547,-0.0038273108625257580217839269944068295555,0.0563743804320642839989893387936235620296,
                   0.02690805972346015511410618324492310435,0.0220415440094200765381218722402119517221,-0.0212648891974946250147857594018918429711,0.02690805972346015511410618324492310435,-0.0419249891825175296416753265638355600796,
                   0.090837762727834262287150610970968707634,0.012916528979010837094053868002458173688,-0.083802504411234553594116257953516723191,-0.0370768692964380335590226670591698557551,0.012916528979010837094053868002458173688,
                   0.0220190737660949683606102629266228615052,-0.0103960234520655480325777646943381414974,-0.0430778557104414350232912584825543198863,-0.083802504411234553594116257953516723191,-0.0103960234520655480325777646943381414974,
                   0.0612729731380361277045452664620534214896,0.0514638311426651515233536484336128693892,-0.0370768692964380335590226670591698557551,-0.0430778557104414350232912584825543198863,0.0514638311426651515233536484336128693892,
                   0.0458159758650418048308947231473710038761,0.0752061760336923723842951502703505151112,0.0195182363678833716833974064187078057163,-0.0717839586355501261581433406735638235737,-0.0287047762731822196055342949661465999411,
                   0.0195182363678833716833974064187078057163,0.0298799804484617625908133833269761143753,-0.0094657908939260131488568344617800019625,-0.0495150061372886531569385943671583822943,-0.0717839586355501261581433406735638235737,
                   -0.0094657908939260131488568344617800019625,0.0515516013393902637069002855965838895497,0.0392807284049554076316845286220144001517,-0.0287047762731822196055342949661465999411,-0.0495150061372886531569385943671583822943,
                   0.0392807284049554076316845286220144001517,0.044703376512672066826773439661942684771,-0.0098123525896343690925018886260378576419,0.0023588179136706465736575298352456720213,-0.0379007226168513665463956751972642751724,
                   0.047157823079187518398586988519281719164,0.0023588179136706465736575298352456720213,0.0296097967831255894187321801937182101223,-0.0010366131881071533706211362967324109508,-0.0328231759210742540992720708531723846255,
                   -0.0379007226168513665463956751972642751724,-0.0010366131881071533706211362967324109508,0.0349634295930887173225100923997466117793,0.0058650806242549740720102162151909877765,0.047157823079187518398586988519281719164,
                   -0.0328231759210742540992720708531723846255,0.0058650806242549740720102162151909877765,-0.022003293568740667704672088412525580686,-0.0027083524138994210408261636647905333263,-0.00538465312530664311647532136198606613679,
                   -0.00538465312530664311647532136198606613679,0.0134776586645127072737768063887626655999,-0.0021696417868755764720890426756824968235,-0.0007334903346089686978706701991953684624,-0.0007334903346089686978706701991953684624,
                   0.0036366224560935138678303830740732337482,-0.0008295647783472188944659058447027203823,-0.00108081552306556688617102623343722721013,-0.01143009465765248824977941776698708894281,0.0147095877739612214564842578407458755276,
                   0.043349483955996584450688737939844841772,-0.0414391036545837986700518058617048941796,0.0368614062403651449452427561912431343831,-0.0401408993566738781519475962650019209679,-0.0008295647783472188944659058447027203823,
                   -0.00108081552306556688617102623343722721013,-0.01143009465765248824977941776698708894281,0.0147095877739612214564842578407458755276,0.043349483955996584450688737939844841772,-0.0414391036545837986700518058617048941796,
                   0.0368614062403651449452427561912431343831,-0.0401408993566738781519475962650019209679,0.0164763571691830613593614966193889481979,-0.0103531076012169189007120346067072345939,0.0193706701976142579997038470505958044386,
                   -0.0143628223336383508718417975700269432159,0.0123081271668582688143944495132056837908,-0.0510334416341028148016260359035858100739,-0.0125384342423057325973454177141678646699,0.0389154306295262274408706907522957866905,
                   0.0162159976848591818130198165997592885692,-0.0325526283683208935851752287381546590267,-0.009722177938549127836590813080382283243,0.0269624244480428460184927760760739578074,-0.0181093527591378170911494077576127293986,
                   0.020653545641137622989079447825825742556,0.0252571874812252462092327884610940886677,-0.0309105575784149255248843475166488739769,0.0123081271668582688143944495132056837908,-0.0510334416341028148016260359035858100739,
                   -0.0125384342423057325973454177141678646699,0.0389154306295262274408706907522957866905,0.0294125515425689054916156527130737095178,-0.0825075769851558131259214071475776523817,-0.042994504557354850934231748628576446176,
                   0.0902068442998401590264043407730748700223,0.0426135266757478706540777609864082336249,-0.062923532521917358732975630947264316183,-0.0429515993346227256552100230611745023824,0.0759171958980917917931796229798364921134,
                   0.0094783133625525595136216167978125869652,-0.0128405125208976607241217982651404546848,-0.0110953879232871701747872654015546705561,0.0192343510024357080135155795558900316371,0.0162159976848591818130198165997592885692,
                   -0.0325526283683208935851752287381546590267,-0.009722177938549127836590813080382283243,0.0269624244480428460184927760760739578074,0.0426135266757478706540777609864082336249,-0.062923532521917358732975630947264316183,
                   -0.0429515993346227256552100230611745023824,0.0759171958980917917931796229798364921134,0.0557322233279141775886917679939199729219,-0.0477350576789320688969918800634066761869,-0.0473042637321437649331503228102282547913,
                   0.0615452690153307663321068039207753293945,0.017015930715996510486930528188163660098,-0.0036698258373942205203292189478105694604,-0.0175892582018090464266908886158703549519,0.0109110916563953645658212091834656711084,
                   -0.0181093527591378170911494077576127293986,0.020653545641137622989079447825825742556,0.0252571874812252462092327884610940886677,-0.0309105575784149255248843475166488739769,0.0094783133625525595136216167978125869652,
                   -0.0128405125208976607241217982651404546848,-0.0110953879232871701747872654015546705561,0.0192343510024357080135155795558900316371,0.017015930715996510486930528188163660098,-0.0036698258373942205203292189478105694604,
                   -0.0175892582018090464266908886158703549519,0.0109110916563953645658212091834656711084,-0.0301760753456865436429261349194245335773,0.0353595659720997598155797910178523905918,0.0148430718534438293554705906310860437828,
                   -0.0249393054814193482791927202480809313202,0.0006257861410851786745769070805391972806,-0.0043300107069642238967106725950227872641,0.0086279851132198563228479027584844590375,-0.0042392041398928373876801832461914495579,
                   -0.0042500129252735948673596667303673877605,0.0024561686608073354258628332940237453694,0.0011572493380003232872404428716285069078,-0.0000479614809820375587775634330942840129,0.0006257861410851786745769070805391972806,
                   -0.0043300107069642238967106725950227872641,0.0086279851132198563228479027584844590375,-0.0042392041398928373876801832461914495579,-0.0042500129252735948673596667303673877605,0.0024561686608073354258628332940237453694,
                   0.0011572493380003232872404428716285069078,-0.0000479614809820375587775634330942840129,0.0594951805519509379586597184503800976353,0.009256083597912940600145114404684329955,0.009256083597912940600145114404684329955,
                   -0.0324387095586567023279897428055683997506,-0.0316640799397024162842934264396261618997,-0.02394490216331378448200165448907251560904,-0.023944902163313784482001654489072515609,-0.0076814205894567982493301357030758313437,
                   -0.01679014316760713741040109806313863047125,0.01537951782088931954215874707969670096659,0.01009255892106122140182431982543822307942,-0.0107774849397590361445783132530120481928,0.01537951782088931954215874707969670096659,
                   0.0450043475857570845115962293533337201681,0.0358408064074479737130339539978094194962,0.0097624292132943828332979875860258148906,0.0100925589210612214018243198254382230794,0.0358408064074479737130339539978094194962,
                   0.0266772652291388629144716786422851188242,0.00447547031346628469296356033176733700347,-0.0107774849397590361445783132530120481928,0.0097624292132943828332979875860258148906,0.00447547031346628469296356033176733700347,
                   -0.00476482671191093487875552844288546591429,0.01641834259961039230013862721910101301083,0.0063212524639323607347602060631393151346,0.0010997334765905885828614718859241252612,0.0080449639466958744067177802117561153706,
                   0.0063212524639323607347602060631393151346,-0.00115189252406634593610344716033293126476,-0.0076656056042351223074114640379700620664,-0.00592870846746316981688848904547140529043,0.0010997334765905885828614718859241252612,
                   -0.0076656056042351223074114640379700620664,-0.01417931868440389867871948091560719286814,-0.01115022745480494196878722322268659516385,0.0080449639466958744067177802117561153706,-0.00592870846746316981688848904547140529043,
                   -0.01115022745480494196878722322268659516385,-0.00032841470621864348670306679558878226968,-0.0033975092662770589668052938857676796775,-0.019342085797265694068093539396472648468,-0.0141205668099239219161948052192574585946,
                   0.00497586938663745892661555312157721796276,-0.019342085797265694068093539396472648468,-0.0274939408092669873972298861730004020686,-0.02098022772909821102592186929536327126689,-0.00709212486587016351644484428786192804291,
                   -0.01412056680992392191619480521925745859455,-0.02098022772909821102592186929536327126689,-0.0144665146489294346546138524177261404652,-0.00187060587852839136454611011064673816949,0.00497586938663745892661555312157721796276,
                   -0.00709212486587016351644484428786192804291,-0.00187060587852839136454611011064673816949,0.01334924803955197682003640012892211560302,0.0037693098342738040770677647298052971379,-0.0023586844875559862088254137463633676333,
                   0.00292827441227211193150901350789511025391,-0.00224334839357429718875502008032128514056,-0.0023586844875559862088254137463633676333,0.00447481908090958215507043731333294649852,0.0136383602592186929536327126688572471705,
                   0.00325840412003895050003534574730751844269,0.00292827441227211193150901350789511025391,0.0136383602592186929536327126688572471705,0.02280190143752780375219498802438154784248,0.00854536301986704864036977300156599632986,
                   -0.00224334839357429718875502008032128514056,0.00325840412003895050003534574730751844269,0.00854536301986704864036977300156599632986,-0.00825600662142239845457780489044786741904,-0.0408030868305511180809843816300020865117,
                   -0.0214836619204089083607155896312522818547,-0.0214836619204089083607155896312522818547,-0.0021642370102666986404467976325024771977,0.01996975349721778474765104829666875317834,0.00065032858707557502738225629791894852136,
                   0.00065032858707557502738225629791894852136,-0.01866909632306663469288653570083085613562,-0.0016519322826831341904291932251873281816,-0.00428906308437089790350936200155863619846,0.00165193228268313419042919322518732818158,
                   0.00428906308437089790350936200155863619846,0.00223539386202697676440750000813037770996,-0.00040173693966078694867266876824093030692,-0.00223539386202697676440750000813037770996,0.00040173693966078694867266876824093030692,
                   -0.00165193228268313419042919322518732818158,-0.0042890630843708979035093620015586361985,0.0016519322826831341904291932251873281816,0.00428906308437089790350936200155863619846,0.00223539386202697676440750000813037770996,
                   -0.00040173693966078694867266876824093030692,-0.00223539386202697676440750000813037770996,0.00040173693966078694867266876824093030692,-0.0502107099283208784505109043770016775965,-0.0184548138043057093007728327153930968256,
                   -0.0148250669688626713260892884115956284711,-0.0276917225865487265517767271618118041787,0.0195328705180009129986253670865536047784,0.0288259751331789651624621082733345241071,0.028743696652166306934614007007511739297,
                   0.0231626173534439509733089113903510731328,0.0117171294819990870013746329134463952216,0.0024240248668210348375378917266654758929,0.002506303347833693065385992992488260703,0.0080873826465560490266910886096489268672,
                   0.0189607099283208784505109043770016775965,-0.0127951861956942906992271672846069031744,-0.0164249330311373286739107115884043715289,-0.0035582774134512734482232728381881958213,0.0466316532423545418897914824325329771498,
                   0.0538179291132051540775042376384453553891,0.0445078025309266730648460097903440895663,0.0303310203309621368265003431920266480359,0.0028558132657819212164508071654311510472,0.0039590444632384543929457762003133361986,
                   -0.0004365251570147101640162491161423600039,-0.0064543133164965597962074206826701147756,-0.0341058132657819212164508071654311510472,-0.0352090444632384543929457762003133361986,-0.0308134748429852898359837508838576399961,
                   -0.0247956866835034402037925793173298852244,-0.0153816532423545418897914824325329771498,-0.0225679291132051540775042376384453553891,-0.0132578025309266730648460097903440895663,0.00091897966903786317349965680797335196411,
                   0.0215115172836109882095289414104982842291,0.0170680463190185624225581523126131941139,0.0077579197367400814098999244645119282911,0.0052108843722185831462378021699919551151,-0.0128087436962433952392453953662144185731,
                   -0.0155820947772678747209782744325980562064,-0.0199776643975210392779402997490537524089,-0.0221188702785218762519036232143156843959,-0.0184412563037566047607546046337855814269,-0.0156679052227321252790217255674019437936,
                   -0.0112723356024789607220597002509462475911,-0.0091311297214781237480963767856843156041,0.0097384827163890117904710585895017157709,0.0141819536809814375774418476873868058861,0.0234920802632599185901000755354880717089,
                   0.0260391156277814168537621978300080448849,-0.032172735244776574653042549946621930761,-0.0353060796270905194273551111964057550534,-0.031676332791647481452671566892608286699,-0.0096537479030044227543083727314320573433,
                   0.0036719938185166185776220853237781711169,0.0013353515982516327667752822067616220912,0.001253073117238974538927180940938837281,0.00730174065395965655230562962757563947132,0.0275780061814833814223779146762218288831,
                   0.0299146484017483672332247177932383779088,0.029996926882761025461072819059061162719,0.0239482593460403434476943703724243605287,0.000922735244776574653042549946621930761,0.0040560796270905194273551111964057550534,
                   0.000426332791647481452671566892608286699,-0.0215962520969955772456916272685679426567,-0.0502390969540379973585146040456957676803,-0.0416947931565696429281348572102527297056,0.0502390969540379973585146040456957676803,
                   0.0416947931565696429281348572102527297056,0.0077191777763886318022917719505536462906,0.0162634815738569862326715187859966842653,-0.0077191777763886318022917719505536462906,-0.0162634815738569862326715187859966842653,
                   -0.0048757990663587735419365738109065850411,-0.0074707357752195330356074598868559521297,0.0067861793677715593225735058890465326335,0.0041912426589107998289026198130971655449,0.0048757990663587735419365738109065850411,
                   0.00747073577521953303560745988685595212971,-0.00678617936777155932257350588904653263354,-0.0041912426589107998289026198130971655449,0.0080021989721346776482709956779452617569,-0.0080021989721346776482709956779452617569,
                   0.0000908065670713865090304893488313377062,-0.0000908065670713865090304893488313377062,0.00540726226327391815460010960199589466825,-0.00540726226327391815460010960199589466825,-0.00250413014178937298464039672711802938238,
                   0.00250413014178937298464039672711802938238,-0.0502107099283208784505109043770016775965,0.0195328705180009129986253670865536047784,0.0117171294819990870013746329134463952216,0.0189607099283208784505109043770016775965,
                   -0.0184548138043057093007728327153930968256,0.0288259751331789651624621082733345241071,0.0024240248668210348375378917266654758929,-0.0127951861956942906992271672846069031744,-0.0148250669688626713260892884115956284711,
                   0.028743696652166306934614007007511739297,0.002506303347833693065385992992488260703,-0.0164249330311373286739107115884043715289,-0.0276917225865487265517767271618118041787,0.0231626173534439509733089113903510731328,
                   0.0080873826465560490266910886096489268672,-0.0035582774134512734482232728381881958213,0.0466316532423545418897914824325329771498,0.0028558132657819212164508071654311510472,-0.0341058132657819212164508071654311510472,
                   -0.0153816532423545418897914824325329771498,0.0538179291132051540775042376384453553891,0.0039590444632384543929457762003133361986,-0.0352090444632384543929457762003133361986,-0.0225679291132051540775042376384453553891,
                   0.0445078025309266730648460097903440895663,-0.0004365251570147101640162491161423600039,-0.0308134748429852898359837508838576399961,-0.0132578025309266730648460097903440895663,0.0303310203309621368265003431920266480359,
                   -0.0064543133164965597962074206826701147756,-0.0247956866835034402037925793173298852244,0.00091897966903786317349965680797335196411,0.0215115172836109882095289414104982842291,-0.0128087436962433952392453953662144185731,
                   -0.0184412563037566047607546046337855814269,0.0097384827163890117904710585895017157709,0.0170680463190185624225581523126131941139,-0.0155820947772678747209782744325980562064,-0.0156679052227321252790217255674019437936,
                   0.0141819536809814375774418476873868058861,0.0077579197367400814098999244645119282911,-0.0199776643975210392779402997490537524089,-0.0112723356024789607220597002509462475911,0.0234920802632599185901000755354880717089,
                   0.0052108843722185831462378021699919551151,-0.0221188702785218762519036232143156843959,-0.0091311297214781237480963767856843156041,0.0260391156277814168537621978300080448849,-0.032172735244776574653042549946621930761,
                   0.0036719938185166185776220853237781711169,0.0275780061814833814223779146762218288831,0.000922735244776574653042549946621930761,-0.0353060796270905194273551111964057550534,0.0013353515982516327667752822067616220912,
                   0.0299146484017483672332247177932383779088,0.0040560796270905194273551111964057550534,-0.031676332791647481452671566892608286699,0.001253073117238974538927180940938837281,0.029996926882761025461072819059061162719,
                   0.000426332791647481452671566892608286699,-0.0096537479030044227543083727314320573433,0.00730174065395965655230562962757563947132,0.0239482593460403434476943703724243605287,-0.0215962520969955772456916272685679426567,
                   -0.0502390969540379973585146040456957676803,0.0502390969540379973585146040456957676803,-0.0416947931565696429281348572102527297056,0.0416947931565696429281348572102527297056,0.0077191777763886318022917719505536462906,
                   -0.0077191777763886318022917719505536462906,0.0162634815738569862326715187859966842653,-0.0162634815738569862326715187859966842653,0.0080021989721346776482709956779452617569,-0.0080021989721346776482709956779452617569,
                   0.0000908065670713865090304893488313377062,-0.0000908065670713865090304893488313377062,0.0054072622632739181546001096019958946682,-0.00540726226327391815460010960199589466825,-0.0025041301417893729846403967271180293824,
                   0.0025041301417893729846403967271180293824,-0.0048757990663587735419365738109065850411,-0.0074707357752195330356074598868559521297,0.0067861793677715593225735058890465326335,0.0041912426589107998289026198130971655449,
                   0.0048757990663587735419365738109065850411,0.00747073577521953303560745988685595212971,-0.00678617936777155932257350588904653263354,-0.0041912426589107998289026198130971655449,-0.0502107099283208784505109043770016775965,
                   -0.0184548138043057093007728327153930968256,-0.0148250669688626713260892884115956284711,-0.0276917225865487265517767271618118041787,0.0466316532423545418897914824325329771498,0.0538179291132051540775042376384453553891,
                   0.0445078025309266730648460097903440895663,0.0303310203309621368265003431920266480359,0.0215115172836109882095289414104982842291,0.0170680463190185624225581523126131941139,0.0077579197367400814098999244645119282911,
                   0.0052108843722185831462378021699919551151,-0.032172735244776574653042549946621930761,-0.0353060796270905194273551111964057550534,-0.031676332791647481452671566892608286699,-0.0096537479030044227543083727314320573433,
                   0.0195328705180009129986253670865536047784,0.0288259751331789651624621082733345241071,0.028743696652166306934614007007511739297,0.0231626173534439509733089113903510731328,0.0028558132657819212164508071654311510472,
                   0.0039590444632384543929457762003133361986,-0.0004365251570147101640162491161423600039,-0.0064543133164965597962074206826701147756,-0.0128087436962433952392453953662144185731,-0.0155820947772678747209782744325980562064,
                   -0.0199776643975210392779402997490537524089,-0.0221188702785218762519036232143156843959,0.0036719938185166185776220853237781711169,0.0013353515982516327667752822067616220912,0.001253073117238974538927180940938837281,
                   0.00730174065395965655230562962757563947132,0.0117171294819990870013746329134463952216,0.0024240248668210348375378917266654758929,0.002506303347833693065385992992488260703,0.0080873826465560490266910886096489268672,
                   -0.0341058132657819212164508071654311510472,-0.0352090444632384543929457762003133361986,-0.0308134748429852898359837508838576399961,-0.0247956866835034402037925793173298852244,-0.0184412563037566047607546046337855814269,
                   -0.0156679052227321252790217255674019437936,-0.0112723356024789607220597002509462475911,-0.0091311297214781237480963767856843156041,0.0275780061814833814223779146762218288831,0.0299146484017483672332247177932383779088,
                   0.029996926882761025461072819059061162719,0.02394825934604034344769437037242436052868,0.0189607099283208784505109043770016775965,-0.0127951861956942906992271672846069031744,-0.0164249330311373286739107115884043715289,
                   -0.0035582774134512734482232728381881958213,-0.0153816532423545418897914824325329771498,-0.02256792911320515407750423763844535538908,-0.0132578025309266730648460097903440895663,0.00091897966903786317349965680797335196411,
                   0.0097384827163890117904710585895017157709,0.0141819536809814375774418476873868058861,0.0234920802632599185901000755354880717089,0.0260391156277814168537621978300080448849,0.000922735244776574653042549946621930761,
                   0.0040560796270905194273551111964057550534,0.000426332791647481452671566892608286699,-0.0215962520969955772456916272685679426567,-0.0048757990663587735419365738109065850411,-0.0074707357752195330356074598868559521297,
                   0.0067861793677715593225735058890465326335,0.0041912426589107998289026198130971655449,0.0048757990663587735419365738109065850411,0.00747073577521953303560745988685595212971,-0.00678617936777155932257350588904653263354,
                   -0.0041912426589107998289026198130971655449,-0.0502390969540379973585146040456957676803,-0.0416947931565696429281348572102527297056,0.0502390969540379973585146040456957676803,0.0416947931565696429281348572102527297056,
                   0.0077191777763886318022917719505536462906,0.01626348157385698623267151878599668426531,-0.00771917777638863180229177195055364629062,-0.0162634815738569862326715187859966842653,0.0080021989721346776482709956779452617569,
                   0.0000908065670713865090304893488313377062,-0.0080021989721346776482709956779452617569,-0.0000908065670713865090304893488313377062,0.00540726226327391815460010960199589466825,-0.00250413014178937298464039672711802938238,
                   -0.00540726226327391815460010960199589466825,0.00250413014178937298464039672711802938238,-0.092673241095083897556667647924639387695,0.0006690462172505151604356785398881365911,0.03193221960553429496614659994112452163674,
                   0.0476352664115395937591992934942596408596,0.0006690462172505151604356785398881365911,0.015911833971151015602001766264350897851,0.00270841919340594642331468943185163379452,-0.00792853988813659110980276714748307329997,
                   0.03193221960553429496614659994112452163674,0.00270841919340594642331468943185163379452,-0.02132867235796290844863114512805416544009,-0.02467272593464821901677951133352958492788,0.0476352664115395937591992934942596408596,
                   -0.00792853988813659110980276714748307329997,-0.02467272593464821901677951133352958492788,-0.00259729172799528996173093906387989402414,0.0006690462172505151604356785398881365911,0.015911833971151015602001766264350897851,
                   0.00270841919340594642331468943185163379452,-0.00792853988813659110980276714748307329997,0.015911833971151015602001766264350897851,0.0122564763026199587871651457168089490727,-0.00346533706211362967324109508389755666765,
                   -0.01574727700912569914630556373270532823079,0.00270841919340594642331468943185163379452,-0.00346533706211362967324109508389755666765,-0.00532580217839269944068295554901383573742,-0.00287297615543126287901089196349720341478,
                   -0.00792853988813659110980276714748307329997,-0.01574727700912569914630556373270532823079,-0.00287297615543126287901089196349720341478,0.01518803355902266705916985575507801000883,0.03193221960553429496614659994112452163674,
                   0.00270841919340594642331468943185163379452,-0.02132867235796290844863114512805416544009,-0.02467272593464821901677951133352958492788,0.00270841919340594642331468943185163379452,-0.00346533706211362967324109508389755666765,
                   -0.00532580217839269944068295554901383573742,-0.00287297615543126287901089196349720341478,-0.02132867235796290844863114512805416544009,-0.00532580217839269944068295554901383573742,0.0141169414188990285546070061819252281425,
                   0.02149322931998822490432734765969973506035,-0.02467272593464821901677951133352958492788,-0.00287297615543126287901089196349720341478,0.02149322931998822490432734765969973506035,0.01741323226376214306741242272593464821902,
                   0.0476352664115395937591992934942596408596,-0.00792853988813659110980276714748307329997,-0.02467272593464821901677951133352958492788,-0.00259729172799528996173093906387989402414,-0.00792853988813659110980276714748307329997,
                   -0.01574727700912569914630556373270532823079,-0.00287297615543126287901089196349720341478,0.01518803355902266705916985575507801000883,-0.02467272593464821901677951133352958492788,-0.00287297615543126287901089196349720341478,
                   0.02149322931998822490432734765969973506035,0.01741323226376214306741242272593464821902,-0.00259729172799528996173093906387989402414,0.01518803355902266705916985575507801000883,0.01741323226376214306741242272593464821902,
                   -0.0424406829555490138357374153664998528113,-0.00800559317044450986164262584633500147189,0.0131954665881660288489843979982337356491,0.0131954665881660288489843979982337356491,-0.0183853400058875478363261701501324698263,
                   -0.00908301442449219899911686782455107447748,0.00389314100677068001177509567265234030026,0.00389314100677068001177509567265234030026,0.00129673241095083897556667647924639387695,-0.00800559317044450986164262584633500147189,
                   0.0131954665881660288489843979982337356491,0.0131954665881660288489843979982337356491,-0.0183853400058875478363261701501324698263,-0.00908301442449219899911686782455107447748,0.00389314100677068001177509567265234030026,
                   0.00389314100677068001177509567265234030026,0.00129673241095083897556667647924639387695,-0.00800559317044450986164262584633500147189,0.0131954665881660288489843979982337356491,0.0131954665881660288489843979982337356491,
                   -0.0183853400058875478363261701501324698263,-0.00908301442449219899911686782455107447748,0.00389314100677068001177509567265234030026,0.00389314100677068001177509567265234030026,0.00129673241095083897556667647924639387695,
                   -0.0502107099283208784505109043770016775965,0.0466316532423545418897914824325329771498,0.0215115172836109882095289414104982842291,-0.032172735244776574653042549946621930761,-0.0184548138043057093007728327153930968256,
                   0.0538179291132051540775042376384453553891,0.0170680463190185624225581523126131941139,-0.0353060796270905194273551111964057550534,-0.0148250669688626713260892884115956284711,0.0445078025309266730648460097903440895663,
                   0.0077579197367400814098999244645119282911,-0.031676332791647481452671566892608286699,-0.0276917225865487265517767271618118041787,0.0303310203309621368265003431920266480359,0.0052108843722185831462378021699919551151,
                   -0.0096537479030044227543083727314320573433,0.0195328705180009129986253670865536047784,0.0028558132657819212164508071654311510472,-0.0128087436962433952392453953662144185731,0.0036719938185166185776220853237781711169,
                   0.0288259751331789651624621082733345241071,0.0039590444632384543929457762003133361986,-0.0155820947772678747209782744325980562064,0.0013353515982516327667752822067616220912,0.028743696652166306934614007007511739297,
                   -0.0004365251570147101640162491161423600039,-0.0199776643975210392779402997490537524089,0.001253073117238974538927180940938837281,0.0231626173534439509733089113903510731328,-0.0064543133164965597962074206826701147756,
                   -0.0221188702785218762519036232143156843959,0.00730174065395965655230562962757563947132,0.0117171294819990870013746329134463952216,-0.0341058132657819212164508071654311510472,-0.0184412563037566047607546046337855814269,
                   0.0275780061814833814223779146762218288831,0.0024240248668210348375378917266654758929,-0.0352090444632384543929457762003133361986,-0.0156679052227321252790217255674019437936,0.0299146484017483672332247177932383779088,
                   0.002506303347833693065385992992488260703,-0.0308134748429852898359837508838576399961,-0.0112723356024789607220597002509462475911,0.029996926882761025461072819059061162719,0.0080873826465560490266910886096489268672,
                   -0.0247956866835034402037925793173298852244,-0.0091311297214781237480963767856843156041,0.02394825934604034344769437037242436052868,0.0189607099283208784505109043770016775965,-0.0153816532423545418897914824325329771498,
                   0.0097384827163890117904710585895017157709,0.000922735244776574653042549946621930761,-0.0127951861956942906992271672846069031744,-0.0225679291132051540775042376384453553891,0.0141819536809814375774418476873868058861,
                   0.0040560796270905194273551111964057550534,-0.0164249330311373286739107115884043715289,-0.0132578025309266730648460097903440895663,0.0234920802632599185901000755354880717089,0.000426332791647481452671566892608286699,
                   -0.0035582774134512734482232728381881958213,0.00091897966903786317349965680797335196411,0.0260391156277814168537621978300080448849,-0.0215962520969955772456916272685679426567,-0.0048757990663587735419365738109065850411,
                   0.0067861793677715593225735058890465326335,-0.0074707357752195330356074598868559521297,0.0041912426589107998289026198130971655449,0.0048757990663587735419365738109065850411,-0.00678617936777155932257350588904653263354,
                   0.00747073577521953303560745988685595212971,-0.0041912426589107998289026198130971655449,0.0080021989721346776482709956779452617569,0.0000908065670713865090304893488313377062,-0.0080021989721346776482709956779452617569,
                   -0.0000908065670713865090304893488313377062,0.0054072622632739181546001096019958946682,-0.00250413014178937298464039672711802938238,-0.00540726226327391815460010960199589466825,0.0025041301417893729846403967271180293824,
                   -0.0502390969540379973585146040456957676803,-0.0416947931565696429281348572102527297056,0.0502390969540379973585146040456957676803,0.0416947931565696429281348572102527297056,0.00771917777638863180229177195055364629062,
                   0.01626348157385698623267151878599668426531,-0.00771917777638863180229177195055364629062,-0.01626348157385698623267151878599668426531,-0.01679014316760713741040109806313863047125,0.01537951782088931954215874707969670096659,
                   0.01009255892106122140182431982543822307942,-0.0107774849397590361445783132530120481928,0.01641834259961039230013862721910101301083,0.0063212524639323607347602060631393151346,0.0010997334765905885828614718859241252612,
                   0.00804496394669587440671778021175611537057,-0.0033975092662770589668052938857676796775,-0.019342085797265694068093539396472648468,-0.0141205668099239219161948052192574585946,0.00497586938663745892661555312157721796276,
                   0.0037693098342738040770677647298052971379,-0.0023586844875559862088254137463633676333,0.00292827441227211193150901350789511025391,-0.00224334839357429718875502008032128514056,0.01537951782088931954215874707969670096659,
                   0.0450043475857570845115962293533337201681,0.0358408064074479737130339539978094194962,0.00976242921329438283329798758602581489064,0.0063212524639323607347602060631393151346,-0.00115189252406634593610344716033293126476,
                   -0.0076656056042351223074114640379700620664,-0.00592870846746316981688848904547140529043,-0.019342085797265694068093539396472648468,-0.0274939408092669873972298861730004020686,-0.02098022772909821102592186929536327126689,
                   -0.00709212486587016351644484428786192804291,-0.0023586844875559862088254137463633676333,0.00447481908090958215507043731333294649852,0.0136383602592186929536327126688572471705,0.00325840412003895050003534574730751844269,
                   0.0100925589210612214018243198254382230794,0.0358408064074479737130339539978094194962,0.0266772652291388629144716786422851188242,0.00447547031346628469296356033176733700347,0.00109973347659058858286147188592412526122,
                   -0.0076656056042351223074114640379700620664,-0.01417931868440389867871948091560719286814,-0.01115022745480494196878722322268659516385,-0.01412056680992392191619480521925745859455,-0.02098022772909821102592186929536327126689,
                   -0.0144665146489294346546138524177261404652,-0.00187060587852839136454611011064673816949,0.00292827441227211193150901350789511025391,0.0136383602592186929536327126688572471705,0.02280190143752780375219498802438154784248,
                   0.00854536301986704864036977300156599632986,-0.0107774849397590361445783132530120481928,0.00976242921329438283329798758602581489064,0.00447547031346628469296356033176733700347,-0.00476482671191093487875552844288546591429,
                   0.0080449639466958744067177802117561153706,-0.00592870846746316981688848904547140529043,-0.01115022745480494196878722322268659516385,-0.00032841470621864348670306679558878226968,0.00497586938663745892661555312157721796276,
                   -0.00709212486587016351644484428786192804291,-0.00187060587852839136454611011064673816949,0.01334924803955197682003640012892211560302,-0.00224334839357429718875502008032128514056,0.00325840412003895050003534574730751844269,
                   0.00854536301986704864036977300156599632986,-0.00825600662142239845457780489044786741904,-0.0016519322826831341904291932251873281816,-0.00428906308437089790350936200155863619846,0.00165193228268313419042919322518732818158,
                   0.00428906308437089790350936200155863619846,0.00223539386202697676440750000813037770996,-0.00040173693966078694867266876824093030692,-0.00223539386202697676440750000813037770996,0.00040173693966078694867266876824093030692,
                   -0.0408030868305511180809843816300020865117,-0.02148366192040890836071558963125228185469,-0.02148366192040890836071558963125228185469,-0.0021642370102666986404467976325024771977,0.01996975349721778474765104829666875317834,
                   0.00065032858707557502738225629791894852136,0.00065032858707557502738225629791894852136,-0.01866909632306663469288653570083085613562,-0.00165193228268313419042919322518732818158,0.00165193228268313419042919322518732818158,
                   -0.00428906308437089790350936200155863619846,0.00428906308437089790350936200155863619846,0.00223539386202697676440750000813037770996,-0.00223539386202697676440750000813037770996,-0.00040173693966078694867266876824093030692,
                   0.00040173693966078694867266876824093030692,-0.0502107099283208784505109043770016775965,0.0195328705180009129986253670865536047784,0.0117171294819990870013746329134463952216,0.0189607099283208784505109043770016775965,
                   0.0466316532423545418897914824325329771498,0.0028558132657819212164508071654311510472,-0.0341058132657819212164508071654311510472,-0.0153816532423545418897914824325329771498,0.0215115172836109882095289414104982842291,
                   -0.0128087436962433952392453953662144185731,-0.0184412563037566047607546046337855814269,0.0097384827163890117904710585895017157709,-0.032172735244776574653042549946621930761,0.0036719938185166185776220853237781711169,
                   0.0275780061814833814223779146762218288831,0.000922735244776574653042549946621930761,-0.0184548138043057093007728327153930968256,0.0288259751331789651624621082733345241071,0.0024240248668210348375378917266654758929,
                   -0.0127951861956942906992271672846069031744,0.0538179291132051540775042376384453553891,0.0039590444632384543929457762003133361986,-0.0352090444632384543929457762003133361986,-0.0225679291132051540775042376384453553891,
                   0.0170680463190185624225581523126131941139,-0.0155820947772678747209782744325980562064,-0.0156679052227321252790217255674019437936,0.0141819536809814375774418476873868058861,-0.0353060796270905194273551111964057550534,
                   0.0013353515982516327667752822067616220912,0.0299146484017483672332247177932383779088,0.0040560796270905194273551111964057550534,-0.0148250669688626713260892884115956284711,0.028743696652166306934614007007511739297,
                   0.002506303347833693065385992992488260703,-0.0164249330311373286739107115884043715289,0.0445078025309266730648460097903440895663,-0.0004365251570147101640162491161423600039,-0.0308134748429852898359837508838576399961,
                   -0.0132578025309266730648460097903440895663,0.0077579197367400814098999244645119282911,-0.0199776643975210392779402997490537524089,-0.0112723356024789607220597002509462475911,0.0234920802632599185901000755354880717089,
                   -0.031676332791647481452671566892608286699,0.001253073117238974538927180940938837281,0.029996926882761025461072819059061162719,0.000426332791647481452671566892608286699,-0.0276917225865487265517767271618118041787,
                   0.0231626173534439509733089113903510731328,0.0080873826465560490266910886096489268672,-0.0035582774134512734482232728381881958213,0.0303310203309621368265003431920266480359,-0.0064543133164965597962074206826701147756,
                   -0.0247956866835034402037925793173298852244,0.00091897966903786317349965680797335196411,0.0052108843722185831462378021699919551151,-0.0221188702785218762519036232143156843959,-0.0091311297214781237480963767856843156041,
                   0.0260391156277814168537621978300080448849,-0.0096537479030044227543083727314320573433,0.00730174065395965655230562962757563947132,0.0239482593460403434476943703724243605287,-0.0215962520969955772456916272685679426567,
                   0.0080021989721346776482709956779452617569,-0.0080021989721346776482709956779452617569,0.0000908065670713865090304893488313377062,-0.0000908065670713865090304893488313377062,0.0054072622632739181546001096019958946682,
                   -0.00540726226327391815460010960199589466825,-0.0025041301417893729846403967271180293824,0.0025041301417893729846403967271180293824,-0.0502390969540379973585146040456957676803,0.0502390969540379973585146040456957676803,
                   -0.0416947931565696429281348572102527297056,0.0416947931565696429281348572102527297056,0.0077191777763886318022917719505536462906,-0.00771917777638863180229177195055364629062,0.0162634815738569862326715187859966842653,
                   -0.0162634815738569862326715187859966842653,-0.0048757990663587735419365738109065850411,0.0067861793677715593225735058890465326335,-0.0074707357752195330356074598868559521297,0.0041912426589107998289026198130971655449,
                   0.0048757990663587735419365738109065850411,-0.00678617936777155932257350588904653263354,0.00747073577521953303560745988685595212971,-0.0041912426589107998289026198130971655449,-0.0502107099283208784505109043770016775965,
                   0.0466316532423545418897914824325329771498,0.0215115172836109882095289414104982842291,-0.032172735244776574653042549946621930761,0.0195328705180009129986253670865536047784,0.0028558132657819212164508071654311510472,
                   -0.0128087436962433952392453953662144185731,0.0036719938185166185776220853237781711169,0.0117171294819990870013746329134463952216,-0.0341058132657819212164508071654311510472,-0.0184412563037566047607546046337855814269,
                   0.0275780061814833814223779146762218288831,0.0189607099283208784505109043770016775965,-0.0153816532423545418897914824325329771498,0.0097384827163890117904710585895017157709,0.000922735244776574653042549946621930761,
                   -0.0184548138043057093007728327153930968256,0.0538179291132051540775042376384453553891,0.0170680463190185624225581523126131941139,-0.0353060796270905194273551111964057550534,0.0288259751331789651624621082733345241071,
                   0.0039590444632384543929457762003133361986,-0.0155820947772678747209782744325980562064,0.0013353515982516327667752822067616220912,0.0024240248668210348375378917266654758929,-0.0352090444632384543929457762003133361986,
                   -0.0156679052227321252790217255674019437936,0.0299146484017483672332247177932383779088,-0.0127951861956942906992271672846069031744,-0.0225679291132051540775042376384453553891,0.0141819536809814375774418476873868058861,
                   0.0040560796270905194273551111964057550534,-0.0148250669688626713260892884115956284711,0.0445078025309266730648460097903440895663,0.0077579197367400814098999244645119282911,-0.031676332791647481452671566892608286699,
                   0.028743696652166306934614007007511739297,-0.0004365251570147101640162491161423600039,-0.0199776643975210392779402997490537524089,0.001253073117238974538927180940938837281,0.002506303347833693065385992992488260703,
                   -0.0308134748429852898359837508838576399961,-0.0112723356024789607220597002509462475911,0.029996926882761025461072819059061162719,-0.0164249330311373286739107115884043715289,-0.0132578025309266730648460097903440895663,
                   0.0234920802632599185901000755354880717089,0.000426332791647481452671566892608286699,-0.0276917225865487265517767271618118041787,0.0303310203309621368265003431920266480359,0.0052108843722185831462378021699919551151,
                   -0.0096537479030044227543083727314320573433,0.0231626173534439509733089113903510731328,-0.0064543133164965597962074206826701147756,-0.0221188702785218762519036232143156843959,0.00730174065395965655230562962757563947132,
                   0.0080873826465560490266910886096489268672,-0.0247956866835034402037925793173298852244,-0.0091311297214781237480963767856843156041,0.0239482593460403434476943703724243605287,-0.0035582774134512734482232728381881958213,
                   0.00091897966903786317349965680797335196411,0.0260391156277814168537621978300080448849,-0.0215962520969955772456916272685679426567,0.0080021989721346776482709956779452617569,0.0000908065670713865090304893488313377062,
                   -0.0080021989721346776482709956779452617569,-0.0000908065670713865090304893488313377062,0.0054072622632739181546001096019958946682,-0.00250413014178937298464039672711802938238,-0.0054072622632739181546001096019958946682,
                   0.0025041301417893729846403967271180293824,-0.0048757990663587735419365738109065850411,0.0067861793677715593225735058890465326335,-0.0074707357752195330356074598868559521297,0.0041912426589107998289026198130971655449,
                   0.0048757990663587735419365738109065850411,-0.00678617936777155932257350588904653263354,0.00747073577521953303560745988685595212971,-0.0041912426589107998289026198130971655449,-0.0502390969540379973585146040456957676803,
                   0.0502390969540379973585146040456957676803,-0.0416947931565696429281348572102527297056,0.0416947931565696429281348572102527297056,0.0077191777763886318022917719505536462906,-0.00771917777638863180229177195055364629062,
                   0.01626348157385698623267151878599668426531,-0.0162634815738569862326715187859966842653,-0.01679014316760713741040109806313863047125,0.01641834259961039230013862721910101301083,-0.0033975092662770589668052938857676796775,
                   0.0037693098342738040770677647298052971379,0.01537951782088931954215874707969670096659,0.0063212524639323607347602060631393151346,-0.019342085797265694068093539396472648468,-0.00235868448755598620882541374636336763326,
                   0.01009255892106122140182431982543822307942,0.00109973347659058858286147188592412526122,-0.01412056680992392191619480521925745859455,0.00292827441227211193150901350789511025391,-0.0107774849397590361445783132530120481928,
                   0.00804496394669587440671778021175611537057,0.00497586938663745892661555312157721796276,-0.00224334839357429718875502008032128514056,0.01537951782088931954215874707969670096659,0.0063212524639323607347602060631393151346,
                   -0.019342085797265694068093539396472648468,-0.00235868448755598620882541374636336763326,0.0450043475857570845115962293533337201681,-0.00115189252406634593610344716033293126476,-0.0274939408092669873972298861730004020686,
                   0.00447481908090958215507043731333294649852,0.0358408064074479737130339539978094194962,-0.0076656056042351223074114640379700620664,-0.02098022772909821102592186929536327126689,0.0136383602592186929536327126688572471705,
                   0.00976242921329438283329798758602581489064,-0.00592870846746316981688848904547140529043,-0.00709212486587016351644484428786192804291,0.00325840412003895050003534574730751844269,0.01009255892106122140182431982543822307942,
                   0.00109973347659058858286147188592412526122,-0.01412056680992392191619480521925745859455,0.00292827441227211193150901350789511025391,0.0358408064074479737130339539978094194962,-0.0076656056042351223074114640379700620664,
                   -0.02098022772909821102592186929536327126689,0.0136383602592186929536327126688572471705,0.0266772652291388629144716786422851188242,-0.01417931868440389867871948091560719286814,-0.0144665146489294346546138524177261404652,
                   0.02280190143752780375219498802438154784248,0.00447547031346628469296356033176733700347,-0.01115022745480494196878722322268659516385,-0.00187060587852839136454611011064673816949,0.00854536301986704864036977300156599632986,
                   -0.0107774849397590361445783132530120481928,0.0080449639466958744067177802117561153706,0.00497586938663745892661555312157721796276,-0.00224334839357429718875502008032128514056,0.00976242921329438283329798758602581489064,
                   -0.00592870846746316981688848904547140529043,-0.00709212486587016351644484428786192804291,0.00325840412003895050003534574730751844269,0.00447547031346628469296356033176733700347,-0.01115022745480494196878722322268659516385,
                   -0.00187060587852839136454611011064673816949,0.00854536301986704864036977300156599632986,-0.00476482671191093487875552844288546591429,-0.00032841470621864348670306679558878226968,0.01334924803955197682003640012892211560302,
                   -0.00825600662142239845457780489044786741904,-0.00165193228268313419042919322518732818158,0.0016519322826831341904291932251873281816,-0.00428906308437089790350936200155863619846,0.00428906308437089790350936200155863619846,
                   0.00223539386202697676440750000813037770996,-0.00223539386202697676440750000813037770996,-0.00040173693966078694867266876824093030692,0.00040173693966078694867266876824093030692,-0.00165193228268313419042919322518732818158,
                   0.00165193228268313419042919322518732818158,-0.00428906308437089790350936200155863619846,0.00428906308437089790350936200155863619846,0.00223539386202697676440750000813037770996,-0.00223539386202697676440750000813037770996,
                   -0.00040173693966078694867266876824093030692,0.00040173693966078694867266876824093030692,-0.0408030868305511180809843816300020865117,-0.02148366192040890836071558963125228185469,-0.02148366192040890836071558963125228185469,
                   -0.0021642370102666986404467976325024771977,0.01996975349721778474765104829666875317834,0.00065032858707557502738225629791894852136,0.00065032858707557502738225629791894852136,-0.01866909632306663469288653570083085613562,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   -0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,
                   -0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,
                   -0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00520833333333333333333333333333333333333,
                   0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,
                   0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.00601265822784810126582278481012658227848,-0.0056170886075949367088607594936708860759,-0.0056170886075949367088607594936708860759,
                   0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,-0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,-0.00528695889982809814033442725425847788717,
                   0.00528695889982809814033442725425847788717,0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00528695889982809814033442725425847788717,-0.00601265822784810126582278481012658227848,
                   0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,-0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,-0.01224996093139553055164869510861072042507,
                   -0.01224996093139553055164869510861072042507,-0.0083733786529145178934208470073448976403,-0.00522151898734177215189873417721518987342,-0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00522151898734177215189873417721518987342,0.00522151898734177215189873417721518987342,0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.00522151898734177215189873417721518987342,
                   0.00837337865291451789342084700734489764026,0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,
                   0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00837337865291451789342084700734489764026,0.00522151898734177215189873417721518987342,0.00651371308016877637130801687763713080169,
                   0.00651371308016877637130801687763713080169,0.00522151898734177215189873417721518987342,-0.0052215189873417721518987341772151898734,-0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00522151898734177215189873417721518987342,-0.00837337865291451789342084700734489764026,-0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,-0.00837337865291451789342084700734489764026,
                   -0.0060126582278481012658227848101265822785,0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,-0.00601265822784810126582278481012658227848,0.00528695889982809814033442725425847788717,
                   0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00528695889982809814033442725425847788717,-0.00528695889982809814033442725425847788717,-0.00916354117830911079856227535552430067198,
                   -0.00916354117830911079856227535552430067198,-0.00528695889982809814033442725425847788717,0.00601265822784810126582278481012658227848,-0.00561708860759493670886075949367088607595,-0.00561708860759493670886075949367088607595,
                   0.00601265822784810126582278481012658227848,0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,
                   -0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,0.00388732614471011095483669323331770589155,
                   0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,
                   0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,
                   -0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,
                   0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,-0.00601265822784810126582278481012658227848,-0.0056170886075949367088607594936708860759,
                   -0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00561708860759493670886075949367088607595,-0.0056170886075949367088607594936708860759,-0.00916354117830911079856227535552430067198,
                   0.00916354117830911079856227535552430067198,0.00561708860759493670886075949367088607595,0.0060126582278481012658227848101265822785,-0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,
                   -0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,-0.0052215189873417721518987341772151898734,0.00522151898734177215189873417721518987342,0.00837337865291451789342084700734489764026,
                   -0.01224996093139553055164869510861072042507,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,
                   -0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.01224996093139553055164869510861072042507,-0.00837337865291451789342084700734489764026,-0.00522151898734177215189873417721518987342,
                   0.00522151898734177215189873417721518987342,0.00837337865291451789342084700734489764026,0.0083733786529145178934208470073448976403,0.00522151898734177215189873417721518987342,-0.0052215189873417721518987341772151898734,
                   -0.00837337865291451789342084700734489764026,0.01224996093139553055164869510861072042507,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.01224996093139553055164869510861072042507,
                   0.01224996093139553055164869510861072042507,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.01224996093139553055164869510861072042507,0.00837337865291451789342084700734489764026,
                   0.00522151898734177215189873417721518987342,-0.00522151898734177215189873417721518987342,-0.00837337865291451789342084700734489764026,-0.0060126582278481012658227848101265822785,0.00528695889982809814033442725425847788717,
                   -0.00528695889982809814033442725425847788717,0.00601265822784810126582278481012658227848,0.00561708860759493670886075949367088607595,0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,
                   -0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,-0.00561708860759493670886075949367088607595,
                   -0.00601265822784810126582278481012658227848,0.00528695889982809814033442725425847788717,-0.00528695889982809814033442725425847788717,0.00601265822784810126582278481012658227848,0.01931942491014220972026879199874980465698,
                   -0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,
                   -0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,
                   -0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,
                   0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,
                   -0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.015625,0.015625,0.015625,0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,
                   0.015625,0.015625,0.015625,0.015625,-0.015625,-0.015625,-0.015625,-0.015625,0.015625,0.015625,0.015625,0.015625,0.015625,0.015625,0.015625,0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,0.015625,0.015625,
                   0.015625,0.015625,0.015625,0.015625,0.015625,0.015625,-0.015625,-0.015625,-0.015625,-0.015625,0.015625,0.015625,0.015625,0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,-0.015625,0.015625,0.015625,0.015625,0.015625,0,0,0,
                   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.0225189873417721518987341772151898734177,0.00362974683544303797468354430379746835443,-0.00362974683544303797468354430379746835443,-0.0225189873417721518987341772151898734177,
                   0.00362974683544303797468354430379746835443,-0.00008227848101265822784810126582278481013,0.00008227848101265822784810126582278481013,-0.00362974683544303797468354430379746835443,-0.0036297468354430379746835443037974683544,
                   0.00008227848101265822784810126582278481013,-0.0000822784810126582278481012658227848101,0.0036297468354430379746835443037974683544,-0.0225189873417721518987341772151898734177,-0.00362974683544303797468354430379746835443,
                   0.00362974683544303797468354430379746835443,0.0225189873417721518987341772151898734177,-0.01630063291139240506329113924050632911392,-0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,
                   0.01630063291139240506329113924050632911392,-0.00931012658227848101265822784810126582278,-0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,0.00931012658227848101265822784810126582278,
                   0.00931012658227848101265822784810126582278,0.00439556962025316455696202531645569620253,-0.00439556962025316455696202531645569620253,-0.00931012658227848101265822784810126582278,0.01630063291139240506329113924050632911392,
                   0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,-0.01630063291139240506329113924050632911392,-0.0163006329113924050632911392405063291139,-0.00931012658227848101265822784810126582278,
                   0.00931012658227848101265822784810126582278,0.0163006329113924050632911392405063291139,-0.00931012658227848101265822784810126582278,-0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,
                   0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,0.00439556962025316455696202531645569620253,-0.00439556962025316455696202531645569620253,-0.00931012658227848101265822784810126582278,
                   0.0163006329113924050632911392405063291139,0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,-0.01630063291139240506329113924050632911392,0.0225189873417721518987341772151898734177,
                   0.0036297468354430379746835443037974683544,-0.0036297468354430379746835443037974683544,-0.02251898734177215189873417721518987341772,0.00362974683544303797468354430379746835443,-0.00008227848101265822784810126582278481013,
                   0.00008227848101265822784810126582278481013,-0.00362974683544303797468354430379746835443,-0.0036297468354430379746835443037974683544,0.00008227848101265822784810126582278481013,-0.00008227848101265822784810126582278481013,
                   0.00362974683544303797468354430379746835443,-0.0225189873417721518987341772151898734177,-0.00362974683544303797468354430379746835443,0.00362974683544303797468354430379746835443,0.02251898734177215189873417721518987341772,
                   0.0085443037974683544303797468354430379747,-0.00854430379746835443037974683544303797468,-0.0085443037974683544303797468354430379747,0.0085443037974683544303797468354430379747,0.0085443037974683544303797468354430379747,
                   -0.00854430379746835443037974683544303797468,-0.00854430379746835443037974683544303797468,0.00854430379746835443037974683544303797468,-0.0025949367088607594936708860759493670886,0.0025949367088607594936708860759493670886,
                   -0.0025949367088607594936708860759493670886,0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,
                   -0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.0025949367088607594936708860759493670886,-0.0025949367088607594936708860759493670886,0.00259493670886075949367088607594936708861,
                   0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,
                   0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,
                   0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,
                   0.015625,0.015625,-0.015625,-0.015625,0.015625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.00601265822784810126582278481012658227848,-0.0056170886075949367088607594936708860759,-0.0056170886075949367088607594936708860759,
                   0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,-0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,-0.00837337865291451789342084700734489764026,
                   0.00837337865291451789342084700734489764026,0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,
                   0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,-0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,-0.00916354117830911079856227535552430067198,
                   -0.00916354117830911079856227535552430067198,-0.00528695889982809814033442725425847788717,-0.00522151898734177215189873417721518987342,-0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00522151898734177215189873417721518987342,0.00522151898734177215189873417721518987342,0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.00522151898734177215189873417721518987342,
                   0.00528695889982809814033442725425847788717,0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,
                   0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00528695889982809814033442725425847788717,0.00522151898734177215189873417721518987342,0.00651371308016877637130801687763713080169,
                   0.00651371308016877637130801687763713080169,0.00522151898734177215189873417721518987342,-0.0052215189873417721518987341772151898734,-0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00522151898734177215189873417721518987342,-0.00528695889982809814033442725425847788717,-0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,-0.00528695889982809814033442725425847788717,
                   -0.00601265822784810126582278481012658227848,0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,-0.00601265822784810126582278481012658227848,0.00837337865291451789342084700734489764026,
                   0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00837337865291451789342084700734489764026,-0.00837337865291451789342084700734489764026,-0.01224996093139553055164869510861072042507,
                   -0.01224996093139553055164869510861072042507,-0.00837337865291451789342084700734489764026,0.00601265822784810126582278481012658227848,-0.00561708860759493670886075949367088607595,-0.00561708860759493670886075949367088607595,
                   0.00601265822784810126582278481012658227848,0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,
                   -0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.01931942491014220972026879199874980465698,
                   0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,
                   0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,
                   -0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,
                   0.0225189873417721518987341772151898734177,0.00362974683544303797468354430379746835443,-0.00362974683544303797468354430379746835443,-0.0225189873417721518987341772151898734177,-0.01630063291139240506329113924050632911392,
                   -0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,0.01630063291139240506329113924050632911392,-0.0163006329113924050632911392405063291139,-0.00931012658227848101265822784810126582278,
                   0.00931012658227848101265822784810126582278,0.0163006329113924050632911392405063291139,0.0225189873417721518987341772151898734177,0.00362974683544303797468354430379746835443,-0.0036297468354430379746835443037974683544,
                   -0.02251898734177215189873417721518987341772,0.00362974683544303797468354430379746835443,-0.00008227848101265822784810126582278481013,0.00008227848101265822784810126582278481013,-0.00362974683544303797468354430379746835443,
                   -0.00931012658227848101265822784810126582278,-0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,
                   -0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,0.00931012658227848101265822784810126582278,0.00362974683544303797468354430379746835443,-0.00008227848101265822784810126582278481013,
                   0.00008227848101265822784810126582278481013,-0.00362974683544303797468354430379746835443,-0.0036297468354430379746835443037974683544,0.00008227848101265822784810126582278481013,-0.0000822784810126582278481012658227848101,
                   0.00362974683544303797468354430379746835443,0.00931012658227848101265822784810126582278,0.00439556962025316455696202531645569620253,-0.00439556962025316455696202531645569620253,-0.00931012658227848101265822784810126582278,
                   0.00931012658227848101265822784810126582278,0.00439556962025316455696202531645569620253,-0.00439556962025316455696202531645569620253,-0.00931012658227848101265822784810126582278,-0.0036297468354430379746835443037974683544,
                   0.00008227848101265822784810126582278481013,-0.00008227848101265822784810126582278481013,0.00362974683544303797468354430379746835443,-0.0225189873417721518987341772151898734177,-0.00362974683544303797468354430379746835443,
                   0.00362974683544303797468354430379746835443,0.0225189873417721518987341772151898734177,0.01630063291139240506329113924050632911392,0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,
                   -0.01630063291139240506329113924050632911392,0.01630063291139240506329113924050632911392,0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,-0.01630063291139240506329113924050632911392,
                   -0.02251898734177215189873417721518987341772,-0.00362974683544303797468354430379746835443,0.00362974683544303797468354430379746835443,0.02251898734177215189873417721518987341772,-0.0025949367088607594936708860759493670886,
                   0.0025949367088607594936708860759493670886,-0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,
                   0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.0085443037974683544303797468354430379747,-0.00854430379746835443037974683544303797468,-0.00854430379746835443037974683544303797468,
                   0.0085443037974683544303797468354430379747,0.0085443037974683544303797468354430379747,-0.00854430379746835443037974683544303797468,-0.00854430379746835443037974683544303797468,0.00854430379746835443037974683544303797468,
                   -0.00259493670886075949367088607594936708861,-0.0025949367088607594936708860759493670886,0.0025949367088607594936708860759493670886,0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,
                   0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.0225189873417721518987341772151898734177,-0.01630063291139240506329113924050632911392,
                   -0.0163006329113924050632911392405063291139,0.0225189873417721518987341772151898734177,0.00362974683544303797468354430379746835443,-0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,
                   0.00362974683544303797468354430379746835443,-0.00362974683544303797468354430379746835443,0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,-0.00362974683544303797468354430379746835443,
                   -0.0225189873417721518987341772151898734177,0.01630063291139240506329113924050632911392,0.0163006329113924050632911392405063291139,-0.0225189873417721518987341772151898734177,0.00362974683544303797468354430379746835443,
                   -0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,0.00362974683544303797468354430379746835443,-0.00008227848101265822784810126582278481013,-0.00439556962025316455696202531645569620253,
                   -0.00439556962025316455696202531645569620253,-0.00008227848101265822784810126582278481013,0.00008227848101265822784810126582278481013,0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,
                   0.00008227848101265822784810126582278481013,-0.00362974683544303797468354430379746835443,0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,-0.00362974683544303797468354430379746835443,
                   -0.00362974683544303797468354430379746835443,0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,-0.00362974683544303797468354430379746835443,0.00008227848101265822784810126582278481013,
                   0.00439556962025316455696202531645569620253,0.00439556962025316455696202531645569620253,0.00008227848101265822784810126582278481013,-0.0000822784810126582278481012658227848101,-0.00439556962025316455696202531645569620253,
                   -0.00439556962025316455696202531645569620253,-0.00008227848101265822784810126582278481013,0.00362974683544303797468354430379746835443,-0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,
                   0.00362974683544303797468354430379746835443,-0.0225189873417721518987341772151898734177,0.0163006329113924050632911392405063291139,0.0163006329113924050632911392405063291139,-0.02251898734177215189873417721518987341772,
                   -0.00362974683544303797468354430379746835443,0.00931012658227848101265822784810126582278,0.00931012658227848101265822784810126582278,-0.00362974683544303797468354430379746835443,0.0036297468354430379746835443037974683544,
                   -0.00931012658227848101265822784810126582278,-0.00931012658227848101265822784810126582278,0.00362974683544303797468354430379746835443,0.0225189873417721518987341772151898734177,-0.01630063291139240506329113924050632911392,
                   -0.01630063291139240506329113924050632911392,0.02251898734177215189873417721518987341772,-0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.0025949367088607594936708860759493670886,
                   0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,
                   -0.00259493670886075949367088607594936708861,-0.0025949367088607594936708860759493670886,0.0025949367088607594936708860759493670886,0.00259493670886075949367088607594936708861,0.00259493670886075949367088607594936708861,
                   0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,-0.00259493670886075949367088607594936708861,0.0085443037974683544303797468354430379747,-0.00854430379746835443037974683544303797468,
                   -0.00854430379746835443037974683544303797468,0.0085443037974683544303797468354430379747,0.0085443037974683544303797468354430379747,-0.00854430379746835443037974683544303797468,-0.00854430379746835443037974683544303797468,
                   0.00854430379746835443037974683544303797468,0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,
                   -0.0056170886075949367088607594936708860759,-0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00561708860759493670886075949367088607595,-0.0056170886075949367088607594936708860759,
                   -0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00561708860759493670886075949367088607595,0.0060126582278481012658227848101265822785,-0.00837337865291451789342084700734489764026,
                   0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,-0.0052215189873417721518987341772151898734,0.00522151898734177215189873417721518987342,
                   0.00528695889982809814033442725425847788717,-0.00916354117830911079856227535552430067198,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.00916354117830911079856227535552430067198,
                   -0.00916354117830911079856227535552430067198,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.00916354117830911079856227535552430067198,-0.00528695889982809814033442725425847788717,
                   -0.00522151898734177215189873417721518987342,0.00522151898734177215189873417721518987342,0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,0.00522151898734177215189873417721518987342,
                   -0.0052215189873417721518987341772151898734,-0.00528695889982809814033442725425847788717,0.00916354117830911079856227535552430067198,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.00916354117830911079856227535552430067198,
                   0.00528695889982809814033442725425847788717,0.00522151898734177215189873417721518987342,-0.00522151898734177215189873417721518987342,-0.00528695889982809814033442725425847788717,-0.0060126582278481012658227848101265822785,
                   0.00837337865291451789342084700734489764026,-0.00837337865291451789342084700734489764026,0.00601265822784810126582278481012658227848,0.00561708860759493670886075949367088607595,0.01224996093139553055164869510861072042507,
                   -0.01224996093139553055164869510861072042507,-0.00561708860759493670886075949367088607595,0.00561708860759493670886075949367088607595,0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,
                   -0.00561708860759493670886075949367088607595,-0.00601265822784810126582278481012658227848,0.00837337865291451789342084700734489764026,-0.00837337865291451789342084700734489764026,0.00601265822784810126582278481012658227848,
                   0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,
                   0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,
                   0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,
                   -0.00263713080168776371308016877637130801688,0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,
                   -0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00716145833333333333333333333333333333333,
                   0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,
                   -0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,
                   -0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0,0,0,0,0,0,0,0,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,
                   0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,
                   0.00520833333333333333333333333333333333333,0,0,0,0,0,0,0,0,0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,-0.0060126582278481012658227848101265822785,
                   -0.00837337865291451789342084700734489764026,-0.0052215189873417721518987341772151898734,0.00522151898734177215189873417721518987342,0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,
                   0.00522151898734177215189873417721518987342,-0.0052215189873417721518987341772151898734,-0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,0.00528695889982809814033442725425847788717,
                   -0.00528695889982809814033442725425847788717,0.00601265822784810126582278481012658227848,-0.0056170886075949367088607594936708860759,-0.00916354117830911079856227535552430067198,0.009163541178309110798562275355524300672,
                   0.00561708860759493670886075949367088607595,-0.01224996093139553055164869510861072042507,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,0.01224996093139553055164869510861072042507,
                   0.0122499609313955305516486951086107204251,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.01224996093139553055164869510861072042507,0.00561708860759493670886075949367088607595,
                   0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,-0.00561708860759493670886075949367088607595,-0.0056170886075949367088607594936708860759,-0.00916354117830911079856227535552430067198,
                   0.00916354117830911079856227535552430067198,0.00561708860759493670886075949367088607595,-0.01224996093139553055164869510861072042507,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,
                   0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.01224996093139553055164869510861072042507,
                   0.00561708860759493670886075949367088607595,0.00916354117830911079856227535552430067198,-0.00916354117830911079856227535552430067198,-0.00561708860759493670886075949367088607595,0.0060126582278481012658227848101265822785,
                   -0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,-0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,-0.00522151898734177215189873417721518987342,
                   0.00522151898734177215189873417721518987342,0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,0.00522151898734177215189873417721518987342,-0.00522151898734177215189873417721518987342,
                   -0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,0.00528695889982809814033442725425847788717,-0.00528695889982809814033442725425847788717,0.00601265822784810126582278481012658227848,
                   -0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,
                   0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,
                   0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,
                   0.01931942491014220972026879199874980465698,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,
                   -0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.015625,-0.015625,-0.015625,0.015625,-0.015625,0.015625,0.015625,
                   -0.015625,-0.015625,0.015625,0.015625,-0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,
                   -0.015625,-0.015625,0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,-0.015625,0.015625,0.015625,-0.015625,-0.015625,0.015625,0.015625,
                   -0.015625,0.015625,-0.015625,-0.015625,0.015625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.00601265822784810126582278481012658227848,-0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,
                   -0.00601265822784810126582278481012658227848,-0.00528695889982809814033442725425847788717,-0.00522151898734177215189873417721518987342,0.00522151898734177215189873417721518987342,0.00528695889982809814033442725425847788717,
                   0.0052869588998280981403344272542584778872,0.00522151898734177215189873417721518987342,-0.0052215189873417721518987341772151898734,-0.00528695889982809814033442725425847788717,-0.0060126582278481012658227848101265822785,
                   0.00837337865291451789342084700734489764026,-0.00837337865291451789342084700734489764026,0.00601265822784810126582278481012658227848,-0.0056170886075949367088607594936708860759,-0.01224996093139553055164869510861072042507,
                   0.0122499609313955305516486951086107204251,0.00561708860759493670886075949367088607595,-0.00916354117830911079856227535552430067198,-0.00651371308016877637130801687763713080169,0.00651371308016877637130801687763713080169,
                   0.00916354117830911079856227535552430067198,0.009163541178309110798562275355524300672,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,-0.00916354117830911079856227535552430067198,
                   0.00561708860759493670886075949367088607595,0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,-0.00561708860759493670886075949367088607595,-0.0056170886075949367088607594936708860759,
                   -0.01224996093139553055164869510861072042507,0.01224996093139553055164869510861072042507,0.00561708860759493670886075949367088607595,-0.00916354117830911079856227535552430067198,-0.00651371308016877637130801687763713080169,
                   0.00651371308016877637130801687763713080169,0.00916354117830911079856227535552430067198,0.00916354117830911079856227535552430067198,0.00651371308016877637130801687763713080169,-0.00651371308016877637130801687763713080169,
                   -0.00916354117830911079856227535552430067198,0.00561708860759493670886075949367088607595,0.01224996093139553055164869510861072042507,-0.01224996093139553055164869510861072042507,-0.00561708860759493670886075949367088607595,
                   0.0060126582278481012658227848101265822785,-0.00837337865291451789342084700734489764026,0.00837337865291451789342084700734489764026,-0.00601265822784810126582278481012658227848,-0.0052869588998280981403344272542584778872,
                   -0.00522151898734177215189873417721518987342,0.00522151898734177215189873417721518987342,0.00528695889982809814033442725425847788717,0.00528695889982809814033442725425847788717,0.00522151898734177215189873417721518987342,
                   -0.00522151898734177215189873417721518987342,-0.00528695889982809814033442725425847788717,-0.00601265822784810126582278481012658227848,0.00837337865291451789342084700734489764026,-0.00837337865291451789342084700734489764026,
                   0.00601265822784810126582278481012658227848,-0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,
                   -0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,0.00263713080168776371308016877637130801688,-0.00263713080168776371308016877637130801688,0.00388732614471011095483669323331770589155,
                   -0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,-0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,
                   -0.00388732614471011095483669323331770589155,0.00388732614471011095483669323331770589155,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,
                   -0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,-0.01931942491014220972026879199874980465698,0.01931942491014220972026879199874980465698,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   -0.01236979166666666666666666666666666666667,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.01236979166666666666666666666666666666667,
                   0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,
                   -0.01236979166666666666666666666666666666667,0.00716145833333333333333333333333333333333,0.00716145833333333333333333333333333333333,-0.01236979166666666666666666666666666666667,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,
                   -0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,
                   0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,-0.00325520833333333333333333333333333333333,0.00325520833333333333333333333333333333333,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,
                   0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333,0.00520833333333333333333333333333333333333};
                return data;
            }

            static const double* N3_polyDegree5Inverse()
            {
                static const double data[56*88] = {0.0188731300345224395857307249712313003452,-0.0245560989643268124280782508630609896433,-0.0171274453394706559263521288837744533947,0.037383199079401611047180667433831990794,-0.0245560989643268124280782508630609896433,
                   0.1223374568469505178365937859608745684695,0.0124758342922899884925201380897583429229,-0.0663173187571921749136939010356731875719,-0.0171274453394706559263521288837744533947,0.0124758342922899884925201380897583429229,
                   -0.0472891254315304948216340621403912543153,0.0080008630609896432681242807825086306099,0.037383199079401611047180667433831990794,-0.0663173187571921749136939010356731875719,0.0080008630609896432681242807825086306099,
                   0.0063604718066743383199079401611047180667,-0.0245560989643268124280782508630609896433,0.1223374568469505178365937859608745684695,0.0124758342922899884925201380897583429229,-0.0663173187571921749136939010356731875719,
                   0.1223374568469505178365937859608745684695,0.427518124280782508630609896432681242808,0.194766685845799769850402761795166858458,0.0330517836593785960874568469505178365938,0.0124758342922899884925201380897583429229,
                   0.194766685845799769850402761795166858458,0.0079485040276179516685845799769850402762,0.0071349252013808975834292289988492520138,-0.0663173187571921749136939010356731875719,0.0330517836593785960874568469505178365938,
                   0.0071349252013808975834292289988492520138,-0.0178092635212888377445339470655926352129,-0.0171274453394706559263521288837744533947,0.0124758342922899884925201380897583429229,-0.0472891254315304948216340621403912543153,
                   0.0080008630609896432681242807825086306099,0.0124758342922899884925201380897583429229,0.194766685845799769850402761795166858458,0.0079485040276179516685845799769850402762,0.0071349252013808975834292289988492520138,
                   -0.0472891254315304948216340621403912543153,0.0079485040276179516685845799769850402762,-0.16066369390103567318757192174913693901,-0.0223216340621403912543153049482163406214,0.0080008630609896432681242807825086306099,
                   0.0071349252013808975834292289988492520138,-0.0223216340621403912543153049482163406214,0.0511257192174913693901035673187571921749,0.037383199079401611047180667433831990794,-0.0663173187571921749136939010356731875719,
                   0.0080008630609896432681242807825086306099,0.0063604718066743383199079401611047180667,-0.0663173187571921749136939010356731875719,0.0330517836593785960874568469505178365938,0.0071349252013808975834292289988492520138,
                   -0.0178092635212888377445339470655926352129,0.0080008630609896432681242807825086306099,0.0071349252013808975834292289988492520138,-0.0223216340621403912543153049482163406214,0.0511257192174913693901035673187571921749,
                   0.0063604718066743383199079401611047180667,-0.0178092635212888377445339470655926352129,0.0511257192174913693901035673187571921749,-0.0251041426927502876869965477560414269275,-0.0172439585730724971231300345224395857307,
                   -0.00028768699654775604142692750287686996548,-0.00028768699654775604142692750287686996548,0.0178193325661680092059838895281933256617,0.020092059838895281933256616800920598389,-0.00256041426927502876869965477560414269275,
                   -0.00256041426927502876869965477560414269275,-0.0149712313003452243958573072497123130035,-0.0172439585730724971231300345224395857307,-0.00028768699654775604142692750287686996548,-0.00028768699654775604142692750287686996548,
                   0.0178193325661680092059838895281933256617,0.020092059838895281933256616800920598389,-0.00256041426927502876869965477560414269275,-0.00256041426927502876869965477560414269275,-0.0149712313003452243958573072497123130035,
                   -0.0172439585730724971231300345224395857307,-0.00028768699654775604142692750287686996548,-0.00028768699654775604142692750287686996548,0.0178193325661680092059838895281933256617,0.020092059838895281933256616800920598389,
                   -0.00256041426927502876869965477560414269275,-0.00256041426927502876869965477560414269275,-0.0149712313003452243958573072497123130035,0.0638986143076332949750671269658611430763,-0.1213617463564720948687193412931809556282,
                   0.0226415459345311665986771472003539514088,0.0054782316839278864595320291522823168393,-0.1213617463564720948687193412931809556282,-0.2875580624153936390223310054533683225666,-0.1019659159405994849032823716368020165489,
                   -0.0072687720839092147109024746155547843312,0.0226415459345311665986771472003539514088,-0.1019659159405994849032823716368020165489,0.033156560963259275495562415393639022331,0.0643223058391834763142780020417151218839,
                   0.0054782316839278864595320291522823168393,-0.0072687720839092147109024746155547843312,0.0643223058391834763142780020417151218839,-0.0331884110088224012274645186037591100882,-0.0514428862667259291731865571528018785403,
                   -0.0054457122713979254063642249296257735076,-0.0450350562080519886427059422840013557313,0.0243555777028878294701079511206093484574,-0.0054457122713979254063642249296257735076,-0.0173634925201380897583429228998849252014,
                   -0.0650951045262754123513617184503260452628,0.0384238240857439169675456595287818916511,-0.0450350562080519886427059422840013557313,-0.065095104526275412351361718450326045263,-0.133696298427311085538933640199462984273,
                   -0.0400263889396273362518088256484880957455,0.0243555777028878294701079511206093484574,0.0384238240857439169675456595287818916511,-0.0400263889396273362518088256484880957455,0.0548150641942836035663039882449165149587,
                   -0.0378026306109112016285012065602782902361,0.1606216088958704992460266721870097397524,0.0315258262502460814697101616932840561533,-0.0767767274919173653350868540741958463477,0.1606216088958704992460266721870097397524,
                   0.315915443996931338703490602224779439969,0.185293153049482163406214039125431530495,0.0543169459564501758594585598805008087709,0.0315258262502460814697101616932840561533,0.1852931530494821634062140391254315304948,
                   0.0843315832374376678174146528576908323744,-0.0172977144359000899085287270941279380098,-0.0767767274919173653350868540741958463477,0.0543169459564501758594585598805008087709,-0.0172977144359000899085287270941279380098,
                   -0.0378105810719207343679917519579966837351,0.0222377253548139624088991177598772535481,-0.0309739603945827574519557641920511118823,-0.0119725058501429809446687083817885505649,0.0500520953202915228231683927886459532029,
                   -0.0309739603945827574519557641920511118823,-0.071254965010766698530411610580386951695,-0.0413043899665735108773083456627760425229,-0.088312187831702599635089086565879814825,-0.0119725058501429809446687083817885505649,
                   -0.0413043899665735108773083456627760425229,0.0392804116105803869516949685726057034074,-0.0041580125902383286349531075269471893944,0.0500520953202915228231683927886459532029,-0.088312187831702599635089086565879814825,
                   -0.0041580125902383286349531075269471893944,0.0130747506712696586114307633294975067127,0.0565737797910793691384408683986743058473,0.0146631322264233656638719929859170365499,0.0146631322264233656638719929859170365499,
                   -0.0359000442439261004661848543705083789472,0.0016540683101864536463692581836041751654,0.0104423530056441448846512137651378157707,0.0104423530056441448846512137651378157707,0.0107945590118585899176616476194535266265,
                   0.0179688638443912705516081043507203844757,0.015150764222494180300087473083253673971,-0.0166713954899608908047726613127457009314,-0.0164482325769245600469229161212283575153,-0.0223465009751929583153211845194967557837,
                   -0.010773127091692492536374392914477302663,0.0210490326207625785684857414815220722394,0.0120705954461228722832098359524519862073,0.0179688638443912705516081043507203844757,0.015150764222494180300087473083253673971,
                   -0.0166713954899608908047726613127457009314,-0.0164482325769245600469229161212283575153,-0.0223465009751929583153211845194967557837,-0.010773127091692492536374392914477302663,0.0210490326207625785684857414815220722394,
                   0.0120705954461228722832098359524519862073,0.0638986143076332949750671269658611430763,-0.1213617463564720948687193412931809556282,0.0226415459345311665986771472003539514088,0.0054782316839278864595320291522823168393,
                   -0.0514428862667259291731865571528018785403,-0.0054457122713979254063642249296257735076,-0.0450350562080519886427059422840013557313,0.0243555777028878294701079511206093484574,-0.0378026306109112016285012065602782902361,
                   0.1606216088958704992460266721870097397524,0.0315258262502460814697101616932840561533,-0.0767767274919173653350868540741958463477,0.0222377253548139624088991177598772535481,-0.0309739603945827574519557641920511118823,
                   -0.0119725058501429809446687083817885505649,0.0500520953202915228231683927886459532029,-0.1213617463564720948687193412931809556282,-0.2875580624153936390223310054533683225666,-0.1019659159405994849032823716368020165489,
                   -0.0072687720839092147109024746155547843312,-0.0054457122713979254063642249296257735076,-0.0173634925201380897583429228998849252014,-0.0650951045262754123513617184503260452628,0.0384238240857439169675456595287818916511,
                   0.1606216088958704992460266721870097397524,0.315915443996931338703490602224779439969,0.185293153049482163406214039125431530495,0.0543169459564501758594585598805008087709,-0.0309739603945827574519557641920511118823,
                   -0.071254965010766698530411610580386951695,-0.0413043899665735108773083456627760425229,-0.088312187831702599635089086565879814825,0.0226415459345311665986771472003539514088,-0.1019659159405994849032823716368020165489,
                   0.033156560963259275495562415393639022331,0.0643223058391834763142780020417151218839,-0.0450350562080519886427059422840013557313,-0.065095104526275412351361718450326045263,-0.133696298427311085538933640199462984273,
                   -0.0400263889396273362518088256484880957455,0.0315258262502460814697101616932840561533,0.1852931530494821634062140391254315304948,0.0843315832374376678174146528576908323744,-0.0172977144359000899085287270941279380098,
                   -0.0119725058501429809446687083817885505649,-0.0413043899665735108773083456627760425229,0.0392804116105803869516949685726057034074,-0.0041580125902383286349531075269471893944,0.0054782316839278864595320291522823168393,
                   -0.0072687720839092147109024746155547843312,0.0643223058391834763142780020417151218839,-0.0331884110088224012274645186037591100882,0.0243555777028878294701079511206093484574,0.0384238240857439169675456595287818916511,
                   -0.0400263889396273362518088256484880957455,0.0548150641942836035663039882449165149587,-0.0767767274919173653350868540741958463477,0.0543169459564501758594585598805008087709,-0.0172977144359000899085287270941279380098,
                   -0.0378105810719207343679917519579966837351,0.0500520953202915228231683927886459532029,-0.088312187831702599635089086565879814825,-0.0041580125902383286349531075269471893944,0.0130747506712696586114307633294975067127,
                   0.0179688638443912705516081043507203844757,0.015150764222494180300087473083253673971,-0.0166713954899608908047726613127457009314,-0.0164482325769245600469229161212283575153,-0.0223465009751929583153211845194967557837,
                   -0.010773127091692492536374392914477302663,0.0210490326207625785684857414815220722394,0.0120705954461228722832098359524519862073,0.0565737797910793691384408683986743058473,0.0146631322264233656638719929859170365499,
                   0.0146631322264233656638719929859170365499,-0.0359000442439261004661848543705083789472,0.0016540683101864536463692581836041751654,0.0104423530056441448846512137651378157707,0.0104423530056441448846512137651378157707,
                   0.0107945590118585899176616476194535266265,0.0179688638443912705516081043507203844757,-0.0166713954899608908047726613127457009314,0.015150764222494180300087473083253673971,-0.0164482325769245600469229161212283575153,
                   -0.0223465009751929583153211845194967557837,0.0210490326207625785684857414815220722394,-0.010773127091692492536374392914477302663,0.0120705954461228722832098359524519862073,0.0638986143076332949750671269658611430763,
                   -0.0514428862667259291731865571528018785403,-0.0378026306109112016285012065602782902361,0.0222377253548139624088991177598772535481,-0.1213617463564720948687193412931809556282,-0.0054457122713979254063642249296257735076,
                   0.1606216088958704992460266721870097397524,-0.0309739603945827574519557641920511118823,0.0226415459345311665986771472003539514088,-0.0450350562080519886427059422840013557313,0.0315258262502460814697101616932840561533,
                   -0.0119725058501429809446687083817885505649,0.0054782316839278864595320291522823168393,0.0243555777028878294701079511206093484574,-0.0767767274919173653350868540741958463477,0.0500520953202915228231683927886459532029,
                   -0.1213617463564720948687193412931809556282,-0.0054457122713979254063642249296257735076,0.1606216088958704992460266721870097397524,-0.0309739603945827574519557641920511118823,-0.2875580624153936390223310054533683225666,
                   -0.0173634925201380897583429228998849252014,0.315915443996931338703490602224779439969,-0.071254965010766698530411610580386951695,-0.1019659159405994849032823716368020165489,-0.0650951045262754123513617184503260452628,
                   0.1852931530494821634062140391254315304948,-0.0413043899665735108773083456627760425229,-0.0072687720839092147109024746155547843312,0.0384238240857439169675456595287818916511,0.0543169459564501758594585598805008087709,
                   -0.088312187831702599635089086565879814825,0.0226415459345311665986771472003539514088,-0.0450350562080519886427059422840013557313,0.0315258262502460814697101616932840561533,-0.0119725058501429809446687083817885505649,
                   -0.1019659159405994849032823716368020165489,-0.0650951045262754123513617184503260452628,0.1852931530494821634062140391254315304948,-0.0413043899665735108773083456627760425229,0.033156560963259275495562415393639022331,
                   -0.133696298427311085538933640199462984273,0.0843315832374376678174146528576908323744,0.0392804116105803869516949685726057034074,0.0643223058391834763142780020417151218839,-0.0400263889396273362518088256484880957455,
                   -0.0172977144359000899085287270941279380098,-0.0041580125902383286349531075269471893944,0.0054782316839278864595320291522823168393,0.0243555777028878294701079511206093484574,-0.0767767274919173653350868540741958463477,
                   0.0500520953202915228231683927886459532029,-0.0072687720839092147109024746155547843312,0.0384238240857439169675456595287818916511,0.0543169459564501758594585598805008087709,-0.088312187831702599635089086565879814825,
                   0.0643223058391834763142780020417151218839,-0.0400263889396273362518088256484880957455,-0.0172977144359000899085287270941279380098,-0.0041580125902383286349531075269471893944,-0.0331884110088224012274645186037591100882,
                   0.0548150641942836035663039882449165149587,-0.0378105810719207343679917519579966837351,0.0130747506712696586114307633294975067127,0.0179688638443912705516081043507203844757,-0.0166713954899608908047726613127457009314,
                   0.015150764222494180300087473083253673971,-0.0164482325769245600469229161212283575153,-0.0223465009751929583153211845194967557837,0.0210490326207625785684857414815220722394,-0.010773127091692492536374392914477302663,
                   0.0120705954461228722832098359524519862073,0.0179688638443912705516081043507203844757,-0.0166713954899608908047726613127457009314,0.015150764222494180300087473083253673971,-0.0164482325769245600469229161212283575153,
                   -0.0223465009751929583153211845194967557837,0.0210490326207625785684857414815220722394,-0.010773127091692492536374392914477302663,0.0120705954461228722832098359524519862073,0.0565737797910793691384408683986743058473,
                   0.0146631322264233656638719929859170365499,0.0146631322264233656638719929859170365499,-0.0359000442439261004661848543705083789472,0.0016540683101864536463692581836041751654,0.0104423530056441448846512137651378157707,
                   0.0104423530056441448846512137651378157707,0.0107945590118585899176616476194535266265,-0.0419563794591484464902186421173762945915,0.0744867062730986781619693012098075389215,0.0178689108155089167747395595496861319646,
                   -0.0498058831990794016110471806674338319908,0.0744867062730986781619693012098075389215,0.199806223590638147600172916628612831144,0.1299380479752315195353170036714340511809,0.0636766053771591746275290579088047442478,
                   0.0178689108155089167747395595496861319646,0.1299380479752315195353170036714340511809,0.0507343471255654799958597426951857331604,0.0002177775342332304357620813317015848661,-0.0498058831990794016110471806674338319908,
                   0.0636766053771591746275290579088047442478,0.0002177775342332304357620813317015848661,-0.0146818541426927502876869965477560414269,0.0774732694637757928897169403498517422568,-0.1442697508919819046401324882337540565389,
                   -0.0556156071717780578540072210958286907654,0.0914770792235190969368184558057975779495,-0.1442697508919819046401324882337540565389,-0.323653552934407364787111622554660529344,-0.2301274333525124664365170694284618335251,
                   -0.0903854442556974202543822796987353949379,-0.0556156071717780578540072210958286907654,-0.2301274333525124664365170694284618335251,-0.1483832470272343690065209052550824702723,-0.0274375310138759505848113443050151910911,
                   0.0914770792235190969368184558057975779495,-0.0903854442556974202543822796987353949379,-0.0274375310138759505848113443050151910911,0.0572809054225193465699794813718864351776,-0.0025506692528053287546958433034382401471,
                   0.0112292709341759974671367076430367569608,0.0678227537962506316936696683532126570101,-0.0455663461011562277385062195188777467258,0.0112292709341759974671367076430367569608,0.057653684790947449175297276563099347909,
                   0.1242523014959723820483314154200230149597,0.0119675908801699940940447269561193611827,0.0678227537962506316936696683532126570101,0.1242523014959723820483314154200230149597,0.1938417122171077867280398925968546221711,
                   0.0756470510560700434118155637142978915131,-0.0455663461011562277385062195188777467258,0.0119675908801699940940447269561193611827,0.0756470510560700434118155637142978915131,-0.0729833052115488824349583843254729330679,
                   -0.0298570435366321442270809359416954353663,0.0557135838112895074920391376087578619224,-0.0272358675665637690954146650349181994752,0.0007859728615266589950134253931722286153,0.0557135838112895074920391376087578619224,
                   0.0952047205021888566192363660718091097838,-0.0114073254013918570880596197051893254425,0.0175814378717859730517958366059631882417,-0.0272358675665637690954146650349181994752,-0.0114073254013918570880596197051893254425,
                   -0.1088484030327384757764504599947637922321,-0.0512674874498450447817536425131361840223,0.0007859728615266589950134253931722286153,0.0175814378717859730517958366059631882417,-0.0512674874498450447817536425131361840223,
                   0.0334934311469121595703874184886843114691,0.0171812306305977192053141420230027824965,-0.019452093265384404624910954024878075511,-0.019452093265384404624910954024878075511,-0.0199437107664955766221589006399132981411,
                   -0.0441590787318635419901242686052812635091,0.004763274699983560743054413940489889857,0.00476327469998356074305441394048988985698,0.0346325293318964205040154407243014837952,-0.0017403084491692086628795489554983225869,
                   0.0075712976662343750951345888054748814242,0.0004428400947388289160441059175236390426,-0.0062738293118039953482991457675001978799,-0.0022153877533624369067407042090586394384,-0.0036156014637027295255143356409179193989,
                   0.0035128561077928166535761472470333229827,0.0023181331092723497786788926029432358546,-0.0017403084491692086628795489554983225869,0.0075712976662343750951345888054748814242,0.0004428400947388289160441059175236390426,
                   -0.0062738293118039953482991457675001978799,-0.0022153877533624369067407042090586394384,-0.0036156014637027295255143356409179193989,0.0035128561077928166535761472470333229827,0.0023181331092723497786788926029432358546,
                   -0.0318913022631377061756808592251630226314,0.1670833401830237273275247958792262589731,0.0882816387199298591703654994794235300564,-0.0602247314921365554276946682009973149214,0.1070788947339580250972656035947175187682,
                   0.1059335854019398323195791550221929968765,0.0764894103786508849799989040495369609294,0.080350430160556742835223847882075730177,-0.1306711943120170968272234095018905145487,-0.216285378650884979998904049536960929366,
                   -0.1557209504630390706340073428681023617732,-0.0338414639158310044385993753082360677297,0.0617019562715765247410817031070195627158,-0.062411926680914022686174584908762123952,-0.0033697188887062304783823771165543317442,
                   0.0074974108170310701956271576524741081703,0.1070788947339580250972656035947175187682,0.1059335854019398323195791550221929968765,0.0764894103786508849799989040495369609294,0.080350430160556742835223847882075730177,
                   -0.073025350704148172502602882349717792756,-0.0913964806290755657844265439202148062908,-0.111658371691599539700805523590333716916,-0.1158607252452189161049920543591429667379,-0.014409775878130308510055345498383473067,
                   -0.0182074223245109321058688147295742232451,-0.0016543920214806290755657844265439202148,0.0095458518274973971176502822072442325607,-0.0253241478985149871225820592909200504137,0.0408591361170475094525727437119842183133,
                   0.0829678681023617732478491972162858238808,0.0316448230040002191900926078141268014686,-0.1306711943120170968272234095018905145487,-0.216285378650884979998904049536960929366,-0.1557209504630390706340073428681023617732,
                   -0.0338414639158310044385993753082360677297,-0.014409775878130308510055345498383473067,-0.0182074223245109321058688147295742232451,-0.0016543920214806290755657844265439202148,0.0095458518274973971176502822072442325607,
                   0.20184490246040878952271357334648473889,0.202811325278097429996164173379363252781,0.1649671557345607978519370924434215573456,0.1217690215902241218696914899446545016165,-0.0510835525234259411474601348019069538057,
                   -0.0055073428681023617732478491972162858239,-0.0537363280179735875938407583977204230369,-0.1031537892487259575867170803879664639158,0.0617019562715765247410817031070195627158,-0.062411926680914022686174584908762123952,
                   -0.0033697188887062304783823771165543317442,0.0074974108170310701956271576524741081703,-0.0253241478985149871225820592909200504137,0.0408591361170475094525727437119842183133,0.0829678681023617732478491972162858238808,
                   0.0316448230040002191900926078141268014686,-0.0510835525234259411474601348019069538057,-0.0055073428681023617732478491972162858239,-0.0537363280179735875938407583977204230369,-0.1031537892487259575867170803879664639158,
                   0.0084873897199846566935174530111238971998,0.0327405131788043180448243739382979889309,-0.031542200942517398213600745246314866568,0.0702299098580744150364403528960490985807,-0.0426172666995451805578387856868869527097,
                   -0.0459692311907501780919502438489780261932,0.0400223299906844210641678996109375856211,0.0485641678996109375856211299249273932818,0.0094949038303468683215518658556633240178,-0.0042417392733848430050961696531316784481,
                   -0.00689996712148610882788097977971395692915,0.0016468025645240835114252835771823113595,-0.0426172666995451805578387856868869527097,-0.0459692311907501780919502438489780261932,0.0400223299906844210641678996109375856211,
                   0.0485641678996109375856211299249273932818,0.0094949038303468683215518658556633240178,-0.0042417392733848430050961696531316784481,-0.0068999671214861088278809797797139569291,0.0016468025645240835114252835771823113595,
                   -0.0067869472299852046687489725464409008713,0.0041920105211244451750780864704915337827,0.0041920105211244451750780864704915337827,-0.0015970738122636856814072003945421666941,0.0129050906899008164830949641076223354704,
                   -0.01031015398104005698942407803167296838183,-0.0103101539810400569894240780316729683818,0.0077152172721792974957531919557236012932,-0.0318913022631377061756808592251630226314,0.1070788947339580250972656035947175187682,
                   -0.1306711943120170968272234095018905145487,0.0617019562715765247410817031070195627158,0.1670833401830237273275247958792262589731,0.105933585401939832319579155022192996877,-0.216285378650884979998904049536960929366,
                   -0.062411926680914022686174584908762123952,0.0882816387199298591703654994794235300564,0.0764894103786508849799989040495369609294,-0.1557209504630390706340073428681023617732,-0.0033697188887062304783823771165543317442,
                   -0.0602247314921365554276946682009973149214,0.080350430160556742835223847882075730177,-0.0338414639158310044385993753082360677297,0.0074974108170310701956271576524741081703,0.1070788947339580250972656035947175187682,
                   -0.073025350704148172502602882349717792756,-0.014409775878130308510055345498383473067,-0.0253241478985149871225820592909200504137,0.1059335854019398323195791550221929968765,-0.0913964806290755657844265439202148062908,
                   -0.0182074223245109321058688147295742232451,0.0408591361170475094525727437119842183133,0.0764894103786508849799989040495369609294,-0.111658371691599539700805523590333716916,-0.0016543920214806290755657844265439202148,
                   0.0829678681023617732478491972162858238808,0.080350430160556742835223847882075730177,-0.1158607252452189161049920543591429667379,0.0095458518274973971176502822072442325607,0.0316448230040002191900926078141268014686,
                   -0.1306711943120170968272234095018905145487,-0.014409775878130308510055345498383473067,0.20184490246040878952271357334648473889,-0.0510835525234259411474601348019069538057,-0.216285378650884979998904049536960929366,
                   -0.0182074223245109321058688147295742232451,0.202811325278097429996164173379363252781,-0.0055073428681023617732478491972162858239,-0.1557209504630390706340073428681023617732,-0.0016543920214806290755657844265439202148,
                   0.1649671557345607978519370924434215573456,-0.0537363280179735875938407583977204230369,-0.0338414639158310044385993753082360677297,0.0095458518274973971176502822072442325607,0.1217690215902241218696914899446545016165,
                   -0.1031537892487259575867170803879664639158,0.0617019562715765247410817031070195627158,-0.0253241478985149871225820592909200504137,-0.0510835525234259411474601348019069538057,0.0084873897199846566935174530111238971998,
                   -0.062411926680914022686174584908762123952,0.0408591361170475094525727437119842183133,-0.0055073428681023617732478491972162858239,0.0327405131788043180448243739382979889309,-0.0033697188887062304783823771165543317442,
                   0.0829678681023617732478491972162858238808,-0.0537363280179735875938407583977204230369,-0.031542200942517398213600745246314866568,0.0074974108170310701956271576524741081703,0.0316448230040002191900926078141268014686,
                   -0.1031537892487259575867170803879664639158,0.0702299098580744150364403528960490985807,-0.0426172666995451805578387856868869527097,0.0400223299906844210641678996109375856211,-0.0459692311907501780919502438489780261932,
                   0.0485641678996109375856211299249273932818,0.0094949038303468683215518658556633240178,-0.0068999671214861088278809797797139569291,-0.0042417392733848430050961696531316784481,0.0016468025645240835114252835771823113595,
                   -0.0067869472299852046687489725464409008713,0.0041920105211244451750780864704915337827,0.0041920105211244451750780864704915337827,-0.0015970738122636856814072003945421666941,0.0129050906899008164830949641076223354704,
                   -0.01031015398104005698942407803167296838183,-0.01031015398104005698942407803167296838183,0.0077152172721792974957531919557236012932,-0.0426172666995451805578387856868869527097,-0.0459692311907501780919502438489780261932,
                   0.0400223299906844210641678996109375856211,0.0485641678996109375856211299249273932818,0.0094949038303468683215518658556633240178,-0.0042417392733848430050961696531316784481,-0.0068999671214861088278809797797139569291,
                   0.0016468025645240835114252835771823113595,-0.0419563794591484464902186421173762945915,0.0744867062730986781619693012098075389215,0.0178689108155089167747395595496861319646,-0.0498058831990794016110471806674338319908,
                   0.0774732694637757928897169403498517422568,-0.1442697508919819046401324882337540565389,-0.0556156071717780578540072210958286907654,0.0914770792235190969368184558057975779495,-0.0025506692528053287546958433034382401471,
                   0.0112292709341759974671367076430367569608,0.0678227537962506316936696683532126570101,-0.0455663461011562277385062195188777467258,-0.0298570435366321442270809359416954353663,0.0557135838112895074920391376087578619224,
                   -0.0272358675665637690954146650349181994752,0.0007859728615266589950134253931722286153,0.0744867062730986781619693012098075389215,0.199806223590638147600172916628612831144,0.1299380479752315195353170036714340511809,
                   0.0636766053771591746275290579088047442478,-0.1442697508919819046401324882337540565389,-0.323653552934407364787111622554660529344,-0.2301274333525124664365170694284618335251,-0.0903854442556974202543822796987353949379,
                   0.0112292709341759974671367076430367569608,0.057653684790947449175297276563099347909,0.1242523014959723820483314154200230149597,0.0119675908801699940940447269561193611827,0.0557135838112895074920391376087578619224,
                   0.0952047205021888566192363660718091097838,-0.0114073254013918570880596197051893254425,0.0175814378717859730517958366059631882417,0.0178689108155089167747395595496861319646,0.1299380479752315195353170036714340511809,
                   0.0507343471255654799958597426951857331604,0.0002177775342332304357620813317015848661,-0.0556156071717780578540072210958286907654,-0.2301274333525124664365170694284618335251,-0.1483832470272343690065209052550824702723,
                   -0.0274375310138759505848113443050151910911,0.0678227537962506316936696683532126570101,0.1242523014959723820483314154200230149597,0.1938417122171077867280398925968546221711,0.0756470510560700434118155637142978915131,
                   -0.0272358675665637690954146650349181994752,-0.0114073254013918570880596197051893254425,-0.1088484030327384757764504599947637922321,-0.0512674874498450447817536425131361840223,-0.0498058831990794016110471806674338319908,
                   0.0636766053771591746275290579088047442478,0.0002177775342332304357620813317015848661,-0.0146818541426927502876869965477560414269,0.0914770792235190969368184558057975779495,-0.0903854442556974202543822796987353949379,
                   -0.0274375310138759505848113443050151910911,0.0572809054225193465699794813718864351776,-0.0455663461011562277385062195188777467258,0.0119675908801699940940447269561193611827,0.0756470510560700434118155637142978915131,
                   -0.0729833052115488824349583843254729330679,0.0007859728615266589950134253931722286153,0.0175814378717859730517958366059631882417,-0.0512674874498450447817536425131361840223,0.0334934311469121595703874184886843114691,
                   -0.0017403084491692086628795489554983225869,0.0075712976662343750951345888054748814242,0.0004428400947388289160441059175236390426,-0.0062738293118039953482991457675001978799,-0.0022153877533624369067407042090586394384,
                   -0.0036156014637027295255143356409179193989,0.00351285610779281665357614724703332298269,0.0023181331092723497786788926029432358546,0.0171812306305977192053141420230027824965,-0.019452093265384404624910954024878075511,
                   -0.019452093265384404624910954024878075511,-0.0199437107664955766221589006399132981411,-0.0441590787318635419901242686052812635091,0.004763274699983560743054413940489889857,0.004763274699983560743054413940489889857,
                   0.0346325293318964205040154407243014837952,-0.0017403084491692086628795489554983225869,0.0004428400947388289160441059175236390426,0.0075712976662343750951345888054748814242,-0.0062738293118039953482991457675001978799,
                   -0.0022153877533624369067407042090586394384,0.0035128561077928166535761472470333229827,-0.0036156014637027295255143356409179193989,0.0023181331092723497786788926029432358546,-0.0318913022631377061756808592251630226314,
                   0.1070788947339580250972656035947175187682,-0.1306711943120170968272234095018905145487,0.0617019562715765247410817031070195627158,0.1070788947339580250972656035947175187682,-0.073025350704148172502602882349717792756,
                   -0.014409775878130308510055345498383473067,-0.0253241478985149871225820592909200504137,-0.1306711943120170968272234095018905145487,-0.014409775878130308510055345498383473067,0.20184490246040878952271357334648473889,
                   -0.0510835525234259411474601348019069538057,0.0617019562715765247410817031070195627158,-0.0253241478985149871225820592909200504137,-0.0510835525234259411474601348019069538057,0.0084873897199846566935174530111238971998,
                   0.1670833401830237273275247958792262589731,0.105933585401939832319579155022192996877,-0.216285378650884979998904049536960929366,-0.062411926680914022686174584908762123952,0.1059335854019398323195791550221929968765,
                   -0.0913964806290755657844265439202148062908,-0.0182074223245109321058688147295742232451,0.0408591361170475094525727437119842183133,-0.216285378650884979998904049536960929366,-0.0182074223245109321058688147295742232451,
                   0.202811325278097429996164173379363252781,-0.0055073428681023617732478491972162858239,-0.062411926680914022686174584908762123952,0.0408591361170475094525727437119842183133,-0.0055073428681023617732478491972162858239,
                   0.0327405131788043180448243739382979889309,0.0882816387199298591703654994794235300564,0.0764894103786508849799989040495369609294,-0.1557209504630390706340073428681023617732,-0.0033697188887062304783823771165543317442,
                   0.0764894103786508849799989040495369609294,-0.111658371691599539700805523590333716916,-0.0016543920214806290755657844265439202148,0.0829678681023617732478491972162858238808,-0.1557209504630390706340073428681023617732,
                   -0.0016543920214806290755657844265439202148,0.1649671557345607978519370924434215573456,-0.0537363280179735875938407583977204230369,-0.0033697188887062304783823771165543317442,0.0829678681023617732478491972162858238808,
                   -0.0537363280179735875938407583977204230369,-0.031542200942517398213600745246314866568,-0.0602247314921365554276946682009973149214,0.080350430160556742835223847882075730177,-0.0338414639158310044385993753082360677297,
                   0.0074974108170310701956271576524741081703,0.080350430160556742835223847882075730177,-0.1158607252452189161049920543591429667379,0.0095458518274973971176502822072442325607,0.0316448230040002191900926078141268014686,
                   -0.0338414639158310044385993753082360677297,0.0095458518274973971176502822072442325607,0.1217690215902241218696914899446545016165,-0.1031537892487259575867170803879664639158,0.0074974108170310701956271576524741081703,
                   0.0316448230040002191900926078141268014686,-0.1031537892487259575867170803879664639158,0.0702299098580744150364403528960490985807,-0.0067869472299852046687489725464409008713,0.0041920105211244451750780864704915337827,
                   0.0041920105211244451750780864704915337827,-0.0015970738122636856814072003945421666941,0.0129050906899008164830949641076223354704,-0.0103101539810400569894240780316729683818,-0.01031015398104005698942407803167296838183,
                   0.0077152172721792974957531919557236012932,-0.0426172666995451805578387856868869527097,0.0400223299906844210641678996109375856211,-0.0459692311907501780919502438489780261932,0.0485641678996109375856211299249273932818,
                   0.0094949038303468683215518658556633240178,-0.0068999671214861088278809797797139569291,-0.0042417392733848430050961696531316784481,0.0016468025645240835114252835771823113595,-0.0426172666995451805578387856868869527097,
                   0.0400223299906844210641678996109375856211,-0.0459692311907501780919502438489780261932,0.0485641678996109375856211299249273932818,0.0094949038303468683215518658556633240178,-0.0068999671214861088278809797797139569291,
                   -0.0042417392733848430050961696531316784481,0.0016468025645240835114252835771823113595,-0.0419563794591484464902186421173762945915,0.0774732694637757928897169403498517422568,-0.0025506692528053287546958433034382401471,
                   -0.0298570435366321442270809359416954353663,0.0744867062730986781619693012098075389215,-0.1442697508919819046401324882337540565389,0.0112292709341759974671367076430367569608,0.0557135838112895074920391376087578619224,
                   0.0178689108155089167747395595496861319646,-0.0556156071717780578540072210958286907654,0.0678227537962506316936696683532126570101,-0.0272358675665637690954146650349181994752,-0.0498058831990794016110471806674338319908,
                   0.0914770792235190969368184558057975779495,-0.0455663461011562277385062195188777467258,0.0007859728615266589950134253931722286153,0.0744867062730986781619693012098075389215,-0.1442697508919819046401324882337540565389,
                   0.0112292709341759974671367076430367569608,0.0557135838112895074920391376087578619224,0.199806223590638147600172916628612831144,-0.323653552934407364787111622554660529344,0.057653684790947449175297276563099347909,
                   0.0952047205021888566192363660718091097838,0.1299380479752315195353170036714340511809,-0.2301274333525124664365170694284618335251,0.1242523014959723820483314154200230149597,-0.0114073254013918570880596197051893254425,
                   0.0636766053771591746275290579088047442478,-0.0903854442556974202543822796987353949379,0.0119675908801699940940447269561193611827,0.0175814378717859730517958366059631882417,0.0178689108155089167747395595496861319646,
                   -0.0556156071717780578540072210958286907654,0.0678227537962506316936696683532126570101,-0.0272358675665637690954146650349181994752,0.1299380479752315195353170036714340511809,-0.2301274333525124664365170694284618335251,
                   0.1242523014959723820483314154200230149597,-0.0114073254013918570880596197051893254425,0.0507343471255654799958597426951857331604,-0.1483832470272343690065209052550824702723,0.1938417122171077867280398925968546221711,
                   -0.1088484030327384757764504599947637922321,0.0002177775342332304357620813317015848661,-0.0274375310138759505848113443050151910911,0.0756470510560700434118155637142978915131,-0.0512674874498450447817536425131361840223,
                   -0.0498058831990794016110471806674338319908,0.0914770792235190969368184558057975779495,-0.0455663461011562277385062195188777467258,0.0007859728615266589950134253931722286153,0.0636766053771591746275290579088047442478,
                   -0.0903854442556974202543822796987353949379,0.0119675908801699940940447269561193611827,0.0175814378717859730517958366059631882417,0.0002177775342332304357620813317015848661,-0.0274375310138759505848113443050151910911,
                   0.0756470510560700434118155637142978915131,-0.0512674874498450447817536425131361840223,-0.0146818541426927502876869965477560414269,0.0572809054225193465699794813718864351776,-0.0729833052115488824349583843254729330679,
                   0.0334934311469121595703874184886843114691,-0.0017403084491692086628795489554983225869,0.0004428400947388289160441059175236390426,0.0075712976662343750951345888054748814242,-0.0062738293118039953482991457675001978799,
                   -0.0022153877533624369067407042090586394384,0.0035128561077928166535761472470333229827,-0.0036156014637027295255143356409179193989,0.0023181331092723497786788926029432358546,-0.0017403084491692086628795489554983225869,
                   0.0004428400947388289160441059175236390426,0.0075712976662343750951345888054748814242,-0.0062738293118039953482991457675001978799,-0.0022153877533624369067407042090586394384,0.0035128561077928166535761472470333229827,
                   -0.0036156014637027295255143356409179193989,0.0023181331092723497786788926029432358546,0.0171812306305977192053141420230027824965,-0.019452093265384404624910954024878075511,-0.019452093265384404624910954024878075511,
                   -0.0199437107664955766221589006399132981411,-0.0441590787318635419901242686052812635091,0.00476327469998356074305441394048988985698,0.00476327469998356074305441394048988985698,0.0346325293318964205040154407243014837952,
                   -0.0133279991369390103567318757192174913694,0.0189159036238413875544677232440945521114,-0.0132988150162464508456069637504236660355,0.0073153409090909090909090909090909090909,0.0189159036238413875544677232440945521114,
                   0.0432578323038502363396878164810654262131,-0.0080526244588744588744588744588744588745,0.0019960552759867105858667040101639257757,-0.0132988150162464508456069637504236660355,-0.0080526244588744588744588744588744588745,
                   -0.0531942500527679852574367342299831751309,-0.0232381438835816472947274635038348118517,0.0073153409090909090909090909090909090909,0.0019960552759867105858667040101639257757,-0.0232381438835816472947274635038348118517,
                   0.0143223173187571921749136939010356731876,-0.0032371462388603738814709278844299941346,0.0205830072753859673690897319589302711666,0.0333336203226762298492256298163471159252,-0.0300561417748917748917748917748917748918,
                   0.0205830072753859673690897319589302711666,0.068942690352896049098580744150364403529,0.0781960227272727272727272727272727272727,0.0106801783786224714494756688849515853735,0.0333336203226762298492256298163471159252,
                   0.0781960227272727272727272727272727272727,0.0923735975258918296892980437284234752589,0.0343615273566486646655423026731043608681,-0.0300561417748917748917748917748917748918,0.0106801783786224714494756688849515853735,
                   0.0343615273566486646655423026731043608681,-0.0356089035446894096683126218991197894151,0.031882979572193707214804261217763327468,-0.0648538406087193007024230652922636044999,-0.0307294536560095631825589631496804492585,
                   0.0430769751082251082251082251082251082251,-0.0648538406087193007024230652922636044999,-0.1340468570195627157652474108170310701956,-0.1068418560606060606060606060606060606061,-0.0393260117119558047828090022182849187069,
                   -0.0307294536560095631825589631496804492585,-0.1068418560606060606060606060606060606061,-0.0845610975258918296892980437284234752589,-0.0161323606899819979988756360064376942014,0.0430769751082251082251082251082251082251,
                   -0.0393260117119558047828090022182849187069,-0.0161323606899819979988756360064376942014,0.0330047368780227430016459552324531227485,-0.0153178341963943229766014576141158419639,0.0253549297094919457788656100892387812219,
                   0.0106946483495797841789402970837569993688,-0.0203361742424242424242424242424242424242,0.0253549297094919457788656100892387812219,0.0843463343628164303269788501856012404536,0.0575317911255411255411255411255411255411,
                   0.0266497780573466227474666293231694075576,0.0106946483495797841789402970837569993688,0.0575317911255411255411255411255411255411,0.0245484167194346519241034008966498417975,0.005008977216914980628060796837168145185,
                   -0.0203361742424242424242424242424242424242,0.0266497780573466227474666293231694075576,0.005008977216914980628060796837168145185,-0.0117181506520905255082470272343690065209,-0.0551689920097093092873683590983169042241,
                   -0.0058170995670995670995670995670995670996,-0.0058170995670995670995670995670995670996,0.0251365244772417768198358915658493717565,-0.0073310079902906907126316409016830957759,-0.0150162337662337662337662337662337662338,
                   -0.0150162337662337662337662337662337662338,-0.0043031911439084434865025582325160384232,0.0007277161813026791929745516243406538765,-0.0046150423260127901478112448576583597681,-0.0007277161813026791929745516243406538765,
                   0.0046150423260127901478112448576583597681,0.0019094146203850845201056171520306541404,0.0019779115243250264347310760812870517512,-0.0019094146203850845201056171520306541404,-0.00197791152432502643473107608128705175119,
                   0.0007277161813026791929745516243406538765,-0.0046150423260127901478112448576583597681,-0.0007277161813026791929745516243406538765,0.0046150423260127901478112448576583597681,0.0019094146203850845201056171520306541404,
                   0.0019779115243250264347310760812870517512,-0.0019094146203850845201056171520306541404,-0.00197791152432502643473107608128705175119,-0.0536828370732642884541618718833908707326,-0.050393434229272836867773576634336128007,
                   -0.0509844666146090196723108115513178804318,0.0080406957230533179900268507863444572305,-0.0003677954575897613872297416601214069568,-0.0198681756154066280648559129571787799636,-0.0354537008496660395394572609762483180205,
                   -0.0259588278429259441917669765771031593816,0.0816177954575897613872297416601214069569,0.0886181756154066280648559129571787799636,0.0917037008496660395394572609762483180205,0.0697088278429259441917669765771031593816,
                   -0.0275671629267357115458381281166091292674,-0.018356565770727163132226423365663871993,-0.0052655333853909803276891884486821195682,-0.0517906957230533179900268507863444572305,0.091087143093076637380434848789279169026,
                   0.1247176399696178177190835418683519949343,0.0927614094912354406025292101241468330076,0.0241371595323335829664943588994221905614,0.0011541437202038467861252671379253657735,0.0122761195818948983505945531261986958189,
                   0.0067441863252780974299961641733793632528,-0.0058020867581785303304290646062797961532,-0.1449041437202038467861252671379253657735,-0.1435261195818948983505945531261986958189,-0.1254941863252780974299961641733793632528,
                   -0.1004479132418214696695709353937202038468,0.052662856906923362619565151210720830974,0.0065323600303821822809164581316480050657,0.0259885905087645593974707898758531669924,0.0821128404676664170335056411005778094386,
                   -0.0596823435150175656504770428821061732454,-0.0221783467206726700397586473535840624448,-0.0213423694068472549485207713055814321637,-0.0628336257770593213631188314732618530087,0.0628434828620746342265329607101759000493,
                   0.0790986167050249328730341388569236670502,0.0883394107211354046797084771768316072114,0.0886794601759000493177708367581785303304,-0.0190934828620746342265329607101759000493,-0.0228486167050249328730341388569236670502,
                   -0.0195894107211354046797084771768316072114,-0.0074294601759000493177708367581785303304,0.0159323435150175656504770428821061732454,-0.0340716532793273299602413526464159375552,-0.0474076305931527450514792286944185678363,
                   -0.0184163742229406786368811685267381469913,0.0160596830648254698887610280015343306483,-0.0464654792728368677735766343361280070141,-0.0261149532166146090196723108115513178804,0.0368741249520521672420406597621787495205,
                   -0.0579494513778532765874538026436760613976,-0.0878620459035807137072959857769984352263,-0.0849410776313466186883908402895744667897,-0.0625989253216310178335494791190993722639,0.0766994513778532765874538026436760613976,
                   0.0941120459035807137072959857769984352263,0.0786910776313466186883908402895744667897,0.0438489253216310178335494791190993722639,-0.0348096830648254698887610280015343306483,0.0402154792728368677735766343361280070141,
                   0.0323649532166146090196723108115513178804,-0.0181241249520521672420406597621787495205,-0.0366333238959821238302250960478808580074,-0.0004916175011111719972479466150352226302,0.0366333238959821238302250960478808580074,
                   0.0004916175011111719972479466150352226302,0.0489223534318471027331786825457711533661,0.0298692546319128597609610267838115939382,-0.0489223534318471027331786825457711533661,-0.0298692546319128597609610267838115939382,
                   -0.0004750793041932282438611552535603168515,-0.0111868991299371046206489244463928008232,0.0030700160130539877375320413295096839401,0.0085919624210763451269780383704434337346,0.0004750793041932282438611552535603168515,
                   0.0111868991299371046206489244463928008232,-0.0030700160130539877375320413295096839401,-0.0085919624210763451269780383704434337346,0.0093116061154035837580141377609732040112,-0.0093116061154035837580141377609732040112,
                   -0.0067166694065428242643432516850238369226,0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,
                   0.0011947229985204668748972546440900871281,-0.0536828370732642884541618718833908707326,-0.0003677954575897613872297416601214069568,0.0816177954575897613872297416601214069569,-0.0275671629267357115458381281166091292674,
                   -0.050393434229272836867773576634336128007,-0.0198681756154066280648559129571787799636,0.0886181756154066280648559129571787799636,-0.018356565770727163132226423365663871993,-0.0509844666146090196723108115513178804318,
                   -0.0354537008496660395394572609762483180205,0.0917037008496660395394572609762483180205,-0.0052655333853909803276891884486821195682,0.0080406957230533179900268507863444572305,-0.0259588278429259441917669765771031593816,
                   0.0697088278429259441917669765771031593816,-0.0517906957230533179900268507863444572305,0.091087143093076637380434848789279169026,0.0011541437202038467861252671379253657735,-0.1449041437202038467861252671379253657735,
                   0.052662856906923362619565151210720830974,0.1247176399696178177190835418683519949343,0.0122761195818948983505945531261986958189,-0.1435261195818948983505945531261986958189,0.0065323600303821822809164581316480050657,
                   0.0927614094912354406025292101241468330076,0.0067441863252780974299961641733793632528,-0.1254941863252780974299961641733793632528,0.0259885905087645593974707898758531669924,0.0241371595323335829664943588994221905614,
                   -0.0058020867581785303304290646062797961532,-0.1004479132418214696695709353937202038468,0.0821128404676664170335056411005778094386,-0.0596823435150175656504770428821061732454,0.0628434828620746342265329607101759000493,
                   -0.0190934828620746342265329607101759000493,0.0159323435150175656504770428821061732454,-0.0221783467206726700397586473535840624448,0.0790986167050249328730341388569236670502,-0.0228486167050249328730341388569236670502,
                   -0.0340716532793273299602413526464159375552,-0.0213423694068472549485207713055814321637,0.0883394107211354046797084771768316072114,-0.0195894107211354046797084771768316072114,-0.0474076305931527450514792286944185678363,
                   -0.0628336257770593213631188314732618530087,0.0886794601759000493177708367581785303304,-0.0074294601759000493177708367581785303304,-0.0184163742229406786368811685267381469913,0.0160596830648254698887610280015343306483,
                   -0.0579494513778532765874538026436760613976,0.0766994513778532765874538026436760613976,-0.0348096830648254698887610280015343306483,-0.0464654792728368677735766343361280070141,-0.0878620459035807137072959857769984352263,
                   0.0941120459035807137072959857769984352263,0.0402154792728368677735766343361280070141,-0.0261149532166146090196723108115513178804,-0.0849410776313466186883908402895744667897,0.0786910776313466186883908402895744667897,
                   0.0323649532166146090196723108115513178804,0.0368741249520521672420406597621787495205,-0.0625989253216310178335494791190993722639,0.0438489253216310178335494791190993722639,-0.0181241249520521672420406597621787495205,
                   -0.0366333238959821238302250960478808580074,0.0366333238959821238302250960478808580074,-0.0004916175011111719972479466150352226302,0.0004916175011111719972479466150352226302,0.0489223534318471027331786825457711533661,
                   -0.0489223534318471027331786825457711533661,0.0298692546319128597609610267838115939382,-0.0298692546319128597609610267838115939382,0.0093116061154035837580141377609732040112,-0.0093116061154035837580141377609732040112,
                   -0.0067166694065428242643432516850238369226,0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,
                   0.0011947229985204668748972546440900871281,-0.0004750793041932282438611552535603168515,-0.0111868991299371046206489244463928008232,0.0030700160130539877375320413295096839401,0.0085919624210763451269780383704434337346,
                   0.0004750793041932282438611552535603168515,0.0111868991299371046206489244463928008232,-0.0030700160130539877375320413295096839401,-0.0085919624210763451269780383704434337346,-0.0536828370732642884541618718833908707326,
                   -0.050393434229272836867773576634336128007,-0.0509844666146090196723108115513178804318,0.0080406957230533179900268507863444572305,0.091087143093076637380434848789279169026,0.1247176399696178177190835418683519949343,
                   0.0927614094912354406025292101241468330076,0.0241371595323335829664943588994221905614,-0.0596823435150175656504770428821061732454,-0.0221783467206726700397586473535840624448,-0.0213423694068472549485207713055814321637,
                   -0.0628336257770593213631188314732618530087,0.0160596830648254698887610280015343306483,-0.0464654792728368677735766343361280070141,-0.0261149532166146090196723108115513178804,0.0368741249520521672420406597621787495205,
                   -0.0003677954575897613872297416601214069568,-0.0198681756154066280648559129571787799636,-0.0354537008496660395394572609762483180205,-0.0259588278429259441917669765771031593816,0.0011541437202038467861252671379253657735,
                   0.0122761195818948983505945531261986958189,0.0067441863252780974299961641733793632528,-0.0058020867581785303304290646062797961532,0.0628434828620746342265329607101759000493,0.0790986167050249328730341388569236670502,
                   0.0883394107211354046797084771768316072114,0.0886794601759000493177708367581785303304,-0.0579494513778532765874538026436760613976,-0.0878620459035807137072959857769984352263,-0.0849410776313466186883908402895744667897,
                   -0.0625989253216310178335494791190993722639,0.0816177954575897613872297416601214069568,0.0886181756154066280648559129571787799636,0.0917037008496660395394572609762483180205,0.0697088278429259441917669765771031593816,
                   -0.1449041437202038467861252671379253657735,-0.1435261195818948983505945531261986958189,-0.1254941863252780974299961641733793632528,-0.1004479132418214696695709353937202038468,-0.0190934828620746342265329607101759000493,
                   -0.0228486167050249328730341388569236670502,-0.0195894107211354046797084771768316072114,-0.0074294601759000493177708367581785303304,0.0766994513778532765874538026436760613976,0.0941120459035807137072959857769984352263,
                   0.0786910776313466186883908402895744667897,0.0438489253216310178335494791190993722639,-0.0275671629267357115458381281166091292674,-0.018356565770727163132226423365663871993,-0.0052655333853909803276891884486821195682,
                   -0.0517906957230533179900268507863444572305,0.052662856906923362619565151210720830974,0.0065323600303821822809164581316480050657,0.0259885905087645593974707898758531669924,0.0821128404676664170335056411005778094386,
                   0.0159323435150175656504770428821061732454,-0.0340716532793273299602413526464159375552,-0.0474076305931527450514792286944185678363,-0.0184163742229406786368811685267381469913,-0.0348096830648254698887610280015343306483,
                   0.0402154792728368677735766343361280070141,0.0323649532166146090196723108115513178804,-0.0181241249520521672420406597621787495205,-0.0004750793041932282438611552535603168515,-0.0111868991299371046206489244463928008232,
                   0.0030700160130539877375320413295096839401,0.0085919624210763451269780383704434337346,0.0004750793041932282438611552535603168515,0.0111868991299371046206489244463928008232,-0.0030700160130539877375320413295096839401,
                   -0.0085919624210763451269780383704434337346,-0.0366333238959821238302250960478808580074,-0.0004916175011111719972479466150352226302,0.0366333238959821238302250960478808580074,0.0004916175011111719972479466150352226302,
                   0.0489223534318471027331786825457711533661,0.0298692546319128597609610267838115939382,-0.0489223534318471027331786825457711533661,-0.0298692546319128597609610267838115939382,0.0093116061154035837580141377609732040112,
                   -0.0067166694065428242643432516850238369226,-0.0093116061154035837580141377609732040112,0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,
                   0.0014002137103402926187736314318592799605,0.0011947229985204668748972546440900871281,-0.0951705983889528193325661680092059838895,-0.0143679516685845799769850402761795166858,0.1219692174913693901035673187571921749137,
                   -0.0248673762945914844649021864211737629459,-0.0143679516685845799769850402761795166858,-0.0271686804756425009589566551591868047564,0.020788933640199462984273110855389336402,0.032108457997698504027617951668584579977,
                   0.1219692174913693901035673187571921749137,0.020788933640199462984273110855389336402,-0.0894091868047564250095895665515918680476,-0.0647097238204833141542002301495972382048,-0.0248673762945914844649021864211737629459,
                   0.032108457997698504027617951668584579977,-0.0647097238204833141542002301495972382048,0.0699053509781357882623705408515535097814,-0.0143679516685845799769850402761795166858,-0.0271686804756425009589566551591868047564,
                   0.020788933640199462984273110855389336402,0.032108457997698504027617951668584579977,-0.0271686804756425009589566551591868047564,-0.0259042002301495972382048331415420023015,0.0096953394706559263521288837744533947066,
                   0.0523332374376678174146528576908323743767,0.020788933640199462984273110855389336402,0.0096953394706559263521288837744533947066,-0.0184864787111622554660529344073647871116,-0.02095349060222477943996931338703490602225,
                   0.032108457997698504027617951668584579977,0.0523332374376678174146528576908323743767,-0.02095349060222477943996931338703490602225,-0.0748489643268124280782508630609896432681,0.1219692174913693901035673187571921749137,
                   0.020788933640199462984273110855389336402,-0.0894091868047564250095895665515918680476,-0.0647097238204833141542002301495972382048,0.020788933640199462984273110855389336402,0.0096953394706559263521288837744533947066,
                   -0.0184864787111622554660529344073647871116,-0.02095349060222477943996931338703490602225,-0.0894091868047564250095895665515918680476,-0.0184864787111622554660529344073647871116,0.0522776179516685845799769850402761795167,
                   0.0645737437667817414652857690832374376678,-0.0647097238204833141542002301495972382048,-0.02095349060222477943996931338703490602225,0.0645737437667817414652857690832374376678,0.032450230149597238204833141542002301496,
                   -0.0248673762945914844649021864211737629459,0.032108457997698504027617951668584579977,-0.0647097238204833141542002301495972382048,0.0699053509781357882623705408515535097814,0.032108457997698504027617951668584579977,
                   0.0523332374376678174146528576908323743767,-0.02095349060222477943996931338703490602225,-0.0748489643268124280782508630609896432681,-0.0647097238204833141542002301495972382048,-0.02095349060222477943996931338703490602225,
                   0.0645737437667817414652857690832374376678,0.032450230149597238204833141542002301496,0.0699053509781357882623705408515535097814,-0.0748489643268124280782508630609896432681,0.032450230149597238204833141542002301496,
                   -0.0399433256616800920598388952819332566168,0.0019485999232834675872650556194859992328,0.00324127349443805140007671653241273494438,0.00324127349443805140007671653241273494438,-0.0084311469121595703874184886843114691216,
                   -0.0190372075182201764480245492903720751822,0.01384733410049865746068277713847334100499,0.01384733410049865746068277713847334100499,-0.00865746068277713847334100498657460682777,0.0019485999232834675872650556194859992328,
                   0.00324127349443805140007671653241273494438,0.00324127349443805140007671653241273494438,-0.0084311469121595703874184886843114691216,-0.0190372075182201764480245492903720751822,0.01384733410049865746068277713847334100499,
                   0.01384733410049865746068277713847334100499,-0.00865746068277713847334100498657460682777,0.0019485999232834675872650556194859992328,0.00324127349443805140007671653241273494438,0.00324127349443805140007671653241273494438,
                   -0.0084311469121595703874184886843114691216,-0.0190372075182201764480245492903720751822,0.01384733410049865746068277713847334100499,0.01384733410049865746068277713847334100499,-0.00865746068277713847334100498657460682777,
                   -0.0536828370732642884541618718833908707326,0.091087143093076637380434848789279169026,-0.0596823435150175656504770428821061732454,0.0160596830648254698887610280015343306483,-0.050393434229272836867773576634336128007,
                   0.1247176399696178177190835418683519949343,-0.0221783467206726700397586473535840624448,-0.0464654792728368677735766343361280070141,-0.0509844666146090196723108115513178804318,0.0927614094912354406025292101241468330076,
                   -0.0213423694068472549485207713055814321637,-0.0261149532166146090196723108115513178804,0.0080406957230533179900268507863444572305,0.0241371595323335829664943588994221905614,-0.0628336257770593213631188314732618530087,
                   0.0368741249520521672420406597621787495205,-0.0003677954575897613872297416601214069569,0.0011541437202038467861252671379253657735,0.0628434828620746342265329607101759000493,-0.0579494513778532765874538026436760613976,
                   -0.0198681756154066280648559129571787799636,0.0122761195818948983505945531261986958189,0.0790986167050249328730341388569236670502,-0.0878620459035807137072959857769984352263,-0.0354537008496660395394572609762483180205,
                   0.0067441863252780974299961641733793632528,0.0883394107211354046797084771768316072114,-0.0849410776313466186883908402895744667897,-0.0259588278429259441917669765771031593816,-0.0058020867581785303304290646062797961532,
                   0.0886794601759000493177708367581785303304,-0.0625989253216310178335494791190993722639,0.0816177954575897613872297416601214069568,-0.1449041437202038467861252671379253657735,-0.0190934828620746342265329607101759000493,
                   0.0766994513778532765874538026436760613976,0.0886181756154066280648559129571787799636,-0.1435261195818948983505945531261986958189,-0.0228486167050249328730341388569236670502,0.0941120459035807137072959857769984352263,
                   0.0917037008496660395394572609762483180205,-0.1254941863252780974299961641733793632528,-0.0195894107211354046797084771768316072114,0.0786910776313466186883908402895744667897,0.0697088278429259441917669765771031593816,
                   -0.1004479132418214696695709353937202038468,-0.0074294601759000493177708367581785303304,0.0438489253216310178335494791190993722639,-0.0275671629267357115458381281166091292674,0.052662856906923362619565151210720830974,
                   0.0159323435150175656504770428821061732454,-0.0348096830648254698887610280015343306483,-0.018356565770727163132226423365663871993,0.0065323600303821822809164581316480050657,-0.0340716532793273299602413526464159375552,
                   0.0402154792728368677735766343361280070141,-0.0052655333853909803276891884486821195682,0.0259885905087645593974707898758531669924,-0.0474076305931527450514792286944185678363,0.0323649532166146090196723108115513178804,
                   -0.0517906957230533179900268507863444572305,0.0821128404676664170335056411005778094386,-0.0184163742229406786368811685267381469913,-0.0181241249520521672420406597621787495205,-0.0004750793041932282438611552535603168515,
                   0.0030700160130539877375320413295096839401,-0.0111868991299371046206489244463928008232,0.0085919624210763451269780383704434337346,0.0004750793041932282438611552535603168515,-0.0030700160130539877375320413295096839401,
                   0.0111868991299371046206489244463928008232,-0.0085919624210763451269780383704434337346,0.0093116061154035837580141377609732040112,-0.0067166694065428242643432516850238369226,-0.0093116061154035837580141377609732040112,
                   0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,0.0014002137103402926187736314318592799605,0.0011947229985204668748972546440900871281,
                   -0.0366333238959821238302250960478808580074,-0.0004916175011111719972479466150352226302,0.0366333238959821238302250960478808580074,0.0004916175011111719972479466150352226302,0.0489223534318471027331786825457711533661,
                   0.0298692546319128597609610267838115939382,-0.0489223534318471027331786825457711533661,-0.0298692546319128597609610267838115939382,-0.0133279991369390103567318757192174913694,0.0189159036238413875544677232440945521114,
                   -0.0132988150162464508456069637504236660355,0.0073153409090909090909090909090909090909,-0.0032371462388603738814709278844299941346,0.0205830072753859673690897319589302711666,0.0333336203226762298492256298163471159252,
                   -0.0300561417748917748917748917748917748918,0.031882979572193707214804261217763327468,-0.0648538406087193007024230652922636044999,-0.0307294536560095631825589631496804492585,0.0430769751082251082251082251082251082251,
                   -0.0153178341963943229766014576141158419639,0.0253549297094919457788656100892387812219,0.0106946483495797841789402970837569993688,-0.0203361742424242424242424242424242424242,0.0189159036238413875544677232440945521114,
                   0.0432578323038502363396878164810654262131,-0.0080526244588744588744588744588744588745,0.0019960552759867105858667040101639257757,0.0205830072753859673690897319589302711666,0.068942690352896049098580744150364403529,
                   0.0781960227272727272727272727272727272727,0.0106801783786224714494756688849515853735,-0.0648538406087193007024230652922636044999,-0.1340468570195627157652474108170310701956,-0.1068418560606060606060606060606060606061,
                   -0.0393260117119558047828090022182849187069,0.0253549297094919457788656100892387812219,0.0843463343628164303269788501856012404536,0.0575317911255411255411255411255411255411,0.0266497780573466227474666293231694075576,
                   -0.0132988150162464508456069637504236660355,-0.0080526244588744588744588744588744588745,-0.0531942500527679852574367342299831751309,-0.0232381438835816472947274635038348118517,0.0333336203226762298492256298163471159252,
                   0.0781960227272727272727272727272727272727,0.0923735975258918296892980437284234752589,0.0343615273566486646655423026731043608681,-0.0307294536560095631825589631496804492585,-0.1068418560606060606060606060606060606061,
                   -0.0845610975258918296892980437284234752589,-0.0161323606899819979988756360064376942014,0.0106946483495797841789402970837569993688,0.0575317911255411255411255411255411255411,0.0245484167194346519241034008966498417975,
                   0.005008977216914980628060796837168145185,0.0073153409090909090909090909090909090909,0.0019960552759867105858667040101639257757,-0.0232381438835816472947274635038348118517,0.0143223173187571921749136939010356731876,
                   -0.0300561417748917748917748917748917748918,0.0106801783786224714494756688849515853735,0.0343615273566486646655423026731043608681,-0.0356089035446894096683126218991197894151,0.0430769751082251082251082251082251082251,
                   -0.0393260117119558047828090022182849187069,-0.0161323606899819979988756360064376942014,0.0330047368780227430016459552324531227485,-0.0203361742424242424242424242424242424242,0.0266497780573466227474666293231694075576,
                   0.005008977216914980628060796837168145185,-0.0117181506520905255082470272343690065209,0.0007277161813026791929745516243406538765,-0.0046150423260127901478112448576583597681,-0.0007277161813026791929745516243406538765,
                   0.0046150423260127901478112448576583597681,0.0019094146203850845201056171520306541404,0.0019779115243250264347310760812870517512,-0.0019094146203850845201056171520306541404,-0.00197791152432502643473107608128705175119,
                   -0.0551689920097093092873683590983169042241,-0.0058170995670995670995670995670995670996,-0.0058170995670995670995670995670995670996,0.0251365244772417768198358915658493717565,-0.0073310079902906907126316409016830957759,
                   -0.01501623376623376623376623376623376623377,-0.01501623376623376623376623376623376623377,-0.0043031911439084434865025582325160384232,0.0007277161813026791929745516243406538765,-0.0007277161813026791929745516243406538765,
                   -0.0046150423260127901478112448576583597681,0.0046150423260127901478112448576583597681,0.0019094146203850845201056171520306541404,-0.0019094146203850845201056171520306541404,0.0019779115243250264347310760812870517512,
                   -0.00197791152432502643473107608128705175119,-0.0536828370732642884541618718833908707326,-0.0003677954575897613872297416601214069568,0.0816177954575897613872297416601214069568,-0.0275671629267357115458381281166091292674,
                   0.091087143093076637380434848789279169026,0.0011541437202038467861252671379253657735,-0.1449041437202038467861252671379253657735,0.052662856906923362619565151210720830974,-0.0596823435150175656504770428821061732454,
                   0.0628434828620746342265329607101759000493,-0.0190934828620746342265329607101759000493,0.0159323435150175656504770428821061732454,0.0160596830648254698887610280015343306483,-0.0579494513778532765874538026436760613976,
                   0.0766994513778532765874538026436760613976,-0.0348096830648254698887610280015343306483,-0.050393434229272836867773576634336128007,-0.0198681756154066280648559129571787799636,0.0886181756154066280648559129571787799636,
                   -0.018356565770727163132226423365663871993,0.1247176399696178177190835418683519949343,0.0122761195818948983505945531261986958189,-0.1435261195818948983505945531261986958189,0.0065323600303821822809164581316480050657,
                   -0.0221783467206726700397586473535840624448,0.0790986167050249328730341388569236670502,-0.0228486167050249328730341388569236670502,-0.0340716532793273299602413526464159375552,-0.0464654792728368677735766343361280070141,
                   -0.0878620459035807137072959857769984352263,0.0941120459035807137072959857769984352263,0.0402154792728368677735766343361280070141,-0.0509844666146090196723108115513178804318,-0.0354537008496660395394572609762483180205,
                   0.0917037008496660395394572609762483180205,-0.0052655333853909803276891884486821195682,0.0927614094912354406025292101241468330076,0.0067441863252780974299961641733793632528,-0.1254941863252780974299961641733793632528,
                   0.0259885905087645593974707898758531669924,-0.0213423694068472549485207713055814321637,0.0883394107211354046797084771768316072114,-0.0195894107211354046797084771768316072114,-0.0474076305931527450514792286944185678363,
                   -0.0261149532166146090196723108115513178804,-0.0849410776313466186883908402895744667897,0.0786910776313466186883908402895744667897,0.0323649532166146090196723108115513178804,0.0080406957230533179900268507863444572305,
                   -0.0259588278429259441917669765771031593816,0.0697088278429259441917669765771031593816,-0.0517906957230533179900268507863444572305,0.0241371595323335829664943588994221905614,-0.0058020867581785303304290646062797961532,
                   -0.1004479132418214696695709353937202038468,0.0821128404676664170335056411005778094386,-0.0628336257770593213631188314732618530087,0.0886794601759000493177708367581785303304,-0.0074294601759000493177708367581785303304,
                   -0.0184163742229406786368811685267381469913,0.0368741249520521672420406597621787495205,-0.0625989253216310178335494791190993722639,0.0438489253216310178335494791190993722639,-0.0181241249520521672420406597621787495205,
                   0.0093116061154035837580141377609732040112,-0.0093116061154035837580141377609732040112,-0.0067166694065428242643432516850238369226,0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,
                   0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,0.0011947229985204668748972546440900871281,-0.0366333238959821238302250960478808580074,0.0366333238959821238302250960478808580074,
                   -0.0004916175011111719972479466150352226302,0.0004916175011111719972479466150352226302,0.0489223534318471027331786825457711533661,-0.0489223534318471027331786825457711533661,0.0298692546319128597609610267838115939382,
                   -0.0298692546319128597609610267838115939382,-0.0004750793041932282438611552535603168515,0.0030700160130539877375320413295096839401,-0.0111868991299371046206489244463928008232,0.0085919624210763451269780383704434337346,
                   0.0004750793041932282438611552535603168515,-0.0030700160130539877375320413295096839401,0.0111868991299371046206489244463928008232,-0.0085919624210763451269780383704434337346,-0.0536828370732642884541618718833908707326,
                   0.091087143093076637380434848789279169026,-0.0596823435150175656504770428821061732454,0.0160596830648254698887610280015343306483,-0.0003677954575897613872297416601214069568,0.0011541437202038467861252671379253657735,
                   0.0628434828620746342265329607101759000493,-0.0579494513778532765874538026436760613976,0.0816177954575897613872297416601214069569,-0.1449041437202038467861252671379253657735,-0.0190934828620746342265329607101759000493,
                   0.0766994513778532765874538026436760613976,-0.0275671629267357115458381281166091292674,0.052662856906923362619565151210720830974,0.0159323435150175656504770428821061732454,-0.0348096830648254698887610280015343306483,
                   -0.050393434229272836867773576634336128007,0.1247176399696178177190835418683519949343,-0.0221783467206726700397586473535840624448,-0.0464654792728368677735766343361280070141,-0.0198681756154066280648559129571787799636,
                   0.0122761195818948983505945531261986958189,0.0790986167050249328730341388569236670502,-0.0878620459035807137072959857769984352263,0.0886181756154066280648559129571787799636,-0.1435261195818948983505945531261986958189,
                   -0.0228486167050249328730341388569236670502,0.0941120459035807137072959857769984352263,-0.018356565770727163132226423365663871993,0.0065323600303821822809164581316480050657,-0.0340716532793273299602413526464159375552,
                   0.0402154792728368677735766343361280070141,-0.0509844666146090196723108115513178804318,0.0927614094912354406025292101241468330076,-0.0213423694068472549485207713055814321637,-0.0261149532166146090196723108115513178804,
                   -0.0354537008496660395394572609762483180205,0.0067441863252780974299961641733793632528,0.0883394107211354046797084771768316072114,-0.0849410776313466186883908402895744667897,0.0917037008496660395394572609762483180205,
                   -0.1254941863252780974299961641733793632528,-0.0195894107211354046797084771768316072114,0.0786910776313466186883908402895744667897,-0.0052655333853909803276891884486821195682,0.0259885905087645593974707898758531669924,
                   -0.0474076305931527450514792286944185678363,0.0323649532166146090196723108115513178804,0.0080406957230533179900268507863444572305,0.0241371595323335829664943588994221905614,-0.0628336257770593213631188314732618530087,
                   0.0368741249520521672420406597621787495205,-0.0259588278429259441917669765771031593816,-0.0058020867581785303304290646062797961532,0.0886794601759000493177708367581785303304,-0.0625989253216310178335494791190993722639,
                   0.0697088278429259441917669765771031593816,-0.1004479132418214696695709353937202038468,-0.0074294601759000493177708367581785303304,0.0438489253216310178335494791190993722639,-0.0517906957230533179900268507863444572305,
                   0.0821128404676664170335056411005778094386,-0.0184163742229406786368811685267381469913,-0.0181241249520521672420406597621787495205,0.0093116061154035837580141377609732040112,-0.0067166694065428242643432516850238369226,
                   -0.0093116061154035837580141377609732040112,0.0067166694065428242643432516850238369226,-0.0014002137103402926187736314318592799605,-0.0011947229985204668748972546440900871281,0.0014002137103402926187736314318592799605,
                   0.0011947229985204668748972546440900871281,-0.0004750793041932282438611552535603168515,0.0030700160130539877375320413295096839401,-0.0111868991299371046206489244463928008232,0.0085919624210763451269780383704434337346,
                   0.0004750793041932282438611552535603168515,-0.0030700160130539877375320413295096839401,0.0111868991299371046206489244463928008232,-0.0085919624210763451269780383704434337346,-0.0366333238959821238302250960478808580074,
                   0.0366333238959821238302250960478808580074,-0.0004916175011111719972479466150352226302,0.0004916175011111719972479466150352226302,0.0489223534318471027331786825457711533661,-0.0489223534318471027331786825457711533661,
                   0.0298692546319128597609610267838115939382,-0.0298692546319128597609610267838115939382,-0.0133279991369390103567318757192174913694,-0.0032371462388603738814709278844299941346,0.031882979572193707214804261217763327468,
                   -0.0153178341963943229766014576141158419639,0.0189159036238413875544677232440945521114,0.0205830072753859673690897319589302711666,-0.0648538406087193007024230652922636044999,0.0253549297094919457788656100892387812219,
                   -0.0132988150162464508456069637504236660355,0.0333336203226762298492256298163471159252,-0.0307294536560095631825589631496804492585,0.0106946483495797841789402970837569993688,0.0073153409090909090909090909090909090909,
                   -0.0300561417748917748917748917748917748918,0.0430769751082251082251082251082251082251,-0.0203361742424242424242424242424242424242,0.0189159036238413875544677232440945521114,0.0205830072753859673690897319589302711666,
                   -0.0648538406087193007024230652922636044999,0.0253549297094919457788656100892387812219,0.0432578323038502363396878164810654262131,0.068942690352896049098580744150364403529,-0.1340468570195627157652474108170310701956,
                   0.0843463343628164303269788501856012404536,-0.0080526244588744588744588744588744588745,0.0781960227272727272727272727272727272727,-0.1068418560606060606060606060606060606061,0.0575317911255411255411255411255411255411,
                   0.0019960552759867105858667040101639257757,0.0106801783786224714494756688849515853735,-0.0393260117119558047828090022182849187069,0.0266497780573466227474666293231694075576,-0.0132988150162464508456069637504236660355,
                   0.0333336203226762298492256298163471159252,-0.0307294536560095631825589631496804492585,0.0106946483495797841789402970837569993688,-0.0080526244588744588744588744588744588745,0.0781960227272727272727272727272727272727,
                   -0.1068418560606060606060606060606060606061,0.0575317911255411255411255411255411255411,-0.0531942500527679852574367342299831751309,0.0923735975258918296892980437284234752589,-0.0845610975258918296892980437284234752589,
                   0.0245484167194346519241034008966498417975,-0.0232381438835816472947274635038348118517,0.0343615273566486646655423026731043608681,-0.0161323606899819979988756360064376942014,0.005008977216914980628060796837168145185,
                   0.0073153409090909090909090909090909090909,-0.0300561417748917748917748917748917748918,0.0430769751082251082251082251082251082251,-0.0203361742424242424242424242424242424242,0.0019960552759867105858667040101639257757,
                   0.0106801783786224714494756688849515853735,-0.0393260117119558047828090022182849187069,0.0266497780573466227474666293231694075576,-0.0232381438835816472947274635038348118517,0.0343615273566486646655423026731043608681,
                   -0.0161323606899819979988756360064376942014,0.005008977216914980628060796837168145185,0.0143223173187571921749136939010356731876,-0.0356089035446894096683126218991197894151,0.0330047368780227430016459552324531227485,
                   -0.0117181506520905255082470272343690065209,0.0007277161813026791929745516243406538765,-0.0007277161813026791929745516243406538765,-0.0046150423260127901478112448576583597681,0.0046150423260127901478112448576583597681,
                   0.0019094146203850845201056171520306541404,-0.0019094146203850845201056171520306541404,0.0019779115243250264347310760812870517512,-0.00197791152432502643473107608128705175119,0.0007277161813026791929745516243406538765,
                   -0.0007277161813026791929745516243406538765,-0.0046150423260127901478112448576583597681,0.0046150423260127901478112448576583597681,0.0019094146203850845201056171520306541404,-0.0019094146203850845201056171520306541404,
                   0.0019779115243250264347310760812870517512,-0.00197791152432502643473107608128705175119,-0.0551689920097093092873683590983169042241,-0.0058170995670995670995670995670995670996,-0.0058170995670995670995670995670995670996,
                   0.0251365244772417768198358915658493717565,-0.0073310079902906907126316409016830957759,-0.01501623376623376623376623376623376623377,-0.0150162337662337662337662337662337662338,-0.0043031911439084434865025582325160384232,0.01171875,-0.015625,-0.00390625,
                   0.0078125,-0.015625,-0.0533854166666666666666666666666666666667,-0.0338541666666666666666666666666666666667,-0.01171875,-0.00390625,-0.0338541666666666666666666666666666666667,-0.0143229166666666666666666666666666666667,0,0.0078125,-0.01171875,0,
                   0.00390625,-0.0208333333333333333333333333333333333333,0.0247395833333333333333333333333333333333,0.0130208333333333333333333333333333333333,-0.0169270833333333333333333333333333333333,0.0247395833333333333333333333333333333333,0.0546875,
                   0.0455729166666666666666666666666666666667,0.0208333333333333333333333333333333333333,0.0130208333333333333333333333333333333333,0.0455729166666666666666666666666666666667,0.0364583333333333333333333333333333333333,
                   0.0091145833333333333333333333333333333333,-0.0169270833333333333333333333333333333333,0.0208333333333333333333333333333333333333,0.00911458333333333333333333333333333333333,-0.0130208333333333333333333333333333333333,
                   0.0065104166666666666666666666666666666667,-0.0026041666666666666666666666666666666667,-0.0143229166666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0026041666666666666666666666666666666667,
                   -0.0221354166666666666666666666666666666667,-0.03125,-0.0065104166666666666666666666666666666667,-0.0143229166666666666666666666666666666667,-0.03125,-0.0403645833333333333333333333333333333333,-0.0182291666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.0065104166666666666666666666666666666667,-0.0182291666666666666666666666666666666667,0.0143229166666666666666666666666666666667,0.0026041666666666666666666666666666666667,
                   -0.0065104166666666666666666666666666666667,0.0052083333333333333333333333333333333333,-0.00130208333333333333333333333333333333333,-0.0065104166666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.00911458333333333333333333333333333333333,-0.0026041666666666666666666666666666666667,0.0052083333333333333333333333333333333333,0.0091145833333333333333333333333333333333,0.0286458333333333333333333333333333333333,
                   0.0091145833333333333333333333333333333333,-0.00130208333333333333333333333333333333333,-0.00260416666666666666666666666666666666667,0.0091145833333333333333333333333333333333,-0.00520833333333333333333333333333333333333,
                   0.0208333333333333333333333333333333333333,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0.01041666666666666666666666666666666666667,0,0,-0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   0,0,0,0.031060006712696586114307633294975067127,-0.006503181681188010301934352567263959669,0.0004773377993314702175461669132555208505,0.0174236430763329497506712696586114307633,-0.0426313853067545050667413536611848848136,
                   -0.0617271234293913618808133576066065517542,-0.0555582922605601930496445264377753829231,-0.0356508658262350245472608341806654042941,0.0217980519734211717334080203278515514802,0.0408937900960580285474800242732732184209,
                   0.0347249589272268597163111931044420495897,0.0148175324929016912139275008473320709608,-0.0102266733793632527809742999616417337936,0.0273365150145213436352676859005972930024,0.0203559955340018631157871664200778124829,
                   0.00340969025700038358266206367472190257,-0.0580689955360314010103039638904617807571,-0.041152828896763495919614063073978685793,-0.0302220929660275651836833271432427550571,-0.0368027617697976347765377301242280145234,
                   0.0440006130472902624801358978574168447586,0.0405033323743766781741465285769083237438,0.0454275747986191024165707710011507479862,0.0549313489780261932160666337881527754945,0.0184993869527097375198641021425831552414,
                   0.0219966676256233218258534714230916762562,0.0170724252013808975834292289988492520138,0.0075686510219738067839333662118472245055,-0.0044310044639685989896960361095382192429,-0.021347171103236504080385936926021314207,
                   -0.0322779070339724348163166728567572449429,-0.0256972382302023652234622698757719854766,0.0424439955360314010103039638904617807571,0.056777828896763495919614063073978685793,0.0458470929660275651836833271432427550571,
                   0.0211777617697976347765377301242280145234,0.0028743869527097375198641021425831552414,-0.0040449990410433448408131952435749904104,-0.0089692414652857690832374376678174146529,-0.0080563489780261932160666337881527754945,
                   -0.0653743869527097375198641021425831552414,-0.0584550009589566551591868047564250095896,-0.0535307585347142309167625623321825853471,-0.0544436510219738067839333662118472245055,0.0200560044639685989896960361095382192429,
                   0.005722171103236504080385936926021314207,0.0166529070339724348163166728567572449429,0.0413222382302023652234622698757719854766,-0.015435006712696586114307633294975067127,-0.009121818318811989698065647432736040331,
                   -0.0161023377993314702175461669132555208505,-0.00179864307633294975067126965861143076333,-0.0042436146932454949332586463388151151864,-0.0163978765706086381191866423933934482458,-0.0225667077394398069503554735622246170769,
                   -0.0112241341737649754527391658193345957059,0.0250769480265788282665919796721484485198,0.0372312099039419714525199757267267815791,0.0434000410727731402836888068955579504103,0.0320574675070983087860724991526679290392,
                   -0.00539832662063674721902570003835826620637,-0.0117115150145213436352676859005972930024,-0.0047309955340018631157871664200778124829,-0.01903469025700038358266206367472190257,0.0493518924426097421878012595312173371245,
                   0.0309536240443413439194029911329489388561,-0.0493518924426097421878012595312173371245,-0.0309536240443413439194029911329489388561,-0.0076852257759430755211345928645506704578,0.01071304262232532274726367553371772781055,
                   0.00768522577594307552113459286455067045784,-0.01071304262232532274726367553371772781055,0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,-0.0011816984390824053271310655276900002638,
                   -0.0065929538503378165825423209389454115193,-0.0011816984390824053271310655276900002638,-0.00659295385033781658254232093894541151925,0.00118169843908240532713106552769000026384,0.00659295385033781658254232093894541151925,
                   -0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,
                   -0.00006849690393994191462545892925639761083,-0.00006849690393994191462545892925639761083,0.00006849690393994191462545892925639761083,0.031060006712696586114307633294975067127,-0.0426313853067545050667413536611848848136,
                   0.0217980519734211717334080203278515514802,-0.0102266733793632527809742999616417337936,-0.006503181681188010301934352567263959669,-0.0617271234293913618808133576066065517542,0.0408937900960580285474800242732732184209,
                   0.0273365150145213436352676859005972930024,0.0004773377993314702175461669132555208505,-0.0555582922605601930496445264377753829231,0.0347249589272268597163111931044420495897,0.0203559955340018631157871664200778124829,
                   0.0174236430763329497506712696586114307633,-0.0356508658262350245472608341806654042941,0.0148175324929016912139275008473320709608,0.00340969025700038358266206367472190257,-0.0580689955360314010103039638904617807571,
                   0.0440006130472902624801358978574168447586,0.0184993869527097375198641021425831552414,-0.0044310044639685989896960361095382192429,-0.041152828896763495919614063073978685793,0.0405033323743766781741465285769083237438,
                   0.0219966676256233218258534714230916762562,-0.021347171103236504080385936926021314207,-0.0302220929660275651836833271432427550571,0.0454275747986191024165707710011507479862,0.0170724252013808975834292289988492520138,
                   -0.0322779070339724348163166728567572449429,-0.0368027617697976347765377301242280145234,0.0549313489780261932160666337881527754945,0.0075686510219738067839333662118472245055,-0.0256972382302023652234622698757719854766,
                   0.0424439955360314010103039638904617807571,0.0028743869527097375198641021425831552414,-0.0653743869527097375198641021425831552414,0.0200560044639685989896960361095382192429,0.056777828896763495919614063073978685793,
                   -0.0040449990410433448408131952435749904104,-0.0584550009589566551591868047564250095896,0.005722171103236504080385936926021314207,0.0458470929660275651836833271432427550571,-0.0089692414652857690832374376678174146529,
                   -0.0535307585347142309167625623321825853471,0.0166529070339724348163166728567572449429,0.0211777617697976347765377301242280145234,-0.0080563489780261932160666337881527754945,-0.0544436510219738067839333662118472245055,
                   0.0413222382302023652234622698757719854766,-0.015435006712696586114307633294975067127,-0.0042436146932454949332586463388151151864,0.0250769480265788282665919796721484485198,-0.00539832662063674721902570003835826620637,
                   -0.009121818318811989698065647432736040331,-0.0163978765706086381191866423933934482458,0.0372312099039419714525199757267267815791,-0.0117115150145213436352676859005972930024,-0.0161023377993314702175461669132555208505,
                   -0.0225667077394398069503554735622246170769,0.0434000410727731402836888068955579504103,-0.0047309955340018631157871664200778124829,-0.00179864307633294975067126965861143076333,-0.0112241341737649754527391658193345957059,
                   0.0320574675070983087860724991526679290392,-0.01903469025700038358266206367472190257,0.0493518924426097421878012595312173371245,-0.0493518924426097421878012595312173371245,0.0309536240443413439194029911329489388561,
                   -0.0309536240443413439194029911329489388561,-0.0076852257759430755211345928645506704578,0.00768522577594307552113459286455067045784,0.01071304262232532274726367553371772781055,-0.01071304262232532274726367553371772781055,
                   -0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,
                   -0.00006849690393994191462545892925639761083,-0.00006849690393994191462545892925639761083,0.00006849690393994191462545892925639761083,0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,
                   -0.0011816984390824053271310655276900002638,-0.0065929538503378165825423209389454115193,-0.0011816984390824053271310655276900002638,-0.00659295385033781658254232093894541151925,0.00118169843908240532713106552769000026384,
                   0.00659295385033781658254232093894541151925,0.05625,0.05,0.04375,0.0375,-0.0875,-0.08125,-0.075,-0.06875,0.00625,0.0125,0.01875,0.025,0.025,0.01875,0.0125,0.00625,-0.0875,-0.08125,-0.075,-0.06875,0.11875,0.1125,0.10625,0.1,0.025,0.01875,0.0125,0.00625,
                   -0.05625,-0.05,-0.04375,-0.0375,0.00625,0.0125,0.01875,0.025,0.025,0.01875,0.0125,0.00625,-0.06875,-0.075,-0.08125,-0.0875,0.0375,0.04375,0.05,0.05625,0.025,0.01875,0.0125,0.00625,-0.05625,-0.05,-0.04375,-0.0375,0.0375,0.04375,0.05,0.05625,-0.00625,
                   -0.0125,-0.01875,-0.025,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.0804735327963176064441887226697353279632,-0.0193410323853361828045372349169817524248,-0.0181589676146638171954627650830182475752,-0.0429735327963176064441887226697353279632,
                   -0.0193410323853361828045372349169817524248,-0.0218355252342594114746013480190695380569,0.0093355252342594114746013480190695380569,0.0318410323853361828045372349169817524248,-0.0181589676146638171954627650830182475752,
                   0.0093355252342594114746013480190695380569,0.0031644747657405885253986519809304619431,0.0056589676146638171954627650830182475752,-0.0429735327963176064441887226697353279632,0.0318410323853361828045372349169817524248,
                   0.0056589676146638171954627650830182475752,0.0054735327963176064441887226697353279632,-0.0856999835607430544139404898898569784646,-0.0132062304783823771165543317442051619267,0.0507062304783823771165543317442051619267,
                   0.0481999835607430544139404898898569784646,-0.0132062304783823771165543317442051619267,0.0007180667433831990794016110471806674338,0.0117819332566168009205983889528193325662,0.0007062304783823771165543317442051619267,
                   0.0507062304783823771165543317442051619267,0.0117819332566168009205983889528193325662,-0.0242819332566168009205983889528193325662,-0.0382062304783823771165543317442051619267,0.0481999835607430544139404898898569784646,
                   0.0007062304783823771165543317442051619267,-0.0382062304783823771165543317442051619267,-0.0106999835607430544139404898898569784646,-0.0219012822620417557126417885911556797633,0.0195859773138254150912378760480026302811,
                   0.0179140226861745849087621239519973697189,-0.0155987177379582442873582114088443202367,0.0195859773138254150912378760480026302811,0.0154907940161104718066743383199079401611,-0.0029907940161104718066743383199079401611,
                   -0.0320859773138254150912378760480026302811,0.0179140226861745849087621239519973697189,-0.0029907940161104718066743383199079401611,-0.0095092059838895281933256616800920598389,-0.0054140226861745849087621239519973697189,
                   -0.0155987177379582442873582114088443202367,-0.0320859773138254150912378760480026302811,-0.0054140226861745849087621239519973697189,0.0530987177379582442873582114088443202367,0.0395644418872266973532796317606444188723,
                   0.0016005260562222587539043235245766891337,-0.0391005260562222587539043235245766891337,-0.0020644418872266973532796317606444188723,0.0016005260562222587539043235245766891337,-0.0033290317277659049810948545125760315634,
                   -0.0091709682722340950189051454874239684366,0.0108994739437777412460956764754233108663,-0.0391005260562222587539043235245766891337,-0.0091709682722340950189051454874239684366,0.0216709682722340950189051454874239684366,
                   0.0266005260562222587539043235245766891337,-0.0020644418872266973532796317606444188723,0.0108994739437777412460956764754233108663,0.0266005260562222587539043235245766891337,-0.0354355581127733026467203682393555811277,
                   0.0361417063948709518329771494328456353773,-0.0361417063948709518329771494328456353773,-0.0361417063948709518329771494328456353773,0.0361417063948709518329771494328456353773,-0.0190530987999342429722176557619595594279,
                   0.0190530987999342429722176557619595594279,0.01905309879993424297221765576195955942791,-0.0190530987999342429722176557619595594279,-0.0107118198257438763767877691928324839717,0.0107118198257438763767877691928324839717,
                   0.0055219464080223573894459970409337497945,-0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,-0.0107118198257438763767877691928324839717,-0.00552194640802235738944599704093374979451,
                   0.0055219464080223573894459970409337497945,-0.0107118198257438763767877691928324839717,0.0107118198257438763767877691928324839717,0.0055219464080223573894459970409337497945,-0.0055219464080223573894459970409337497945,
                   0.0107118198257438763767877691928324839717,-0.01071181982574387637678776919283248397172,-0.00552194640802235738944599704093374979451,0.0055219464080223573894459970409337497945,0.05625,-0.0875,0.00625,0.025,0.05,-0.08125,0.0125,0.01875,0.04375,-0.075,
                   0.01875,0.0125,0.0375,-0.06875,0.025,0.00625,-0.0875,0.11875,0.025,-0.05625,-0.08125,0.1125,0.01875,-0.05,-0.075,0.10625,0.0125,-0.04375,-0.06875,0.1,0.00625,-0.0375,0.00625,0.025,-0.06875,0.0375,0.0125,0.01875,-0.075,0.04375,0.01875,0.0125,-0.08125,
                   0.05,0.025,0.00625,-0.0875,0.05625,0.025,-0.05625,0.0375,-0.00625,0.01875,-0.05,0.04375,-0.0125,0.0125,-0.04375,0.05,-0.01875,0.00625,-0.0375,0.05625,-0.025,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.031060006712696586114307633294975067127,
                   -0.006503181681188010301934352567263959669,0.0004773377993314702175461669132555208505,0.0174236430763329497506712696586114307633,-0.0580689955360314010103039638904617807571,-0.041152828896763495919614063073978685793,
                   -0.0302220929660275651836833271432427550571,-0.0368027617697976347765377301242280145234,0.0424439955360314010103039638904617807571,0.056777828896763495919614063073978685793,0.0458470929660275651836833271432427550571,
                   0.0211777617697976347765377301242280145234,-0.015435006712696586114307633294975067127,-0.009121818318811989698065647432736040331,-0.0161023377993314702175461669132555208505,-0.00179864307633294975067126965861143076333,
                   -0.0426313853067545050667413536611848848136,-0.0617271234293913618808133576066065517542,-0.0555582922605601930496445264377753829231,-0.0356508658262350245472608341806654042941,0.0440006130472902624801358978574168447586,
                   0.0405033323743766781741465285769083237438,0.0454275747986191024165707710011507479862,0.0549313489780261932160666337881527754945,0.0028743869527097375198641021425831552414,-0.0040449990410433448408131952435749904104,
                   -0.0089692414652857690832374376678174146529,-0.0080563489780261932160666337881527754945,-0.0042436146932454949332586463388151151864,-0.0163978765706086381191866423933934482458,-0.0225667077394398069503554735622246170769,
                   -0.0112241341737649754527391658193345957059,0.0217980519734211717334080203278515514802,0.0408937900960580285474800242732732184209,0.0347249589272268597163111931044420495897,0.0148175324929016912139275008473320709608,
                   0.0184993869527097375198641021425831552414,0.0219966676256233218258534714230916762562,0.0170724252013808975834292289988492520138,0.0075686510219738067839333662118472245055,-0.0653743869527097375198641021425831552414,
                   -0.0584550009589566551591868047564250095896,-0.0535307585347142309167625623321825853471,-0.0544436510219738067839333662118472245055,0.0250769480265788282665919796721484485198,0.0372312099039419714525199757267267815791,
                   0.0434000410727731402836888068955579504103,0.0320574675070983087860724991526679290392,-0.0102266733793632527809742999616417337936,0.0273365150145213436352676859005972930024,0.0203559955340018631157871664200778124829,
                   0.00340969025700038358266206367472190257,-0.0044310044639685989896960361095382192429,-0.021347171103236504080385936926021314207,-0.0322779070339724348163166728567572449429,-0.0256972382302023652234622698757719854766,
                   0.0200560044639685989896960361095382192429,0.005722171103236504080385936926021314207,0.0166529070339724348163166728567572449429,0.0413222382302023652234622698757719854766,-0.00539832662063674721902570003835826620637,
                   -0.0117115150145213436352676859005972930024,-0.0047309955340018631157871664200778124829,-0.01903469025700038358266206367472190257,0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,
                   -0.0011816984390824053271310655276900002638,-0.0065929538503378165825423209389454115193,-0.0011816984390824053271310655276900002638,-0.00659295385033781658254232093894541151925,0.00118169843908240532713106552769000026384,
                   0.00659295385033781658254232093894541151925,0.0493518924426097421878012595312173371245,0.0309536240443413439194029911329489388561,-0.0493518924426097421878012595312173371245,-0.0309536240443413439194029911329489388561,
                   -0.0076852257759430755211345928645506704578,0.01071304262232532274726367553371772781055,0.00768522577594307552113459286455067045784,-0.01071304262232532274726367553371772781055,-0.0053427585073154693407857964819990136446,
                   0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,-0.00006849690393994191462545892925639761083,
                   -0.00006849690393994191462545892925639761083,0.00006849690393994191462545892925639761083,0.0804735327963176064441887226697353279632,-0.0193410323853361828045372349169817524248,-0.0181589676146638171954627650830182475752,
                   -0.0429735327963176064441887226697353279632,-0.0856999835607430544139404898898569784646,-0.0132062304783823771165543317442051619267,0.0507062304783823771165543317442051619267,0.0481999835607430544139404898898569784646,
                   -0.0219012822620417557126417885911556797633,0.0195859773138254150912378760480026302811,0.0179140226861745849087621239519973697189,-0.0155987177379582442873582114088443202367,0.0395644418872266973532796317606444188723,
                   0.0016005260562222587539043235245766891337,-0.0391005260562222587539043235245766891337,-0.0020644418872266973532796317606444188723,-0.0193410323853361828045372349169817524248,-0.0218355252342594114746013480190695380569,
                   0.0093355252342594114746013480190695380569,0.0318410323853361828045372349169817524248,-0.0132062304783823771165543317442051619267,0.0007180667433831990794016110471806674338,0.0117819332566168009205983889528193325662,
                   0.0007062304783823771165543317442051619267,0.0195859773138254150912378760480026302811,0.0154907940161104718066743383199079401611,-0.0029907940161104718066743383199079401611,-0.0320859773138254150912378760480026302811,
                   0.0016005260562222587539043235245766891337,-0.0033290317277659049810948545125760315634,-0.0091709682722340950189051454874239684366,0.0108994739437777412460956764754233108663,-0.0181589676146638171954627650830182475752,
                   0.0093355252342594114746013480190695380569,0.0031644747657405885253986519809304619431,0.0056589676146638171954627650830182475752,0.0507062304783823771165543317442051619267,0.0117819332566168009205983889528193325662,
                   -0.0242819332566168009205983889528193325662,-0.0382062304783823771165543317442051619267,0.0179140226861745849087621239519973697189,-0.0029907940161104718066743383199079401611,-0.0095092059838895281933256616800920598389,
                   -0.0054140226861745849087621239519973697189,-0.0391005260562222587539043235245766891337,-0.0091709682722340950189051454874239684366,0.0216709682722340950189051454874239684366,0.0266005260562222587539043235245766891337,
                   -0.0429735327963176064441887226697353279632,0.0318410323853361828045372349169817524248,0.0056589676146638171954627650830182475752,0.0054735327963176064441887226697353279632,0.0481999835607430544139404898898569784646,
                   0.0007062304783823771165543317442051619267,-0.0382062304783823771165543317442051619267,-0.0106999835607430544139404898898569784646,-0.0155987177379582442873582114088443202367,-0.0320859773138254150912378760480026302811,
                   -0.0054140226861745849087621239519973697189,0.0530987177379582442873582114088443202367,-0.0020644418872266973532796317606444188723,0.0108994739437777412460956764754233108663,0.0266005260562222587539043235245766891337,
                   -0.0354355581127733026467203682393555811277,-0.0107118198257438763767877691928324839717,0.0107118198257438763767877691928324839717,0.0055219464080223573894459970409337497945,-0.0055219464080223573894459970409337497945,
                   0.0107118198257438763767877691928324839717,-0.01071181982574387637678776919283248397172,-0.00552194640802235738944599704093374979451,0.0055219464080223573894459970409337497945,0.0361417063948709518329771494328456353773,
                   -0.0361417063948709518329771494328456353773,-0.0361417063948709518329771494328456353773,0.0361417063948709518329771494328456353773,-0.0190530987999342429722176557619595594279,0.0190530987999342429722176557619595594279,
                   0.01905309879993424297221765576195955942791,-0.0190530987999342429722176557619595594279,-0.0107118198257438763767877691928324839717,0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,
                   -0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,-0.00552194640802235738944599704093374979451,-0.01071181982574387637678776919283248397172,0.0055219464080223573894459970409337497945,
                   0.0804735327963176064441887226697353279632,-0.0856999835607430544139404898898569784646,-0.0219012822620417557126417885911556797633,0.0395644418872266973532796317606444188723,-0.0193410323853361828045372349169817524248,
                   -0.0132062304783823771165543317442051619267,0.0195859773138254150912378760480026302811,0.0016005260562222587539043235245766891337,-0.0181589676146638171954627650830182475752,0.0507062304783823771165543317442051619267,
                   0.0179140226861745849087621239519973697189,-0.0391005260562222587539043235245766891337,-0.0429735327963176064441887226697353279632,0.0481999835607430544139404898898569784646,-0.0155987177379582442873582114088443202367,
                   -0.0020644418872266973532796317606444188723,-0.0193410323853361828045372349169817524248,-0.0132062304783823771165543317442051619267,0.0195859773138254150912378760480026302811,0.0016005260562222587539043235245766891337,
                   -0.0218355252342594114746013480190695380569,0.0007180667433831990794016110471806674338,0.0154907940161104718066743383199079401611,-0.0033290317277659049810948545125760315634,0.0093355252342594114746013480190695380569,
                   0.0117819332566168009205983889528193325662,-0.0029907940161104718066743383199079401611,-0.0091709682722340950189051454874239684366,0.0318410323853361828045372349169817524248,0.0007062304783823771165543317442051619267,
                   -0.0320859773138254150912378760480026302811,0.0108994739437777412460956764754233108663,-0.0181589676146638171954627650830182475752,0.0507062304783823771165543317442051619267,0.0179140226861745849087621239519973697189,
                   -0.0391005260562222587539043235245766891337,0.0093355252342594114746013480190695380569,0.0117819332566168009205983889528193325662,-0.0029907940161104718066743383199079401611,-0.0091709682722340950189051454874239684366,
                   0.0031644747657405885253986519809304619431,-0.0242819332566168009205983889528193325662,-0.0095092059838895281933256616800920598389,0.0216709682722340950189051454874239684366,0.0056589676146638171954627650830182475752,
                   -0.0382062304783823771165543317442051619267,-0.0054140226861745849087621239519973697189,0.0266005260562222587539043235245766891337,-0.0429735327963176064441887226697353279632,0.0481999835607430544139404898898569784646,
                   -0.0155987177379582442873582114088443202367,-0.0020644418872266973532796317606444188723,0.0318410323853361828045372349169817524248,0.0007062304783823771165543317442051619267,-0.0320859773138254150912378760480026302811,
                   0.0108994739437777412460956764754233108663,0.0056589676146638171954627650830182475752,-0.0382062304783823771165543317442051619267,-0.0054140226861745849087621239519973697189,0.0266005260562222587539043235245766891337,
                   0.0054735327963176064441887226697353279632,-0.0106999835607430544139404898898569784646,0.0530987177379582442873582114088443202367,-0.0354355581127733026467203682393555811277,-0.0107118198257438763767877691928324839717,
                   0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,-0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,-0.00552194640802235738944599704093374979451,
                   -0.01071181982574387637678776919283248397172,0.0055219464080223573894459970409337497945,-0.0107118198257438763767877691928324839717,0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,
                   -0.0055219464080223573894459970409337497945,0.0107118198257438763767877691928324839717,-0.00552194640802235738944599704093374979451,-0.01071181982574387637678776919283248397172,0.0055219464080223573894459970409337497945,
                   0.0361417063948709518329771494328456353773,-0.0361417063948709518329771494328456353773,-0.0361417063948709518329771494328456353773,0.0361417063948709518329771494328456353773,-0.0190530987999342429722176557619595594279,
                   0.01905309879993424297221765576195955942791,0.0190530987999342429722176557619595594279,-0.0190530987999342429722176557619595594279,0.031060006712696586114307633294975067127,-0.0580689955360314010103039638904617807571,
                   0.0424439955360314010103039638904617807571,-0.015435006712696586114307633294975067127,-0.006503181681188010301934352567263959669,-0.041152828896763495919614063073978685793,0.056777828896763495919614063073978685793,
                   -0.009121818318811989698065647432736040331,0.0004773377993314702175461669132555208505,-0.0302220929660275651836833271432427550571,0.0458470929660275651836833271432427550571,-0.0161023377993314702175461669132555208505,
                   0.0174236430763329497506712696586114307633,-0.0368027617697976347765377301242280145234,0.0211777617697976347765377301242280145234,-0.00179864307633294975067126965861143076333,-0.0426313853067545050667413536611848848136,
                   0.0440006130472902624801358978574168447586,0.0028743869527097375198641021425831552414,-0.0042436146932454949332586463388151151864,-0.0617271234293913618808133576066065517542,0.0405033323743766781741465285769083237438,
                   -0.0040449990410433448408131952435749904104,-0.0163978765706086381191866423933934482458,-0.0555582922605601930496445264377753829231,0.0454275747986191024165707710011507479862,-0.0089692414652857690832374376678174146529,
                   -0.0225667077394398069503554735622246170769,-0.0356508658262350245472608341806654042941,0.0549313489780261932160666337881527754945,-0.0080563489780261932160666337881527754945,-0.0112241341737649754527391658193345957059,
                   0.0217980519734211717334080203278515514802,0.0184993869527097375198641021425831552414,-0.0653743869527097375198641021425831552414,0.0250769480265788282665919796721484485198,0.0408937900960580285474800242732732184209,
                   0.0219966676256233218258534714230916762562,-0.0584550009589566551591868047564250095896,0.0372312099039419714525199757267267815791,0.0347249589272268597163111931044420495897,0.0170724252013808975834292289988492520138,
                   -0.0535307585347142309167625623321825853471,0.0434000410727731402836888068955579504103,0.0148175324929016912139275008473320709608,0.0075686510219738067839333662118472245055,-0.0544436510219738067839333662118472245055,
                   0.0320574675070983087860724991526679290392,-0.0102266733793632527809742999616417337936,-0.0044310044639685989896960361095382192429,0.0200560044639685989896960361095382192429,-0.00539832662063674721902570003835826620637,
                   0.0273365150145213436352676859005972930024,-0.021347171103236504080385936926021314207,0.005722171103236504080385936926021314207,-0.0117115150145213436352676859005972930024,0.0203559955340018631157871664200778124829,
                   -0.0322779070339724348163166728567572449429,0.0166529070339724348163166728567572449429,-0.0047309955340018631157871664200778124829,0.00340969025700038358266206367472190257,-0.0256972382302023652234622698757719854766,
                   0.0413222382302023652234622698757719854766,-0.01903469025700038358266206367472190257,0.0011816984390824053271310655276900002638,-0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,
                   -0.0065929538503378165825423209389454115193,-0.0011816984390824053271310655276900002638,0.00118169843908240532713106552769000026384,-0.00659295385033781658254232093894541151925,0.00659295385033781658254232093894541151925,
                   -0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,
                   -0.00006849690393994191462545892925639761083,-0.00006849690393994191462545892925639761083,0.00006849690393994191462545892925639761083,0.0493518924426097421878012595312173371245,0.0309536240443413439194029911329489388561,
                   -0.0493518924426097421878012595312173371245,-0.0309536240443413439194029911329489388561,-0.0076852257759430755211345928645506704578,0.01071304262232532274726367553371772781055,0.00768522577594307552113459286455067045784,
                   -0.01071304262232532274726367553371772781055,0.01171875,-0.015625,-0.00390625,0.0078125,-0.0208333333333333333333333333333333333333,0.0247395833333333333333333333333333333333,0.0130208333333333333333333333333333333333,
                   -0.0169270833333333333333333333333333333333,0.0065104166666666666666666666666666666667,-0.0026041666666666666666666666666666666667,-0.0143229166666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.0026041666666666666666666666666666666667,-0.0065104166666666666666666666666666666667,0.0052083333333333333333333333333333333333,-0.00130208333333333333333333333333333333333,-0.015625,-0.0533854166666666666666666666666666666667,
                   -0.0338541666666666666666666666666666666667,-0.01171875,0.0247395833333333333333333333333333333333,0.0546875,0.0455729166666666666666666666666666666667,0.0208333333333333333333333333333333333333,-0.0026041666666666666666666666666666666667,
                   -0.0221354166666666666666666666666666666667,-0.03125,-0.0065104166666666666666666666666666666667,-0.0065104166666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.00911458333333333333333333333333333333333,
                   -0.0026041666666666666666666666666666666667,-0.00390625,-0.0338541666666666666666666666666666666667,-0.0143229166666666666666666666666666666667,0,0.0130208333333333333333333333333333333333,0.0455729166666666666666666666666666666667,
                   0.0364583333333333333333333333333333333333,0.0091145833333333333333333333333333333333,-0.0143229166666666666666666666666666666667,-0.03125,-0.0403645833333333333333333333333333333333,-0.0182291666666666666666666666666666666667,
                   0.0052083333333333333333333333333333333333,0.0091145833333333333333333333333333333333,0.0286458333333333333333333333333333333333,0.0091145833333333333333333333333333333333,0.0078125,-0.01171875,0,0.00390625,-0.0169270833333333333333333333333333333333,
                   0.0208333333333333333333333333333333333333,0.00911458333333333333333333333333333333333,-0.0130208333333333333333333333333333333333,0.0104166666666666666666666666666666666667,-0.0065104166666666666666666666666666666667,
                   -0.0182291666666666666666666666666666666667,0.0143229166666666666666666666666666666667,-0.00130208333333333333333333333333333333333,-0.00260416666666666666666666666666666666667,0.0091145833333333333333333333333333333333,
                   -0.00520833333333333333333333333333333333333,0,0,0,0,0,0,0,0,0.0208333333333333333333333333333333333333,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0.01041666666666666666666666666666666666667,0,0,
                   -0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0.031060006712696586114307633294975067127,-0.0426313853067545050667413536611848848136,0.0217980519734211717334080203278515514802,-0.0102266733793632527809742999616417337936,
                   -0.0580689955360314010103039638904617807571,0.0440006130472902624801358978574168447586,0.0184993869527097375198641021425831552414,-0.0044310044639685989896960361095382192429,0.0424439955360314010103039638904617807571,
                   0.0028743869527097375198641021425831552414,-0.0653743869527097375198641021425831552414,0.0200560044639685989896960361095382192429,-0.015435006712696586114307633294975067127,-0.0042436146932454949332586463388151151864,
                   0.0250769480265788282665919796721484485198,-0.00539832662063674721902570003835826620637,-0.006503181681188010301934352567263959669,-0.0617271234293913618808133576066065517542,0.0408937900960580285474800242732732184209,
                   0.0273365150145213436352676859005972930024,-0.041152828896763495919614063073978685793,0.0405033323743766781741465285769083237438,0.0219966676256233218258534714230916762562,-0.021347171103236504080385936926021314207,
                   0.056777828896763495919614063073978685793,-0.0040449990410433448408131952435749904104,-0.0584550009589566551591868047564250095896,0.005722171103236504080385936926021314207,-0.009121818318811989698065647432736040331,
                   -0.0163978765706086381191866423933934482458,0.0372312099039419714525199757267267815791,-0.0117115150145213436352676859005972930024,0.0004773377993314702175461669132555208505,-0.0555582922605601930496445264377753829231,
                   0.0347249589272268597163111931044420495897,0.0203559955340018631157871664200778124829,-0.0302220929660275651836833271432427550571,0.0454275747986191024165707710011507479862,0.0170724252013808975834292289988492520138,
                   -0.0322779070339724348163166728567572449429,0.0458470929660275651836833271432427550571,-0.0089692414652857690832374376678174146529,-0.0535307585347142309167625623321825853471,0.0166529070339724348163166728567572449429,
                   -0.0161023377993314702175461669132555208505,-0.0225667077394398069503554735622246170769,0.0434000410727731402836888068955579504103,-0.0047309955340018631157871664200778124829,0.0174236430763329497506712696586114307633,
                   -0.0356508658262350245472608341806654042941,0.0148175324929016912139275008473320709608,0.00340969025700038358266206367472190257,-0.0368027617697976347765377301242280145234,0.0549313489780261932160666337881527754945,
                   0.0075686510219738067839333662118472245055,-0.0256972382302023652234622698757719854766,0.0211777617697976347765377301242280145234,-0.0080563489780261932160666337881527754945,-0.0544436510219738067839333662118472245055,
                   0.0413222382302023652234622698757719854766,-0.00179864307633294975067126965861143076333,-0.0112241341737649754527391658193345957059,0.0320574675070983087860724991526679290392,-0.01903469025700038358266206367472190257,
                   -0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,
                   -0.00006849690393994191462545892925639761083,-0.00006849690393994191462545892925639761083,0.00006849690393994191462545892925639761083,0.0493518924426097421878012595312173371245,-0.0493518924426097421878012595312173371245,
                   0.0309536240443413439194029911329489388561,-0.0309536240443413439194029911329489388561,-0.0076852257759430755211345928645506704578,0.00768522577594307552113459286455067045784,0.01071304262232532274726367553371772781055,
                   -0.0107130426223253227472636755337177278106,0.0011816984390824053271310655276900002638,-0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,-0.0065929538503378165825423209389454115193,
                   -0.0011816984390824053271310655276900002638,0.00118169843908240532713106552769000026384,-0.00659295385033781658254232093894541151925,0.00659295385033781658254232093894541151925,0.05625,-0.0875,0.00625,0.025,-0.0875,0.11875,0.025,-0.05625,0.00625,0.025,
                   -0.06875,0.0375,0.025,-0.05625,0.0375,-0.00625,0.05,-0.08125,0.0125,0.01875,-0.08125,0.1125,0.01875,-0.05,0.0125,0.01875,-0.075,0.04375,0.01875,-0.05,0.04375,-0.0125,0.04375,-0.075,0.01875,0.0125,-0.075,0.10625,0.0125,-0.04375,0.01875,0.0125,-0.08125,
                   0.05,0.0125,-0.04375,0.05,-0.01875,0.0375,-0.06875,0.025,0.00625,-0.06875,0.1,0.00625,-0.0375,0.025,0.00625,-0.0875,0.05625,0.00625,-0.0375,0.05625,-0.025,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.031060006712696586114307633294975067127,
                   -0.0580689955360314010103039638904617807571,0.0424439955360314010103039638904617807571,-0.015435006712696586114307633294975067127,-0.0426313853067545050667413536611848848136,0.0440006130472902624801358978574168447586,
                   0.0028743869527097375198641021425831552414,-0.0042436146932454949332586463388151151864,0.0217980519734211717334080203278515514802,0.0184993869527097375198641021425831552414,-0.0653743869527097375198641021425831552414,
                   0.0250769480265788282665919796721484485198,-0.0102266733793632527809742999616417337936,-0.0044310044639685989896960361095382192429,0.0200560044639685989896960361095382192429,-0.00539832662063674721902570003835826620637,
                   -0.006503181681188010301934352567263959669,-0.041152828896763495919614063073978685793,0.056777828896763495919614063073978685793,-0.009121818318811989698065647432736040331,-0.0617271234293913618808133576066065517542,
                   0.0405033323743766781741465285769083237438,-0.0040449990410433448408131952435749904104,-0.0163978765706086381191866423933934482458,0.0408937900960580285474800242732732184209,0.0219966676256233218258534714230916762562,
                   -0.0584550009589566551591868047564250095896,0.0372312099039419714525199757267267815791,0.0273365150145213436352676859005972930024,-0.021347171103236504080385936926021314207,0.005722171103236504080385936926021314207,
                   -0.0117115150145213436352676859005972930024,0.0004773377993314702175461669132555208505,-0.0302220929660275651836833271432427550571,0.0458470929660275651836833271432427550571,-0.0161023377993314702175461669132555208505,
                   -0.0555582922605601930496445264377753829231,0.0454275747986191024165707710011507479862,-0.0089692414652857690832374376678174146529,-0.0225667077394398069503554735622246170769,0.0347249589272268597163111931044420495897,
                   0.0170724252013808975834292289988492520138,-0.0535307585347142309167625623321825853471,0.0434000410727731402836888068955579504103,0.0203559955340018631157871664200778124829,-0.0322779070339724348163166728567572449429,
                   0.0166529070339724348163166728567572449429,-0.0047309955340018631157871664200778124829,0.0174236430763329497506712696586114307633,-0.0368027617697976347765377301242280145234,0.0211777617697976347765377301242280145234,
                   -0.00179864307633294975067126965861143076333,-0.0356508658262350245472608341806654042941,0.0549313489780261932160666337881527754945,-0.0080563489780261932160666337881527754945,-0.0112241341737649754527391658193345957059,
                   0.0148175324929016912139275008473320709608,0.0075686510219738067839333662118472245055,-0.0544436510219738067839333662118472245055,0.0320574675070983087860724991526679290392,0.00340969025700038358266206367472190257,
                   -0.0256972382302023652234622698757719854766,0.0413222382302023652234622698757719854766,-0.01903469025700038358266206367472190257,-0.0053427585073154693407857964819990136446,0.0053427585073154693407857964819990136446,
                   0.0053427585073154693407857964819990136446,-0.0053427585073154693407857964819990136446,0.0000684969039399419146254589292563976108,-0.00006849690393994191462545892925639761083,-0.00006849690393994191462545892925639761083,
                   0.0000684969039399419146254589292563976108,0.0011816984390824053271310655276900002638,-0.0011816984390824053271310655276900002638,0.0065929538503378165825423209389454115193,-0.0065929538503378165825423209389454115193,
                   -0.0011816984390824053271310655276900002638,0.00118169843908240532713106552769000026384,-0.00659295385033781658254232093894541151925,0.00659295385033781658254232093894541151925,0.0493518924426097421878012595312173371245,
                   -0.0493518924426097421878012595312173371245,0.0309536240443413439194029911329489388561,-0.0309536240443413439194029911329489388561,-0.0076852257759430755211345928645506704578,0.00768522577594307552113459286455067045784,
                   0.01071304262232532274726367553371772781055,-0.01071304262232532274726367553371772781055,0.01171875,-0.0208333333333333333333333333333333333333,0.0065104166666666666666666666666666666667,0.0026041666666666666666666666666666666667,-0.015625,
                   0.0247395833333333333333333333333333333333,-0.0026041666666666666666666666666666666667,-0.0065104166666666666666666666666666666667,-0.00390625,0.0130208333333333333333333333333333333333,-0.0143229166666666666666666666666666666667,
                   0.00520833333333333333333333333333333333333,0.0078125,-0.0169270833333333333333333333333333333333,0.0104166666666666666666666666666666666667,-0.00130208333333333333333333333333333333333,-0.015625,0.0247395833333333333333333333333333333333,
                   -0.0026041666666666666666666666666666666667,-0.0065104166666666666666666666666666666667,-0.0533854166666666666666666666666666666667,0.0546875,-0.0221354166666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0338541666666666666666666666666666666667,0.0455729166666666666666666666666666666667,-0.03125,0.00911458333333333333333333333333333333333,-0.01171875,0.0208333333333333333333333333333333333333,-0.0065104166666666666666666666666666666667,
                   -0.00260416666666666666666666666666666666667,-0.00390625,0.0130208333333333333333333333333333333333,-0.0143229166666666666666666666666666666667,0.00520833333333333333333333333333333333333,-0.0338541666666666666666666666666666666667,
                   0.0455729166666666666666666666666666666667,-0.03125,0.00911458333333333333333333333333333333333,-0.0143229166666666666666666666666666666667,0.0364583333333333333333333333333333333333,-0.0403645833333333333333333333333333333333,
                   0.0286458333333333333333333333333333333333,0,0.00911458333333333333333333333333333333333,-0.0182291666666666666666666666666666666667,0.0091145833333333333333333333333333333333,0.0078125,-0.0169270833333333333333333333333333333333,
                   0.0104166666666666666666666666666666666667,-0.00130208333333333333333333333333333333333,-0.01171875,0.0208333333333333333333333333333333333333,-0.0065104166666666666666666666666666666667,-0.00260416666666666666666666666666666666667,0,
                   0.00911458333333333333333333333333333333333,-0.0182291666666666666666666666666666666667,0.0091145833333333333333333333333333333333,0.00390625,-0.0130208333333333333333333333333333333333,0.0143229166666666666666666666666666666667,
                   -0.00520833333333333333333333333333333333333,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.0208333333333333333333333333333333333333,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,
                   0.01041666666666666666666666666666666666667,0,0,-0.01041666666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00859375,0.00859375,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00859375,0.00859375,
                   0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,
                   -0.00546875,0.00546875,-0.00546875,-0.01536458333333333333333333333333333333333,-0.01536458333333333333333333333333333333333,-0.00546875,-0.00546875,-0.01536458333333333333333333333333333333333,-0.01536458333333333333333333333333333333333,-0.00546875,
                   0.00546875,-0.00546875,-0.00546875,0.00546875,-0.00546875,0.00546875,0.00546875,-0.00546875,0.00546875,0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,0.00546875,0.00546875,
                   0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,0.00546875,-0.00546875,0.00546875,0.00546875,-0.00546875,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,-0.00859375,-0.00859375,-0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,
                   -0.00859375,-0.00859375,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,
                   -0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,
                   0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,0.01953125,0.01953125,
                   0.01171875,-0.01171875,-0.01953125,-0.01953125,-0.01171875,0.00390625,-0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,-0.01171875,-0.00911458333333333333333333333333333333333,-0.00911458333333333333333333333333333333333,
                   -0.01171875,0.01171875,0.0091145833333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.01171875,-0.00390625,0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,-0.01171875,
                   -0.00911458333333333333333333333333333333333,-0.0091145833333333333333333333333333333333,-0.01171875,0.01171875,0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.01171875,-0.00390625,0.00390625,0.00390625,
                   -0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,0.01953125,0.01953125,0.01171875,-0.01171875,-0.01953125,-0.01953125,-0.01171875,0.00390625,-0.00390625,-0.00390625,0.00390625,-0.01041666666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.00390625,0.01171875,-0.01171875,0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,0.00390625,0.01953125,-0.01953125,
                   -0.00390625,-0.00390625,0.01171875,-0.01171875,0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,-0.00390625,-0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.00390625,-0.00390625,
                   -0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,-0.00390625,-0.00911458333333333333333333333333333333333,
                   0.00911458333333333333333333333333333333333,0.00390625,-0.00390625,-0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,-0.00390625,0.01171875,-0.01171875,
                   0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,-0.00390625,0.01171875,-0.01171875,0.00390625,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.03125,0.03125,0.03125,0.03125,-0.03125,-0.03125,-0.03125,-0.03125,-0.03125,-0.03125,-0.03125,-0.03125,0.03125,0.03125,0.03125,0.03125,-0.03125,-0.03125,-0.03125,
                   -0.03125,0.03125,0.03125,0.03125,0.03125,0.03125,0.03125,0.03125,0.03125,-0.03125,-0.03125,-0.03125,-0.03125,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.0136363636363636363636363636363636363636,
                   0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,0.0136363636363636363636363636363636363636,0.0069805194805194805194805194805194805195,0.0061688311688311688311688311688311688312,
                   -0.00616883116883116883116883116883116883117,-0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,-0.00616883116883116883116883116883116883117,0.0061688311688311688311688311688311688312,
                   0.0069805194805194805194805194805194805195,0.0136363636363636363636363636363636363636,-0.0069805194805194805194805194805194805195,0.0069805194805194805194805194805194805195,-0.0136363636363636363636363636363636363636,
                   0.0212662337662337662337662337662337662338,0.01093073593073593073593073593073593073593,-0.01093073593073593073593073593073593073593,-0.0212662337662337662337662337662337662338,0.01093073593073593073593073593073593073593,
                   0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0109307359307359307359307359307359307359,-0.01093073593073593073593073593073593073593,-0.00492424242424242424242424242424242424242,
                   0.00492424242424242424242424242424242424242,0.0109307359307359307359307359307359307359,-0.0212662337662337662337662337662337662338,-0.0109307359307359307359307359307359307359,0.0109307359307359307359307359307359307359,
                   0.0212662337662337662337662337662337662338,-0.0212662337662337662337662337662337662338,-0.01093073593073593073593073593073593073593,0.0109307359307359307359307359307359307359,0.0212662337662337662337662337662337662338,
                   -0.01093073593073593073593073593073593073593,-0.00492424242424242424242424242424242424242,0.00492424242424242424242424242424242424242,0.01093073593073593073593073593073593073593,0.0109307359307359307359307359307359307359,
                   0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0109307359307359307359307359307359307359,0.0212662337662337662337662337662337662338,0.01093073593073593073593073593073593073593,
                   -0.0109307359307359307359307359307359307359,-0.0212662337662337662337662337662337662338,0.0136363636363636363636363636363636363636,-0.0069805194805194805194805194805194805195,0.0069805194805194805194805194805194805195,
                   -0.01363636363636363636363636363636363636364,-0.0069805194805194805194805194805194805195,-0.0061688311688311688311688311688311688312,0.0061688311688311688311688311688311688312,0.0069805194805194805194805194805194805195,
                   0.0069805194805194805194805194805194805195,0.0061688311688311688311688311688311688312,-0.0061688311688311688311688311688311688312,-0.00698051948051948051948051948051948051948,-0.01363636363636363636363636363636363636364,
                   0.0069805194805194805194805194805194805195,-0.00698051948051948051948051948051948051948,0.01363636363636363636363636363636363636364,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,
                   0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,
                   0.0183982683982683982683982683982683982684,0.0054112554112554112554112554112554112554,-0.00541125541125541125541125541125541125541,-0.0054112554112554112554112554112554112554,0.00541125541125541125541125541125541125541,
                   -0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.0054112554112554112554112554112554112554,
                   -0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,
                   0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,-0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,
                   -0.03125,0.03125,-0.03125,0.03125,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,-0.03125,0.03125,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   -0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,0.03125,0.03125,0.03125,-0.03125,-0.03125,-0.03125,-0.03125,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,-0.03125,-0.03125,-0.03125,0.03125,0.03125,0.03125,0.03125,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,-0.03125,-0.03125,-0.03125,0.03125,0.03125,0.03125,0.03125,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.03125,0.03125,0.03125,0.03125,-0.03125,-0.03125,-0.03125,-0.03125,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01875,-0.00625,0.00625,0.01875,0.01875,0.00625,-0.00625,-0.01875,0.01875,0.00625,-0.00625,-0.01875,-0.01875,-0.00625,0.00625,0.01875,0.01875,0.00625,-0.00625,
                   -0.01875,-0.01875,-0.00625,0.00625,0.01875,-0.01875,-0.00625,0.00625,0.01875,0.01875,0.00625,-0.00625,-0.01875,0.01875,0.00625,-0.00625,-0.01875,-0.01875,-0.00625,0.00625,0.01875,-0.01875,-0.00625,0.00625,0.01875,0.01875,0.00625,-0.00625,-0.01875,
                   -0.01875,-0.00625,0.00625,0.01875,0.01875,0.00625,-0.00625,-0.01875,0.01875,0.00625,-0.00625,-0.01875,-0.01875,-0.00625,0.00625,0.01875,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01875,0.01875,0.01875,-0.01875,-0.00625,0.00625,0.00625,-0.00625,
                   0.00625,-0.00625,-0.00625,0.00625,0.01875,-0.01875,-0.01875,0.01875,0.01875,-0.01875,-0.01875,0.01875,0.00625,-0.00625,-0.00625,0.00625,-0.00625,0.00625,0.00625,-0.00625,-0.01875,0.01875,0.01875,-0.01875,0.01875,-0.01875,-0.01875,0.01875,0.00625,
                   -0.00625,-0.00625,0.00625,-0.00625,0.00625,0.00625,-0.00625,-0.01875,0.01875,0.01875,-0.01875,-0.01875,0.01875,0.01875,-0.01875,-0.00625,0.00625,0.00625,-0.00625,0.00625,-0.00625,-0.00625,0.00625,0.01875,-0.01875,-0.01875,0.01875,0,0,0,0,0,0,0,0,0,0,0,
                   0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666666,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   -0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,
                   -0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   -0.03125,0.03125,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,
                   0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.03125,-0.03125,0.01041666666666666666666666666666666666667,
                   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.00390625,0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,
                   0.01953125,0.01953125,0.01171875,-0.01171875,-0.00911458333333333333333333333333333333333,-0.00911458333333333333333333333333333333333,-0.01171875,-0.01171875,-0.0091145833333333333333333333333333333333,-0.00911458333333333333333333333333333333333,
                   -0.01171875,0.01171875,0.01953125,0.01953125,0.01171875,-0.01171875,-0.01953125,-0.01953125,-0.01171875,0.01171875,0.0091145833333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.01171875,0.01171875,
                   0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.01171875,-0.01171875,-0.01953125,-0.01953125,-0.01171875,0.00390625,-0.00390625,-0.00390625,0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,
                   0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,
                   -0.0136363636363636363636363636363636363636,0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,0.0136363636363636363636363636363636363636,0.0212662337662337662337662337662337662338,
                   0.01093073593073593073593073593073593073593,-0.01093073593073593073593073593073593073593,-0.0212662337662337662337662337662337662338,-0.0212662337662337662337662337662337662338,-0.01093073593073593073593073593073593073593,
                   0.0109307359307359307359307359307359307359,0.0212662337662337662337662337662337662338,0.0136363636363636363636363636363636363636,-0.0069805194805194805194805194805194805195,0.0069805194805194805194805194805194805195,
                   -0.01363636363636363636363636363636363636364,0.0069805194805194805194805194805194805195,0.0061688311688311688311688311688311688312,-0.00616883116883116883116883116883116883117,-0.0069805194805194805194805194805194805195,
                   0.01093073593073593073593073593073593073593,0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0109307359307359307359307359307359307359,-0.01093073593073593073593073593073593073593,
                   -0.00492424242424242424242424242424242424242,0.00492424242424242424242424242424242424242,0.0109307359307359307359307359307359307359,-0.0069805194805194805194805194805194805195,-0.0061688311688311688311688311688311688312,
                   0.0061688311688311688311688311688311688312,0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,-0.00616883116883116883116883116883116883117,0.0061688311688311688311688311688311688312,
                   0.0069805194805194805194805194805194805195,-0.01093073593073593073593073593073593073593,-0.00492424242424242424242424242424242424242,0.00492424242424242424242424242424242424242,0.0109307359307359307359307359307359307359,
                   0.0109307359307359307359307359307359307359,0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0109307359307359307359307359307359307359,0.0069805194805194805194805194805194805195,
                   0.0061688311688311688311688311688311688312,-0.0061688311688311688311688311688311688312,-0.00698051948051948051948051948051948051948,0.0136363636363636363636363636363636363636,-0.0069805194805194805194805194805194805195,
                   0.0069805194805194805194805194805194805195,-0.0136363636363636363636363636363636363636,-0.0212662337662337662337662337662337662338,-0.0109307359307359307359307359307359307359,0.01093073593073593073593073593073593073593,
                   0.0212662337662337662337662337662337662338,0.0212662337662337662337662337662337662338,0.01093073593073593073593073593073593073593,-0.0109307359307359307359307359307359307359,-0.0212662337662337662337662337662337662338,
                   -0.01363636363636363636363636363636363636364,0.0069805194805194805194805194805194805195,-0.00698051948051948051948051948051948051948,0.01363636363636363636363636363636363636364,0.0054112554112554112554112554112554112554,
                   -0.0054112554112554112554112554112554112554,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,
                   0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,
                   -0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,
                   0.0054112554112554112554112554112554112554,-0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,
                   0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,-0.01875,0.01875,0.01875,-0.01875,0.01875,-0.01875,-0.01875,0.01875,0.01875,-0.01875,-0.01875,0.01875,-0.01875,0.01875,
                   0.01875,-0.01875,-0.00625,0.00625,0.00625,-0.00625,0.00625,-0.00625,-0.00625,0.00625,0.00625,-0.00625,-0.00625,0.00625,-0.00625,0.00625,0.00625,-0.00625,0.00625,-0.00625,-0.00625,0.00625,-0.00625,0.00625,0.00625,-0.00625,-0.00625,0.00625,0.00625,
                   -0.00625,0.00625,-0.00625,-0.00625,0.00625,0.01875,-0.01875,-0.01875,0.01875,-0.01875,0.01875,0.01875,-0.01875,-0.01875,0.01875,0.01875,-0.01875,0.01875,-0.01875,-0.01875,0.01875,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   -0.0136363636363636363636363636363636363636,0.0212662337662337662337662337662337662338,-0.0212662337662337662337662337662337662338,0.0136363636363636363636363636363636363636,0.0069805194805194805194805194805194805195,
                   0.01093073593073593073593073593073593073593,-0.01093073593073593073593073593073593073593,-0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,-0.01093073593073593073593073593073593073593,
                   0.0109307359307359307359307359307359307359,0.0069805194805194805194805194805194805195,0.0136363636363636363636363636363636363636,-0.0212662337662337662337662337662337662338,0.0212662337662337662337662337662337662338,
                   -0.01363636363636363636363636363636363636364,0.0069805194805194805194805194805194805195,0.01093073593073593073593073593073593073593,-0.01093073593073593073593073593073593073593,-0.0069805194805194805194805194805194805195,
                   0.0061688311688311688311688311688311688312,0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0061688311688311688311688311688311688312,-0.00616883116883116883116883116883116883117,
                   -0.00492424242424242424242424242424242424242,0.00492424242424242424242424242424242424242,0.0061688311688311688311688311688311688312,-0.0069805194805194805194805194805194805195,-0.0109307359307359307359307359307359307359,
                   0.0109307359307359307359307359307359307359,0.0069805194805194805194805194805194805195,-0.0069805194805194805194805194805194805195,-0.01093073593073593073593073593073593073593,0.0109307359307359307359307359307359307359,
                   0.0069805194805194805194805194805194805195,-0.00616883116883116883116883116883116883117,-0.00492424242424242424242424242424242424242,0.00492424242424242424242424242424242424242,0.0061688311688311688311688311688311688312,
                   0.0061688311688311688311688311688311688312,0.00492424242424242424242424242424242424242,-0.00492424242424242424242424242424242424242,-0.0061688311688311688311688311688311688312,0.0069805194805194805194805194805194805195,
                   0.01093073593073593073593073593073593073593,-0.0109307359307359307359307359307359307359,-0.00698051948051948051948051948051948051948,0.0136363636363636363636363636363636363636,-0.0212662337662337662337662337662337662338,
                   0.0212662337662337662337662337662337662338,-0.01363636363636363636363636363636363636364,-0.0069805194805194805194805194805194805195,-0.0109307359307359307359307359307359307359,0.0109307359307359307359307359307359307359,
                   0.0069805194805194805194805194805194805195,0.0069805194805194805194805194805194805195,0.01093073593073593073593073593073593073593,-0.0109307359307359307359307359307359307359,-0.00698051948051948051948051948051948051948,
                   -0.0136363636363636363636363636363636363636,0.0212662337662337662337662337662337662338,-0.0212662337662337662337662337662337662338,0.01363636363636363636363636363636363636364,0.0054112554112554112554112554112554112554,
                   -0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,
                   0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.0054112554112554112554112554112554112554,-0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,
                   0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,0.00541125541125541125541125541125541125541,-0.00541125541125541125541125541125541125541,
                   -0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,
                   -0.0183982683982683982683982683982683982684,-0.0183982683982683982683982683982683982684,0.0183982683982683982683982683982683982684,-0.00390625,0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,
                   -0.00390625,0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,-0.01171875,-0.01171875,0.01171875,0.01953125,-0.00911458333333333333333333333333333333333,-0.0091145833333333333333333333333333333333,0.01953125,0.01953125,
                   -0.00911458333333333333333333333333333333333,-0.0091145833333333333333333333333333333333,0.01953125,0.01171875,-0.01171875,-0.01171875,0.01171875,-0.01171875,0.01171875,0.01171875,-0.01171875,-0.01953125,0.0091145833333333333333333333333333333333,
                   0.00911458333333333333333333333333333333333,-0.01953125,-0.01953125,0.0091145833333333333333333333333333333333,0.00911458333333333333333333333333333333333,-0.01953125,-0.01171875,0.01171875,0.01171875,-0.01171875,0.00390625,-0.00390625,-0.00390625,
                   0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,-0.00546875,0.00546875,-0.00546875,0.00546875,0.00546875,-0.00546875,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00859375,0.00859375,0.00182291666666666666666666666666666666667,-0.00546875,
                   -0.01536458333333333333333333333333333333333,-0.01536458333333333333333333333333333333333,-0.00546875,0.00546875,0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,0.00546875,
                   -0.00182291666666666666666666666666666666667,-0.00859375,-0.00859375,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00859375,0.00859375,0.00182291666666666666666666666666666666667,-0.00546875,
                   -0.01536458333333333333333333333333333333333,-0.01536458333333333333333333333333333333333,-0.00546875,0.00546875,0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,0.00546875,
                   -0.00182291666666666666666666666666666666667,-0.00859375,-0.00859375,-0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,-0.00546875,0.00546875,-0.00546875,0.00546875,0.00546875,-0.00546875,0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,0,0,0,0,0,0,0,0,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,
                   -0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0,0,0,0,0,0,0,0,-0.00390625,
                   0.01171875,-0.01171875,0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,-0.00390625,0.01171875,-0.01171875,0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,-0.00390625,
                   -0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.00390625,-0.00390625,-0.00911458333333333333333333333333333333333,0.0091145833333333333333333333333333333333,0.00390625,0.00390625,0.01953125,-0.01953125,
                   -0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,-0.00390625,-0.0091145833333333333333333333333333333333,0.00911458333333333333333333333333333333333,0.00390625,-0.00390625,-0.00911458333333333333333333333333333333333,
                   0.00911458333333333333333333333333333333333,0.00390625,0.00390625,0.01953125,-0.01953125,-0.00390625,-0.00390625,0.01171875,-0.01171875,0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,0.00390625,-0.01171875,0.01171875,-0.00390625,-0.00390625,
                   0.01171875,-0.01171875,0.00390625,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,-0.03125,0.03125,-0.03125,0.03125,0.03125,-0.03125,0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,-0.03125,0.03125,-0.03125,0.03125,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,
                   -0.0104166666666666666666666666666666666667,0.03125,-0.03125,-0.03125,0.03125,-0.03125,0.03125,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.03125,-0.03125,-0.03125,0.03125,
                   -0.03125,0.03125,0.03125,-0.03125,0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                   0,0,0,-0.01041666666666666666666666666666666666666,0.03125,-0.03125,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.01041666666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.03125,-0.03125,0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,
                   -0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.0104166666666666666666666666666666666667,
                   0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,
                   0.03125,-0.03125,0.0104166666666666666666666666666666666667,-0.0104166666666666666666666666666666666667,0.03125,-0.03125,0.01041666666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,
                   -0.0104166666666666666666666666666666666667,0.0104166666666666666666666666666666666667,-0.03125,0.03125,-0.0104166666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.03125,-0.03125,
                   0.01041666666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,-0.01171875,-0.01171875,0.01171875,-0.01171875,0.01171875,0.01171875,-0.01171875,0.00390625,-0.00390625,
                   -0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.01953125,-0.00911458333333333333333333333333333333333,-0.00911458333333333333333333333333333333333,0.01953125,-0.01953125,0.0091145833333333333333333333333333333333,
                   0.0091145833333333333333333333333333333333,-0.01953125,-0.00390625,0.00390625,0.00390625,-0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.01953125,-0.0091145833333333333333333333333333333333,-0.0091145833333333333333333333333333333333,
                   0.01953125,-0.01953125,0.00911458333333333333333333333333333333333,0.00911458333333333333333333333333333333333,-0.01953125,-0.00390625,0.00390625,0.00390625,-0.00390625,-0.00390625,0.00390625,0.00390625,-0.00390625,0.01171875,-0.01171875,-0.01171875,
                   0.01171875,-0.01171875,0.01171875,0.01171875,-0.01171875,0.00390625,-0.00390625,-0.00390625,0.00390625,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,
                   -0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,0.01041666666666666666666666666666666666667,-0.01041666666666666666666666666666666666667,
                   0.01041666666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,-0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,
                   0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,-0.00182291666666666666666666666666666666667,0.00859375,-0.01536458333333333333333333333333333333333,
                   0.01536458333333333333333333333333333333333,-0.00859375,0.00859375,-0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,-0.00859375,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,-0.00182291666666666666666666666666666666667,0.00859375,-0.01536458333333333333333333333333333333333,
                   0.01536458333333333333333333333333333333333,-0.00859375,0.00859375,-0.01536458333333333333333333333333333333333,0.01536458333333333333333333333333333333333,-0.00859375,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,
                   -0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,
                   -0.00182291666666666666666666666666666666667,0.00182291666666666666666666666666666666667,-0.00546875,0.00546875,-0.00182291666666666666666666666666666666667,-0.00182291666666666666666666666666666666667,0.00546875,-0.00546875,
                   0.00182291666666666666666666666666666666667,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,-0.00208333333333333333333333333333333333333,
                   -0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333,0.00208333333333333333333333333333333333333};
                return data;
            }

            static const double* N3_tricubicInverse()
            {
                static const double data[64*64] = 
                {
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, -2.5, 2, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 1.5, -1.5, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, -0.25, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, -2.5, 0, 0, 0, 2, 0, 0, 0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 0, 0.5, 0, 1.25, 0, -1.25, 0, -1, 0, 1, 0, 0.25, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, -2.5, 2, -0.5, -2.5, 6.25, -5, 1.25, 2, -5, 4, -1, -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 1.5, -1.5, 0.5, 1.25, -3.75, 3.75, -1.25, -1, 3, -3, 1, 0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 0, 0, 0, 1.5, 0, 0, 0, -1.5, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, -0.25, 0, -0.75, 0, 0.75, 0, 0.75, 0, -0.75, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 1.25, -1, 0.25, 1.5, -3.75, 3, -0.75, -1.5, 3.75, -3, 0.75, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, -0.75, 0.75, -0.25, -0.75, 2.25, -2.25, 0.75, 0.75, -2.25, 2.25, -0.75, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0.25, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0.25, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, -0.5, 0, 0, 0, 1.25, 0, 0, 0, -1, 0, 0, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, -1.25, 0, 0, 0, 1, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0.25, 0, -0.25, 0, -0.625, 0, 0.625, 0, 0.5, 0, -0.5, 0, -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0.625, 0, -0.625, 0, -0.5, 0, 0.5, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    -0.5, 1.25, -1, 0.25, 1.25, -3.125, 2.5, -0.625, -1, 2.5, -2, 0.5, 0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, -1.25, 3.125, -2.5, 0.625, 1, -2.5, 2, -0.5, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0.25, -0.75, 0.75, -0.25, -0.625, 1.875, -1.875, 0.625, 0.5, -1.5, 1.5, -0.5, -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0.625, -1.875, 1.875, -0.625, -0.5, 1.5, -1.5, 0.5, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0.25, 0, 0, 0, -0.75, 0, 0, 0, 0.75, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0.75, 0, 0, 0, -0.75, 0, 0, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    -0.125, 0, 0.125, 0, 0.375, 0, -0.375, 0, -0.375, 0, 0.375, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, -0.375, 0, 0.375, 0, 0.375, 0, -0.375, 0, -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0.25, -0.625, 0.5, -0.125, -0.75, 1.875, -1.5, 0.375, 0.75, -1.875, 1.5, -0.375, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0.75, -1.875, 1.5, -0.375, -0.75, 1.875, -1.5, 0.375, 0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    -0.125, 0.375, -0.375, 0.125, 0.375, -1.125, 1.125, -0.375, -0.375, 1.125, -1.125, 0.375, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, -0.375, 1.125, -1.125, 0.375, 0.375, -1.125, 1.125, -0.375, -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.25, 0, -1.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 1, -2.5, 2, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2.5, 6.25, -5, 1.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, -5, 4, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 1.5, -1.5, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.25, -3.75, 3.75, -1.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 3, -3, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 1.25, 0, 0, 0, 0, 0, 0, 0, -1.25, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 
                    0.25, 0, -0.25, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, -0.625, 0, 0.625, 0, 0, 0, 0, 0, 0.625, 0, -0.625, 0, 0, 0, 0, 0, 0.5, 0, -0.5, 0, 0, 0, 0, 0, -0.5, 0, 0.5, 0, 0, 0, 0, 0, -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, 
                    -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 1.25, -3.125, 2.5, -0.625, 0, 0, 0, 0, -1.25, 3.125, -2.5, 0.625, 0, 0, 0, 0, -1, 2.5, -2, 0.5, 0, 0, 0, 0, 1, -2.5, 2, -0.5, 0, 0, 0, 0, 0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 
                    0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, -0.625, 1.875, -1.875, 0.625, 0, 0, 0, 0, 0.625, -1.875, 1.875, -0.625, 0, 0, 0, 0, 0.5, -1.5, 1.5, -0.5, 0, 0, 0, 0, -0.5, 1.5, -1.5, 0.5, 0, 0, 0, 0, -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, 
                    0, 1, 0, 0, 0, -2.5, 0, 0, 0, 2, 0, 0, 0, -0.5, 0, 0, 0, -2.5, 0, 0, 0, 6.25, 0, 0, 0, -5, 0, 0, 0, 1.25, 0, 0, 0, 2, 0, 0, 0, -5, 0, 0, 0, 4, 0, 0, 0, -1, 0, 0, 0, -0.5, 0, 0, 0, 1.25, 0, 0, 0, -1, 0, 0, 0, 0.25, 0, 0, 
                    -0.5, 0, 0.5, 0, 1.25, 0, -1.25, 0, -1, 0, 1, 0, 0.25, 0, -0.25, 0, 1.25, 0, -1.25, 0, -3.125, 0, 3.125, 0, 2.5, 0, -2.5, 0, -0.625, 0, 0.625, 0, -1, 0, 1, 0, 2.5, 0, -2.5, 0, -2, 0, 2, 0, 0.5, 0, -0.5, 0, 0.25, 0, -0.25, 0, -0.625, 0, 0.625, 0, 0.5, 0, -0.5, 0, -0.125, 0, 0.125, 0, 
                    1, -2.5, 2, -0.5, -2.5, 6.25, -5, 1.25, 2, -5, 4, -1, -0.5, 1.25, -1, 0.25, -2.5, 6.25, -5, 1.25, 6.25, -15.625, 12.5, -3.125, -5, 12.5, -10, 2.5, 1.25, -3.125, 2.5, -0.625, 2, -5, 4, -1, -5, 12.5, -10, 2.5, 4, -10, 8, -2, -1, 2.5, -2, 0.5, -0.5, 1.25, -1, 0.25, 1.25, -3.125, 2.5, -0.625, -1, 2.5, -2, 0.5, 0.25, -0.625, 0.5, -0.125, 
                    -0.5, 1.5, -1.5, 0.5, 1.25, -3.75, 3.75, -1.25, -1, 3, -3, 1, 0.25, -0.75, 0.75, -0.25, 1.25, -3.75, 3.75, -1.25, -3.125, 9.375, -9.375, 3.125, 2.5, -7.5, 7.5, -2.5, -0.625, 1.875, -1.875, 0.625, -1, 3, -3, 1, 2.5, -7.5, 7.5, -2.5, -2, 6, -6, 2, 0.5, -1.5, 1.5, -0.5, 0.25, -0.75, 0.75, -0.25, -0.625, 1.875, -1.875, 0.625, 0.5, -1.5, 1.5, -0.5, -0.125, 0.375, -0.375, 0.125, 
                    0, -0.5, 0, 0, 0, 1.5, 0, 0, 0, -1.5, 0, 0, 0, 0.5, 0, 0, 0, 1.25, 0, 0, 0, -3.75, 0, 0, 0, 3.75, 0, 0, 0, -1.25, 0, 0, 0, -1, 0, 0, 0, 3, 0, 0, 0, -3, 0, 0, 0, 1, 0, 0, 0, 0.25, 0, 0, 0, -0.75, 0, 0, 0, 0.75, 0, 0, 0, -0.25, 0, 0, 
                    0.25, 0, -0.25, 0, -0.75, 0, 0.75, 0, 0.75, 0, -0.75, 0, -0.25, 0, 0.25, 0, -0.625, 0, 0.625, 0, 1.875, 0, -1.875, 0, -1.875, 0, 1.875, 0, 0.625, 0, -0.625, 0, 0.5, 0, -0.5, 0, -1.5, 0, 1.5, 0, 1.5, 0, -1.5, 0, -0.5, 0, 0.5, 0, -0.125, 0, 0.125, 0, 0.375, 0, -0.375, 0, -0.375, 0, 0.375, 0, 0.125, 0, -0.125, 0, 
                    -0.5, 1.25, -1, 0.25, 1.5, -3.75, 3, -0.75, -1.5, 3.75, -3, 0.75, 0.5, -1.25, 1, -0.25, 1.25, -3.125, 2.5, -0.625, -3.75, 9.375, -7.5, 1.875, 3.75, -9.375, 7.5, -1.875, -1.25, 3.125, -2.5, 0.625, -1, 2.5, -2, 0.5, 3, -7.5, 6, -1.5, -3, 7.5, -6, 1.5, 1, -2.5, 2, -0.5, 0.25, -0.625, 0.5, -0.125, -0.75, 1.875, -1.5, 0.375, 0.75, -1.875, 1.5, -0.375, -0.25, 0.625, -0.5, 0.125, 
                    0.25, -0.75, 0.75, -0.25, -0.75, 2.25, -2.25, 0.75, 0.75, -2.25, 2.25, -0.75, -0.25, 0.75, -0.75, 0.25, -0.625, 1.875, -1.875, 0.625, 1.875, -5.625, 5.625, -1.875, -1.875, 5.625, -5.625, 1.875, 0.625, -1.875, 1.875, -0.625, 0.5, -1.5, 1.5, -0.5, -1.5, 4.5, -4.5, 1.5, 1.5, -4.5, 4.5, -1.5, -0.5, 1.5, -1.5, 0.5, -0.125, 0.375, -0.375, 0.125, 0.375, -1.125, 1.125, -0.375, -0.375, 1.125, -1.125, 0.375, 0.125, -0.375, 0.375, -0.125, 
                    0, 0, 0, 0, 0, -0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0.25, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.75, 0, 0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.75, 0, -0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, -0.5, 1.25, -1, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.5, -3.75, 3, -0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1.5, 3.75, -3, 0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.5, -1.25, 1, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0, 0, 0, 0.25, -0.75, 0.75, -0.25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.75, 2.25, -2.25, 0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.75, -2.25, 2.25, -0.75, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.25, 0.75, -0.75, 0.25, 0, 0, 0, 0, 0, 0, 0, 0, 
                    0, 0.25, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, -0.75, 0, 0, 0, 0, 0, 0, 0, 0.75, 0, 0, 0, 0, 0, 0, 0, 0.75, 0, 0, 0, 0, 0, 0, 0, -0.75, 0, 0, 0, 0, 0, 0, 0, -0.25, 0, 0, 0, 0, 0, 0, 0, 0.25, 0, 0, 0, 0, 0, 0, 
                    -0.125, 0, 0.125, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, 0.375, 0, -0.375, 0, 0, 0, 0, 0, -0.375, 0, 0.375, 0, 0, 0, 0, 0, -0.375, 0, 0.375, 0, 0, 0, 0, 0, 0.375, 0, -0.375, 0, 0, 0, 0, 0, 0.125, 0, -0.125, 0, 0, 0, 0, 0, -0.125, 0, 0.125, 0, 0, 0, 0, 0, 
                    0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, -0.75, 1.875, -1.5, 0.375, 0, 0, 0, 0, 0.75, -1.875, 1.5, -0.375, 0, 0, 0, 0, 0.75, -1.875, 1.5, -0.375, 0, 0, 0, 0, -0.75, 1.875, -1.5, 0.375, 0, 0, 0, 0, -0.25, 0.625, -0.5, 0.125, 0, 0, 0, 0, 0.25, -0.625, 0.5, -0.125, 0, 0, 0, 0, 
                    -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, 0.375, -1.125, 1.125, -0.375, 0, 0, 0, 0, -0.375, 1.125, -1.125, 0.375, 0, 0, 0, 0, -0.375, 1.125, -1.125, 0.375, 0, 0, 0, 0, 0.375, -1.125, 1.125, -0.375, 0, 0, 0, 0, 0.125, -0.375, 0.375, -0.125, 0, 0, 0, 0, -0.125, 0.375, -0.375, 0.125, 0, 0, 0, 0, 
                    0, -0.5, 0, 0, 0, 1.25, 0, 0, 0, -1, 0, 0, 0, 0.25, 0, 0, 0, 1.5, 0, 0, 0, -3.75, 0, 0, 0, 3, 0, 0, 0, -0.75, 0, 0, 0, -1.5, 0, 0, 0, 3.75, 0, 0, 0, -3, 0, 0, 0, 0.75, 0, 0, 0, 0.5, 0, 0, 0, -1.25, 0, 0, 0, 1, 0, 0, 0, -0.25, 0, 0, 
                    0.25, 0, -0.25, 0, -0.625, 0, 0.625, 0, 0.5, 0, -0.5, 0, -0.125, 0, 0.125, 0, -0.75, 0, 0.75, 0, 1.875, 0, -1.875, 0, -1.5, 0, 1.5, 0, 0.375, 0, -0.375, 0, 0.75, 0, -0.75, 0, -1.875, 0, 1.875, 0, 1.5, 0, -1.5, 0, -0.375, 0, 0.375, 0, -0.25, 0, 0.25, 0, 0.625, 0, -0.625, 0, -0.5, 0, 0.5, 0, 0.125, 0, -0.125, 0, 
                    -0.5, 1.25, -1, 0.25, 1.25, -3.125, 2.5, -0.625, -1, 2.5, -2, 0.5, 0.25, -0.625, 0.5, -0.125, 1.5, -3.75, 3, -0.75, -3.75, 9.375, -7.5, 1.875, 3, -7.5, 6, -1.5, -0.75, 1.875, -1.5, 0.375, -1.5, 3.75, -3, 0.75, 3.75, -9.375, 7.5, -1.875, -3, 7.5, -6, 1.5, 0.75, -1.875, 1.5, -0.375, 0.5, -1.25, 1, -0.25, -1.25, 3.125, -2.5, 0.625, 1, -2.5, 2, -0.5, -0.25, 0.625, -0.5, 0.125, 
                    0.25, -0.75, 0.75, -0.25, -0.625, 1.875, -1.875, 0.625, 0.5, -1.5, 1.5, -0.5, -0.125, 0.375, -0.375, 0.125, -0.75, 2.25, -2.25, 0.75, 1.875, -5.625, 5.625, -1.875, -1.5, 4.5, -4.5, 1.5, 0.375, -1.125, 1.125, -0.375, 0.75, -2.25, 2.25, -0.75, -1.875, 5.625, -5.625, 1.875, 1.5, -4.5, 4.5, -1.5, -0.375, 1.125, -1.125, 0.375, -0.25, 0.75, -0.75, 0.25, 0.625, -1.875, 1.875, -0.625, -0.5, 1.5, -1.5, 0.5, 0.125, -0.375, 0.375, -0.125, 
                    0, 0.25, 0, 0, 0, -0.75, 0, 0, 0, 0.75, 0, 0, 0, -0.25, 0, 0, 0, -0.75, 0, 0, 0, 2.25, 0, 0, 0, -2.25, 0, 0, 0, 0.75, 0, 0, 0, 0.75, 0, 0, 0, -2.25, 0, 0, 0, 2.25, 0, 0, 0, -0.75, 0, 0, 0, -0.25, 0, 0, 0, 0.75, 0, 0, 0, -0.75, 0, 0, 0, 0.25, 0, 0, 
                    -0.125, 0, 0.125, 0, 0.375, 0, -0.375, 0, -0.375, 0, 0.375, 0, 0.125, 0, -0.125, 0, 0.375, 0, -0.375, 0, -1.125, 0, 1.125, 0, 1.125, 0, -1.125, 0, -0.375, 0, 0.375, 0, -0.375, 0, 0.375, 0, 1.125, 0, -1.125, 0, -1.125, 0, 1.125, 0, 0.375, 0, -0.375, 0, 0.125, 0, -0.125, 0, -0.375, 0, 0.375, 0, 0.375, 0, -0.375, 0, -0.125, 0, 0.125, 0, 
                    0.25, -0.625, 0.5, -0.125, -0.75, 1.875, -1.5, 0.375, 0.75, -1.875, 1.5, -0.375, -0.25, 0.625, -0.5, 0.125, -0.75, 1.875, -1.5, 0.375, 2.25, -5.625, 4.5, -1.125, -2.25, 5.625, -4.5, 1.125, 0.75, -1.875, 1.5, -0.375, 0.75, -1.875, 1.5, -0.375, -2.25, 5.625, -4.5, 1.125, 2.25, -5.625, 4.5, -1.125, -0.75, 1.875, -1.5, 0.375, -0.25, 0.625, -0.5, 0.125, 0.75, -1.875, 1.5, -0.375, -0.75, 1.875, -1.5, 0.375, 0.25, -0.625, 0.5, -0.125, 
                    -0.125, 0.375, -0.375, 0.125, 0.375, -1.125, 1.125, -0.375, -0.375, 1.125, -1.125, 0.375, 0.125, -0.375, 0.375, -0.125, 0.375, -1.125, 1.125, -0.375, -1.125, 3.375, -3.375, 1.125, 1.125, -3.375, 3.375, -1.125, -0.375, 1.125, -1.125, 0.375, -0.375, 1.125, -1.125, 0.375, 1.125, -3.375, 3.375, -1.125, -1.125, 3.375, -3.375, 1.125, 0.375, -1.125, 1.125, -0.375, 0.125, -0.375, 0.375, -0.125, -0.375, 1.125, -1.125, 0.375, 0.375, -1.125, 1.125, -0.375, -0.125, 0.375, -0.375, 0.125
                };
                return data;
            }
        }; // struct StencilPolyData

        // Calculate the coefficients for a polynomial in N dimensions, based on a given stencil of O points
        // evaluated on a rectangular grid, where M is the number of coefficients in the polynomial
        template<int N, int M, int O, typename F>
        void calculateCoefficients(uvector<double,M>& c, const uvector<int,N>* stencil, const double* matrix, const uvector<int,N>& gp, const F& phi)
        {
            uvector<double,O> values;
            for (int i = 0; i < O; ++i)
                values(i) = phi(uvector<int,N>(gp + stencil[i]));
            for (int i = 0; i < M; ++i)
            {
                c(i) = 0.0;
                for (int j = 0; j < O; ++j)
                    c(i) += matrix[i*O + j]*values(j);
            }
        }

        /* Two-dimensional, degree 2 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2
            based on a 12-point stencil. Pseudoinverse has rank 6. */
        struct N2_PolyDegree2
        {
            enum { order = 3 };

            uvector<double,6> c;

            N2_PolyDegree2() {}

            template<typename F>
            N2_PolyDegree2(const uvector<int,2>& i, const F& phi, const uvector<double,2>& dx)
            {
                detail::calculateCoefficients<2,6,12>(c, detail::StencilPolyData::N2_stencil12points(), detail::StencilPolyData::N2_polyDegree2Inverse(), i, phi);
                uvector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
            }

            double operator() (const uvector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*c(5)) + x(0)*(c(1) + x(1)*c(4) + x(0)*c(3));
            }

            uvector<double,2> grad(const uvector<double,2>& x) const
            {
                return uvector<double,2>(c(1) + x(1)*c(4) + x(0)*(2.0*c(3)), c(2) + x(1)*(2.0*c(5)) + x(0)*c(4));
            }

            uvector<double,3> hessian(const uvector<double,2>& x) const
            {
                return uvector<double,3>(2.0*c(3), c(4), 2.0*c(5));
            }
        };

        /* Two-dimensional, degree 3 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2 + c6 x^3 + c7 x^2 y + c8 x y^2 + c9 y^3
            based on a 12-point stencil. Pseudoinverse has rank 10. */
        struct N2_PolyDegree3
        {
            enum { order = 4 };

            uvector<double,10> c;

            N2_PolyDegree3() {}

            template<typename F>
            N2_PolyDegree3(const uvector<int,2>& i, const F& phi, const uvector<double,2>& dx)
            {
                detail::calculateCoefficients<2,10,12>(c, detail::StencilPolyData::N2_stencil12points(), detail::StencilPolyData::N2_polyDegree3Inverse(), i, phi);
                uvector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const uvector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*c(9))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*c(8)) + x(0)*(c(3) + x(1)*c(7) + x(0)*c(6)));
            }

            uvector<double,2> grad(const uvector<double,2>& x) const
            {
                return uvector<double,2>(c(1) + x(1)*(c(4) + x(1)*c(8)) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7)) + x(0)*(3.0*c(6))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9))) + x(0)*(c(4) + x(1)*(2.0*c(8)) + x(0)*c(7)));
            }

            uvector<double,3> hessian(const uvector<double,2>& x) const
            {
                return uvector<double,3>(2.0*c(3) + x(1)*(2.0*c(7)) + x(0)*(6.0*c(6)), c(4) + x(1)*(2.0*c(8)) + x(0)*(2.0*c(7)), 2.0*c(5) + x(1)*(6.0*c(9)) + x(0)*(2.0*c(8)));
            }
        };

        /* Two-dimensional, degree 4 polynomial of the form 
                p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2 + c6 x^3 + c7 x^2 y + c8 x y^2 + c9 y^3 +
                         c10 x^4 + c11 x^3 y + c12 x^2 y^2 + c13 x y^3 + c14 y^4
            based on a 24-point stencil. Pseudoinverse has rank 15. (Note that for a stencil based on a 4 by 4
            patch, as used in bicubic, the Vandermonde matrix has rank 13.) */
        struct N2_PolyDegree4
        {
            enum { order = 5 };

            uvector<double,15> c;

            N2_PolyDegree4() {};

            template<typename F>
            N2_PolyDegree4(const uvector<int,2>& i, const F& phi, const uvector<double,2>& dx)
            {
                detail::calculateCoefficients<2,15,24>(c, detail::StencilPolyData::N2_stencil24points(), detail::StencilPolyData::N2_polyDegree4Inverse(), i, phi);
                uvector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const uvector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*(c(9) + x(1)*c(14)))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*c(13))) + x(0)*(c(3) + x(1)*(c(7) + x(1)*c(12)) + x(0)*(c(6) + x(1)*c(11) + x(0)*c(10))));
            }

            uvector<double,2> grad(const uvector<double,2>& x) const
            {
                return uvector<double,2>(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*c(13))) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12))) + x(0)*(3.0*c(6) + x(1)*(3.0*c(11)) + x(0)*(4.0*c(10)))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9) + x(1)*(4.0*c(14)))) + x(0)*(c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13))) + x(0)*(c(7) + x(1)*(2.0*c(12)) + x(0)*c(11))));
            }

            uvector<double,3> hessian(const uvector<double,2>& x) const
            {
                return uvector<double,3>(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12))) + x(0)*(6.0*c(6) + x(1)*(6.0*c(11)) + x(0)*(12.0*c(10))),
                    c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13))) + x(0)*(2.0*c(7) + x(1)*(4.0*c(12)) + x(0)*(3.0*c(11))),
                    2.0*c(5) + x(1)*(6.0*c(9) + x(1)*(12.0*c(14))) + x(0)*(2.0*c(8) + x(1)*(6.0*c(13)) + x(0)*(2.0*c(12))));
            }
        };

        /* Two-dimensional, degree 5 polynomial of the form 
                p(x,y) = c0 + c1 x + c2 y + ... + c20 y^5
            based on a 24-point stencil. Pseudoinverse has rank 21. */
        struct N2_PolyDegree5
        {
            enum { order = 6 };

            uvector<double,21> c;

            N2_PolyDegree5() {};

            template<typename F>
            N2_PolyDegree5(const uvector<int,2>& i, const F& phi, const uvector<double,2>& dx)
            {
                detail::calculateCoefficients<2,21,24>(c, detail::StencilPolyData::N2_stencil24points(), detail::StencilPolyData::N2_polyDegree5Inverse(), i, phi);
                uvector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(15) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(16) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(17) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(18) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(19) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(20) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const uvector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*(c(9) + x(1)*(c(14) + x(1)*c(20))))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*(c(13) + x(1)*c(19)))) + x(0)*(c(3) + x(1)*(c(7) + x(1)*(c(12) + x(1)*c(18))) + x(0)*(c(6) + x(1)*(c(11) + x(1)*c(17)) + x(0)*(c(10) + x(1)*c(16) + x(0)*c(15)))));
            }

            uvector<double,2> grad(const uvector<double,2>& x) const
            {
                return uvector<double,2>(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*(c(13) + x(1)*c(19)))) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12) + x(1)*(2.0*c(18)))) + x(0)*(3.0*c(6) + x(1)*(3.0*c(11) + x(1)*(3.0*c(17))) + x(0)*(4.0*c(10) + x(1)*(4.0*c(16)) + x(0)*(5.0*c(15))))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9) + x(1)*(4.0*c(14) + x(1)*(5.0*c(20))))) + x(0)*(c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13) + x(1)*(4.0*c(19)))) + x(0)*(c(7) + x(1)*(2.0*c(12) + x(1)*(3.0*c(18))) + x(0)*(c(11) + x(1)*(2.0*c(17)) + x(0)*c(16)))));
            }

            uvector<double,3> hessian(const uvector<double,2>& x) const
            {
                return uvector<double,3>(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12) + x(1)*(2.0*c(18)))) + x(0)*(6.0*c(6) + x(1)*(6.0*c(11) + x(1)*(6.0*c(17))) + x(0)*(12.0*c(10) + x(1)*(12.0*c(16)) + x(0)*(20.0*c(15)))),
                    c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13) + x(1)*(4.0*c(19)))) + x(0)*(2.0*c(7) + x(1)*(4.0*c(12) + x(1)*(6.0*c(18))) + x(0)*(3.0*c(11) + x(1)*(6.0*c(17)) + x(0)*(4.0*c(16)))),
                    2.0*c(5) + x(1)*(6.0*c(9) + x(1)*(12.0*c(14) + x(1)*(20.0*c(20)))) + x(0)*(2.0*c(8) + x(1)*(6.0*c(13) + x(1)*(12.0*c(19))) + x(0)*(2.0*c(12) + x(1)*(6.0*c(18)) + x(0)*(2.0*c(17)))));
            }
        };

        /* Two-dimensional, bicubic interpolant guaranteeing C^1 continuity between cells on a rectangular
            grid, using second order finite differences evaluated at gird points to ensure continuity of
            the gradient across grid cells. This interpolant is equivalent to that discussed in D. L. Chopp, 
            Some improvements of the Fast Marching Method, SIAM Journal Scientific Computing 23(1) (2001) 230244. */
        struct N2_Bicubic
        {
            enum { order = 3 };

            uvector<double,16> c;

            template<typename F>
            N2_Bicubic(const uvector<int,2>& i, const F& phi, const uvector<double,2>& dx)
            {
                double stencil[16];
                int index = 0;
                for (MultiLoop<2> b(0, 4); ~b; ++b)
                    stencil[index++] = phi(i - 1 + b());

                for (int i = 0; i < 16; ++i)
                {
                    c(i) = 0.0;
                    for (int j = 0; j < 16; ++j)
                        c(i) += detail::StencilPolyData::N2_bicubicInverse()[i*16+j]*stencil[j];
                }

                uvector<double,2> dxinv = 1.0 / dx;
                double sx = 1.0;
                for (int i = 0; i < 4; ++i)
                {
                    double sy = 1.0;
                    for (int j = 0; j < 4; ++j)
                    {
                        c(i*4+j) *= sx*sy;
                        sy *= dxinv(1);
                    }
                    sx *= dxinv(0);
                }
            }

            double operator() (const uvector<double,2>& x) const
            {
                return c(0) + x(1)*(c(1) + x(1)*(c(2) + x(1)*c(3))) + x(0)*(c(4) + x(1)*(c(5) + x(1)*(c(6) + x(1)*c(7))) + x(0)*(c(8) + x(1)*(c(9) + x(1)*(c(10) + x(1)*c(11))) + x(0)*(c(12) + x(1)*(c(13) + x(1)*(c(14) + x(1)*c(15))))));
            }

            uvector<double,2> grad(const uvector<double,2>& x) const
            {
                return uvector<double,2>(c(4) + x(1)*(c(5) + x(1)*(c(6) + x(1)*c(7))) + x(0)*(2.0*c(8) + x(1)*(2.0*c(9) + x(1)*(2.0*c(10) + x(1)*(2.0*c(11)))) + x(0)*(3.0*c(12) + x(1)*(3.0*c(13) + x(1)*(3.0*c(14) + x(1)*(3.0*c(15)))))),
                    c(1) + x(1)*(2.0*c(2) + x(1)*(3.0*c(3))) + x(0)*(c(5) + x(1)*(2.0*c(6) + x(1)*(3.0*c(7))) + x(0)*(c(9) + x(1)*(2.0*c(10) + x(1)*(3.0*c(11))) + x(0)*(c(13) + x(1)*(2.0*c(14) + x(1)*(3.0*c(15)))))));
            }

            uvector<double,3> hessian(const uvector<double,2>& x) const
            {
                return uvector<double,3>(2.0*c(8) + x(1)*(2.0*c(9) + x(1)*(2.0*c(10) + x(1)*(2.0*c(11)))) + x(0)*(6.0*c(12) + x(1)*(6.0*c(13) + x(1)*(6.0*c(14) + x(1)*(6.0*c(15))))),
                    c(5) + x(1)*(2.0*c(6) + x(1)*(3.0*c(7))) + x(0)*(2.0*c(9) + x(1)*(4.0*c(10) + x(1)*(6.0*c(11))) + x(0)*(3.0*c(13) + x(1)*(6.0*c(14) + x(1)*(9.0*c(15))))),
                    2.0*c(2) + x(1)*(6.0*c(3)) + x(0)*(2.0*c(6) + x(1)*(6.0*c(7)) + x(0)*(2.0*c(10) + x(1)*(6.0*c(11)) + x(0)*(2.0*c(14) + x(1)*(6.0*c(15))))));
            }
        };

        /* Three-dimensional, degree 2 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2
            based on a 32-point stencil. Pseudoinverse has rank 10. */
        struct N3_PolyDegree2
        {
            enum { order = 3 };

            uvector<double,10> c;

            N3_PolyDegree2() {}

            template<typename F>
            N3_PolyDegree2(const uvector<int,3>& i, const F& phi, const uvector<double,3>& dx)
            {
                detail::calculateCoefficients<3,10,32>(c, detail::StencilPolyData::N3_stencil32points(), detail::StencilPolyData::N3_polyDegree2Inverse(), i, phi);
                uvector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
            }

            double operator() (const uvector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*c(9)) + x(1)*(c(2) + x(2)*c(8) + x(1)*c(7)) + x(0)*(c(1) + x(2)*c(6) + x(1)*c(5) + x(0)*c(4));
            }

            uvector<double,3> grad(const uvector<double,3>& x) const
            {
                return uvector<double,3>(c(1) + x(2)*c(6) + x(1)*c(5) + x(0)*(2.0*c(4)), c(2) + x(2)*c(8) + x(1)*(2.0*c(7)) + x(0)*c(5), c(3) + x(2)*(2.0*c(9)) + x(1)*c(8) + x(0)*c(6));
            }

            uvector<double,6> hessian(const uvector<double,3>& x) const
            {
                return uvector<double,6>(2.0*c(4), c(5), c(6), 2.0*c(7), c(8), 2.0*c(9));
            }
        };

        /* Three-dimensional, degree 3 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2 +
                         c10 x^3 + c11 x^2 y + c12 x^2 z + c13 x y^2 + c14 x y z + c15 x z^2 + c16 y^3 + c17 y^2 z + c18 y z^2 + c19 z^3
            based on a 32-point stencil. Pseudoinverse has rank 20. */
        struct N3_PolyDegree3
        {
            enum { order = 4 };

            uvector<double,20> c;

            N3_PolyDegree3() {}

            template<typename F>
            N3_PolyDegree3(const uvector<int,3>& i, const F& phi, const uvector<double,3>& dx)
            {
                detail::calculateCoefficients<3,20,32>(c, detail::StencilPolyData::N3_stencil32points(), detail::StencilPolyData::N3_polyDegree3Inverse(), i, phi);
                uvector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const uvector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*c(19))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*c(18)) + x(1)*(c(7) + x(2)*c(17) + x(1)*c(16))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*c(15)) + x(1)*(c(5) + x(2)*c(14) + x(1)*c(13)) + x(0)*(c(4) + x(2)*c(12) + x(1)*c(11) + x(0)*c(10)));
            }

            uvector<double,3> grad(const uvector<double,3>& x) const
            {
                return uvector<double,3>(c(1) + x(2)*(c(6) + x(2)*c(15)) + x(1)*(c(5) + x(2)*c(14) + x(1)*c(13)) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12)) + x(1)*(2.0*c(11)) + x(0)*(3.0*c(10))),
                    c(2) + x(2)*(c(8) + x(2)*c(18)) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17)) + x(1)*(3.0*c(16))) + x(0)*(c(5) + x(2)*c(14) + x(1)*(2.0*c(13)) + x(0)*c(11)),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19))) + x(1)*(c(8) + x(2)*(2.0*c(18)) + x(1)*c(17)) + x(0)*(c(6) + x(2)*(2.0*c(15)) + x(1)*c(14) + x(0)*c(12)));
            }

            uvector<double,6> hessian(const uvector<double,3>& x) const
            {
                return uvector<double,6>(2.0*c(4) + x(2)*(2.0*c(12)) + x(1)*(2.0*c(11)) + x(0)*(6.0*c(10)), c(5) + x(2)*c(14) + x(1)*(2.0*c(13)) + x(0)*(2.0*c(11)),
                    c(6) + x(2)*(2.0*c(15)) + x(1)*c(14) + x(0)*(2.0*c(12)), 2.0*c(7) + x(2)*(2.0*c(17)) + x(1)*(6.0*c(16)) + x(0)*(2.0*c(13)),
                    c(8) + x(2)*(2.0*c(18)) + x(1)*(2.0*c(17)) + x(0)*c(14), 2.0*c(9) + x(2)*(6.0*c(19)) + x(1)*(2.0*c(18)) + x(0)*(2.0*c(15)));
            }
        };

        /* Three-dimensional, degree 4 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2 +
                         c10 x^3 + c11 x^2 y + c12 x^2 z + c13 x y^2 + c14 x y z + c15 x z^2 + c16 y^3 + c17 y^2 z + c18 y z^2 + c19 z^3 +
                         c20 x^4 + c21 x^3 y + c22 x^3 z + c23 x^2 y^2 + c24 x^2 y z + c25 x^2 z^2 + c26 x y^3 + c27 x y^2 z + c28 x y z^2 +
                         c29 x z^3 + c30 y^4 + c31 y^3 z + c32 y^2 z^2 + c33 y z^3 + c34 z^4
            based on an 88-point stencil. Pseudoinverse has rank 35. */
        struct N3_PolyDegree4
        {
            enum { order = 5 };

            uvector<double,35> c;

            N3_PolyDegree4() {}

            template<typename F>
            N3_PolyDegree4(const uvector<int,3>& i, const F& phi, const uvector<double,3>& dx)
            {
                detail::calculateCoefficients<3,35,88>(c, detail::StencilPolyData::N3_stencil88points(), detail::StencilPolyData::N3_polyDegree4Inverse(), i, phi);
                uvector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
                c(20) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(21) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(22) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(23) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(24) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(25) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(26) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(27) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(28) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(29) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(30) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(31) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(32) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(33) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(34) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const uvector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*(c(19) + x(2)*c(34)))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*c(33))) + x(1)*(c(7) + x(2)*(c(17) + x(2)*c(32)) + x(1)*(c(16) + x(2)*c(31) + x(1)*c(30)))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*c(29))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(c(13) + x(2)*c(27) + x(1)*c(26))) + x(0)*(c(4) + x(2)*(c(12) + x(2)*c(25)) + x(1)*(c(11) + x(2)*c(24) + x(1)*c(23)) + x(0)*(c(10) + x(2)*c(22) + x(1)*c(21) + x(0)*c(20))));
            }

            uvector<double,3> grad(const uvector<double,3>& x) const
            {
                return uvector<double,3>(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*c(29))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(c(13) + x(2)*c(27) + x(1)*c(26))) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(2.0*c(23))) + x(0)*(3.0*c(10) + x(2)*(3.0*c(22)) + x(1)*(3.0*c(21)) + x(0)*(4.0*c(20)))),
                    c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*c(33))) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32))) + x(1)*(3.0*c(16) + x(2)*(3.0*c(31)) + x(1)*(4.0*c(30)))) + x(0)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(3.0*c(26))) + x(0)*(c(11) + x(2)*c(24) + x(1)*(2.0*c(23)) + x(0)*c(21))),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19) + x(2)*(4.0*c(34)))) + x(1)*(c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33))) + x(1)*(c(17) + x(2)*(2.0*c(32)) + x(1)*c(31))) + x(0)*(c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29))) + x(1)*(c(14) + x(2)*(2.0*c(28)) + x(1)*c(27)) + x(0)*(c(12) + x(2)*(2.0*c(25)) + x(1)*c(24) + x(0)*c(22))));
            }

            uvector<double,6> hessian(const uvector<double,3>& x) const
            {
                return uvector<double,6>(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(2.0*c(23))) + x(0)*(6.0*c(10) + x(2)*(6.0*c(22)) + x(1)*(6.0*c(21)) + x(0)*(12.0*c(20))),
                    c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(3.0*c(26))) + x(0)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(4.0*c(23)) + x(0)*(3.0*c(21))),
                    c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29))) + x(1)*(c(14) + x(2)*(2.0*c(28)) + x(1)*c(27)) + x(0)*(2.0*c(12) + x(2)*(4.0*c(25)) + x(1)*(2.0*c(24)) + x(0)*(3.0*c(22))),
                    2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32))) + x(1)*(6.0*c(16) + x(2)*(6.0*c(31)) + x(1)*(12.0*c(30))) + x(0)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(6.0*c(26)) + x(0)*(2.0*c(23))),
                    c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33))) + x(1)*(2.0*c(17) + x(2)*(4.0*c(32)) + x(1)*(3.0*c(31))) + x(0)*(c(14) + x(2)*(2.0*c(28)) + x(1)*(2.0*c(27)) + x(0)*c(24)),
                    2.0*c(9) + x(2)*(6.0*c(19) + x(2)*(12.0*c(34))) + x(1)*(2.0*c(18) + x(2)*(6.0*c(33)) + x(1)*(2.0*c(32))) + x(0)*(2.0*c(15) + x(2)*(6.0*c(29)) + x(1)*(2.0*c(28)) + x(0)*(2.0*c(25))));
            }
        };

        /* Three-dimensional, degree 5 polynomial of the form
                p(x,y) = c0 + c1 x + c2 y + c3 z + ... + c55 z^5
            based on an 88-point stencil. Pseudoinverse has rank 56. */
        struct N3_PolyDegree5
        {
            enum { order = 6 };

            uvector<double,56> c;

            N3_PolyDegree5() {}

            template<typename F>
            N3_PolyDegree5(const uvector<int,3>& i, const F& phi, const uvector<double,3>& dx)
            {
                detail::calculateCoefficients<3,56,88>(c, detail::StencilPolyData::N3_stencil88points(), detail::StencilPolyData::N3_polyDegree5Inverse(), i, phi);
                uvector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
                c(20) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(21) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(22) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(23) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(24) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(25) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(26) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(27) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(28) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(29) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(30) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(31) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(32) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(33) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(34) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(35) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(36) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(37) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(38) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(39) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(40) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(41) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(42) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(43) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(44) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(45) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(46) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(47) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(48) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(49) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(50) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(51) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(52) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(53) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(54) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(55) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const uvector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*(c(19) + x(2)*(c(34) + x(2)*c(55))))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*(c(33) + x(2)*c(54)))) + x(1)*(c(7) + x(2)*(c(17) + x(2)*(c(32) + x(2)*c(53))) + x(1)*(c(16) + x(2)*(c(31) + x(2)*c(52)) + x(1)*(c(30) + x(2)*c(51) + x(1)*c(50))))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*(c(29) + x(2)*c(49)))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(c(13) + x(2)*(c(27) + x(2)*c(47)) + x(1)*(c(26) + x(2)*c(46) + x(1)*c(45)))) + x(0)*(c(4) + x(2)*(c(12) + x(2)*(c(25) + x(2)*c(44))) + x(1)*(c(11) + x(2)*(c(24) + x(2)*c(43)) + x(1)*(c(23) + x(2)*c(42) + x(1)*c(41))) + x(0)*(c(10) + x(2)*(c(22) + x(2)*c(40)) + x(1)*(c(21) + x(2)*c(39) + x(1)*c(38)) + x(0)*(c(20) + x(2)*c(37) + x(1)*c(36) + x(0)*c(35)))));
            }

            uvector<double,3> grad(const uvector<double,3>& x) const
            {
                return uvector<double,3>(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*(c(29) + x(2)*c(49)))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(c(13) + x(2)*(c(27) + x(2)*c(47)) + x(1)*(c(26) + x(2)*c(46) + x(1)*c(45)))) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25) + x(2)*(2.0*c(44)))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(2.0*c(41)))) + x(0)*(3.0*c(10) + x(2)*(3.0*c(22) + x(2)*(3.0*c(40))) + x(1)*(3.0*c(21) + x(2)*(3.0*c(39)) + x(1)*(3.0*c(38))) + x(0)*(4.0*c(20) + x(2)*(4.0*c(37)) + x(1)*(4.0*c(36)) + x(0)*(5.0*c(35))))),
                    c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*(c(33) + x(2)*c(54)))) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32) + x(2)*(2.0*c(53)))) + x(1)*(3.0*c(16) + x(2)*(3.0*c(31) + x(2)*(3.0*c(52))) + x(1)*(4.0*c(30) + x(2)*(4.0*c(51)) + x(1)*(5.0*c(50))))) + x(0)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(3.0*c(26) + x(2)*(3.0*c(46)) + x(1)*(4.0*c(45)))) + x(0)*(c(11) + x(2)*(c(24) + x(2)*c(43)) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(3.0*c(41))) + x(0)*(c(21) + x(2)*c(39) + x(1)*(2.0*c(38)) + x(0)*c(36)))),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19) + x(2)*(4.0*c(34) + x(2)*(5.0*c(55))))) + x(1)*(c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33) + x(2)*(4.0*c(54)))) + x(1)*(c(17) + x(2)*(2.0*c(32) + x(2)*(3.0*c(53))) + x(1)*(c(31) + x(2)*(2.0*c(52)) + x(1)*c(51)))) + x(0)*(c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29) + x(2)*(4.0*c(49)))) + x(1)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(c(27) + x(2)*(2.0*c(47)) + x(1)*c(46))) + x(0)*(c(12) + x(2)*(2.0*c(25) + x(2)*(3.0*c(44))) + x(1)*(c(24) + x(2)*(2.0*c(43)) + x(1)*c(42)) + x(0)*(c(22) + x(2)*(2.0*c(40)) + x(1)*c(39) + x(0)*c(37)))));
            }

            uvector<double,6> hessian(const uvector<double,3>& x) const
            {
                return uvector<double,6>(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25) + x(2)*(2.0*c(44)))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(2.0*c(41)))) + x(0)*(6.0*c(10) + x(2)*(6.0*c(22) + x(2)*(6.0*c(40))) + x(1)*(6.0*c(21) + x(2)*(6.0*c(39)) + x(1)*(6.0*c(38))) + x(0)*(12.0*c(20) + x(2)*(12.0*c(37)) + x(1)*(12.0*c(36)) + x(0)*(20.0*c(35)))),
                    c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(3.0*c(26) + x(2)*(3.0*c(46)) + x(1)*(4.0*c(45)))) + x(0)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(4.0*c(23) + x(2)*(4.0*c(42)) + x(1)*(6.0*c(41))) + x(0)*(3.0*c(21) + x(2)*(3.0*c(39)) + x(1)*(6.0*c(38)) + x(0)*(4.0*c(36)))),
                    c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29) + x(2)*(4.0*c(49)))) + x(1)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(c(27) + x(2)*(2.0*c(47)) + x(1)*c(46))) + x(0)*(2.0*c(12) + x(2)*(4.0*c(25) + x(2)*(6.0*c(44))) + x(1)*(2.0*c(24) + x(2)*(4.0*c(43)) + x(1)*(2.0*c(42))) + x(0)*(3.0*c(22) + x(2)*(6.0*c(40)) + x(1)*(3.0*c(39)) + x(0)*(4.0*c(37)))),
                    2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32) + x(2)*(2.0*c(53)))) + x(1)*(6.0*c(16) + x(2)*(6.0*c(31) + x(2)*(6.0*c(52))) + x(1)*(12.0*c(30) + x(2)*(12.0*c(51)) + x(1)*(20.0*c(50)))) + x(0)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(6.0*c(26) + x(2)*(6.0*c(46)) + x(1)*(12.0*c(45))) + x(0)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(6.0*c(41)) + x(0)*(2.0*c(38)))),
                    c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33) + x(2)*(4.0*c(54)))) + x(1)*(2.0*c(17) + x(2)*(4.0*c(32) + x(2)*(6.0*c(53))) + x(1)*(3.0*c(31) + x(2)*(6.0*c(52)) + x(1)*(4.0*c(51)))) + x(0)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(2.0*c(27) + x(2)*(4.0*c(47)) + x(1)*(3.0*c(46))) + x(0)*(c(24) + x(2)*(2.0*c(43)) + x(1)*(2.0*c(42)) + x(0)*c(39))),
                    2.0*c(9) + x(2)*(6.0*c(19) + x(2)*(12.0*c(34) + x(2)*(20.0*c(55)))) + x(1)*(2.0*c(18) + x(2)*(6.0*c(33) + x(2)*(12.0*c(54))) + x(1)*(2.0*c(32) + x(2)*(6.0*c(53)) + x(1)*(2.0*c(52)))) + x(0)*(2.0*c(15) + x(2)*(6.0*c(29) + x(2)*(12.0*c(49))) + x(1)*(2.0*c(28) + x(2)*(6.0*c(48)) + x(1)*(2.0*c(47))) + x(0)*(2.0*c(25) + x(2)*(6.0*c(44)) + x(1)*(2.0*c(43)) + x(0)*(2.0*c(40)))));
            }
        };

        /* Three-dimensional, tricubic interpolant guaranteeing C^1 continuity between cells on a rectangular
            grid, using second order finite differences evaluated at gird points to ensure continuity of
            the gradient across grid cells. This interpolant is the analogue of the bicubic in two dimensions. */
        struct N3_Tricubic
        {
            enum { order = 3 };

            uvector<double,64> c;

            template<typename F>
            N3_Tricubic(const uvector<int,3>& i, const F& phi, const uvector<double,3>& dx)
            {
                double stencil[4*4*4];
                int index = 0;
                for (MultiLoop<3> b(0, 4); ~b; ++b)
                    stencil[index++] = phi(i - 1 + b());

                for (int i = 0; i < 64; ++i)
                {
                    c(i) = 0.0;
                    for (int j = 0; j < 64; ++j)
                        c(i) += detail::StencilPolyData::N3_tricubicInverse()[i*64+j]*stencil[j];
                }

                uvector<double,3> dxinv = 1.0 / dx;
                double sx = 1.0;
                for (int i = 0; i < 4; ++i)
                {
                    double sy = 1.0;
                    for (int j = 0; j < 4; ++j)
                    {
                        double sz = 1.0;
                        for (int k = 0; k < 4; ++k)
                        {
                            c(i*16+j*4+k) *= sx*sy*sz;
                            sz *= dxinv(2);
                        }
                        sy *= dxinv(1);
                    }
                    sx *= dxinv(0);
                }
            }

            double operator() (const uvector<double,3>& x) const
            {
                return c(0) + x(2)*(c(1) + x(2)*(c(2) + x(2)*c(3))) + x(1)*(c(4) + x(2)*(c(5) + x(2)*(c(6) + x(2)*c(7))) + x(1)*(c(8) + x(2)*(c(9) + x(2)*(c(10) + x(2)*c(11))) + x(1)*(c(12) + x(2)*(c(13) + x(2)*(c(14) + x(2)*c(15)))))) + x(0)*(c(16) + x(2)*(c(17) + x(2)*(c(18) + x(2)*c(19))) + x(1)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(c(24) + x(2)*(c(25) + x(2)*(c(26) + x(2)*c(27))) + x(1)*(c(28) + x(2)*(c(29) + x(2)*(c(30) + x(2)*c(31)))))) + x(0)*(c(32) + x(2)*(c(33) + x(2)*(c(34) + x(2)*c(35))) + x(1)*(c(36) + x(2)*(c(37) + x(2)*(c(38) + x(2)*c(39))) + x(1)*(c(40) + x(2)*(c(41) + x(2)*(c(42) + x(2)*c(43))) + x(1)*(c(44) + x(2)*(c(45) + x(2)*(c(46) + x(2)*c(47)))))) + x(0)*(c(48) + x(2)*(c(49) + x(2)*(c(50) + x(2)*c(51))) + x(1)*(c(52) + x(2)*(c(53) + x(2)*(c(54) + x(2)*c(55))) + x(1)*(c(56) + x(2)*(c(57) + x(2)*(c(58) + x(2)*c(59))) + x(1)*(c(60) + x(2)*(c(61) + x(2)*(c(62) + x(2)*c(63)))))))));
            }

            uvector<double,3> grad(const uvector<double,3>& x) const
            {
                return uvector<double,3>(c(16) + x(2)*(c(17) + x(2)*(c(18) + x(2)*c(19))) + x(1)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(c(24) + x(2)*(c(25) + x(2)*(c(26) + x(2)*c(27))) + x(1)*(c(28) + x(2)*(c(29) + x(2)*(c(30) + x(2)*c(31)))))) + x(0)*(2.0*c(32) + x(2)*(2.0*c(33) + x(2)*(2.0*c(34) + x(2)*(2.0*c(35)))) + x(1)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(2.0*c(44) + x(2)*(2.0*c(45) + x(2)*(2.0*c(46) + x(2)*(2.0*c(47))))))) + x(0)*(3.0*c(48) + x(2)*(3.0*c(49) + x(2)*(3.0*c(50) + x(2)*(3.0*c(51)))) + x(1)*(3.0*c(52) + x(2)*(3.0*c(53) + x(2)*(3.0*c(54) + x(2)*(3.0*c(55)))) + x(1)*(3.0*c(56) + x(2)*(3.0*c(57) + x(2)*(3.0*c(58) + x(2)*(3.0*c(59)))) + x(1)*(3.0*c(60) + x(2)*(3.0*c(61) + x(2)*(3.0*c(62) + x(2)*(3.0*c(63))))))))),
                    c(4) + x(2)*(c(5) + x(2)*(c(6) + x(2)*c(7))) + x(1)*(2.0*c(8) + x(2)*(2.0*c(9) + x(2)*(2.0*c(10) + x(2)*(2.0*c(11)))) + x(1)*(3.0*c(12) + x(2)*(3.0*c(13) + x(2)*(3.0*c(14) + x(2)*(3.0*c(15)))))) + x(0)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(3.0*c(28) + x(2)*(3.0*c(29) + x(2)*(3.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(c(36) + x(2)*(c(37) + x(2)*(c(38) + x(2)*c(39))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(3.0*c(44) + x(2)*(3.0*c(45) + x(2)*(3.0*c(46) + x(2)*(3.0*c(47)))))) + x(0)*(c(52) + x(2)*(c(53) + x(2)*(c(54) + x(2)*c(55))) + x(1)*(2.0*c(56) + x(2)*(2.0*c(57) + x(2)*(2.0*c(58) + x(2)*(2.0*c(59)))) + x(1)*(3.0*c(60) + x(2)*(3.0*c(61) + x(2)*(3.0*c(62) + x(2)*(3.0*c(63))))))))),
                    c(1) + x(2)*(2.0*c(2) + x(2)*(3.0*c(3))) + x(1)*(c(5) + x(2)*(2.0*c(6) + x(2)*(3.0*c(7))) + x(1)*(c(9) + x(2)*(2.0*c(10) + x(2)*(3.0*c(11))) + x(1)*(c(13) + x(2)*(2.0*c(14) + x(2)*(3.0*c(15)))))) + x(0)*(c(17) + x(2)*(2.0*c(18) + x(2)*(3.0*c(19))) + x(1)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(c(25) + x(2)*(2.0*c(26) + x(2)*(3.0*c(27))) + x(1)*(c(29) + x(2)*(2.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(c(33) + x(2)*(2.0*c(34) + x(2)*(3.0*c(35))) + x(1)*(c(37) + x(2)*(2.0*c(38) + x(2)*(3.0*c(39))) + x(1)*(c(41) + x(2)*(2.0*c(42) + x(2)*(3.0*c(43))) + x(1)*(c(45) + x(2)*(2.0*c(46) + x(2)*(3.0*c(47)))))) + x(0)*(c(49) + x(2)*(2.0*c(50) + x(2)*(3.0*c(51))) + x(1)*(c(53) + x(2)*(2.0*c(54) + x(2)*(3.0*c(55))) + x(1)*(c(57) + x(2)*(2.0*c(58) + x(2)*(3.0*c(59))) + x(1)*(c(61) + x(2)*(2.0*c(62) + x(2)*(3.0*c(63))))))))));
            }

            uvector<double,6> hessian(const uvector<double,3>& x) const
            {
                return uvector<double,6>(2.0*c(32) + x(2)*(2.0*c(33) + x(2)*(2.0*c(34) + x(2)*(2.0*c(35)))) + x(1)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(2.0*c(44) + x(2)*(2.0*c(45) + x(2)*(2.0*c(46) + x(2)*(2.0*c(47))))))) + x(0)*(6.0*c(48) + x(2)*(6.0*c(49) + x(2)*(6.0*c(50) + x(2)*(6.0*c(51)))) + x(1)*(6.0*c(52) + x(2)*(6.0*c(53) + x(2)*(6.0*c(54) + x(2)*(6.0*c(55)))) + x(1)*(6.0*c(56) + x(2)*(6.0*c(57) + x(2)*(6.0*c(58) + x(2)*(6.0*c(59)))) + x(1)*(6.0*c(60) + x(2)*(6.0*c(61) + x(2)*(6.0*c(62) + x(2)*(6.0*c(63)))))))),
                    c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(3.0*c(28) + x(2)*(3.0*c(29) + x(2)*(3.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(4.0*c(40) + x(2)*(4.0*c(41) + x(2)*(4.0*c(42) + x(2)*(4.0*c(43)))) + x(1)*(6.0*c(44) + x(2)*(6.0*c(45) + x(2)*(6.0*c(46) + x(2)*(6.0*c(47)))))) + x(0)*(3.0*c(52) + x(2)*(3.0*c(53) + x(2)*(3.0*c(54) + x(2)*(3.0*c(55)))) + x(1)*(6.0*c(56) + x(2)*(6.0*c(57) + x(2)*(6.0*c(58) + x(2)*(6.0*c(59)))) + x(1)*(9.0*c(60) + x(2)*(9.0*c(61) + x(2)*(9.0*c(62) + x(2)*(9.0*c(63)))))))),
                    c(17) + x(2)*(2.0*c(18) + x(2)*(3.0*c(19))) + x(1)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(c(25) + x(2)*(2.0*c(26) + x(2)*(3.0*c(27))) + x(1)*(c(29) + x(2)*(2.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(2.0*c(33) + x(2)*(4.0*c(34) + x(2)*(6.0*c(35))) + x(1)*(2.0*c(37) + x(2)*(4.0*c(38) + x(2)*(6.0*c(39))) + x(1)*(2.0*c(41) + x(2)*(4.0*c(42) + x(2)*(6.0*c(43))) + x(1)*(2.0*c(45) + x(2)*(4.0*c(46) + x(2)*(6.0*c(47)))))) + x(0)*(3.0*c(49) + x(2)*(6.0*c(50) + x(2)*(9.0*c(51))) + x(1)*(3.0*c(53) + x(2)*(6.0*c(54) + x(2)*(9.0*c(55))) + x(1)*(3.0*c(57) + x(2)*(6.0*c(58) + x(2)*(9.0*c(59))) + x(1)*(3.0*c(61) + x(2)*(6.0*c(62) + x(2)*(9.0*c(63)))))))),
                    2.0*c(8) + x(2)*(2.0*c(9) + x(2)*(2.0*c(10) + x(2)*(2.0*c(11)))) + x(1)*(6.0*c(12) + x(2)*(6.0*c(13) + x(2)*(6.0*c(14) + x(2)*(6.0*c(15))))) + x(0)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(6.0*c(28) + x(2)*(6.0*c(29) + x(2)*(6.0*c(30) + x(2)*(6.0*c(31))))) + x(0)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(6.0*c(44) + x(2)*(6.0*c(45) + x(2)*(6.0*c(46) + x(2)*(6.0*c(47))))) + x(0)*(2.0*c(56) + x(2)*(2.0*c(57) + x(2)*(2.0*c(58) + x(2)*(2.0*c(59)))) + x(1)*(6.0*c(60) + x(2)*(6.0*c(61) + x(2)*(6.0*c(62) + x(2)*(6.0*c(63)))))))),
                    c(5) + x(2)*(2.0*c(6) + x(2)*(3.0*c(7))) + x(1)*(2.0*c(9) + x(2)*(4.0*c(10) + x(2)*(6.0*c(11))) + x(1)*(3.0*c(13) + x(2)*(6.0*c(14) + x(2)*(9.0*c(15))))) + x(0)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(2.0*c(25) + x(2)*(4.0*c(26) + x(2)*(6.0*c(27))) + x(1)*(3.0*c(29) + x(2)*(6.0*c(30) + x(2)*(9.0*c(31))))) + x(0)*(c(37) + x(2)*(2.0*c(38) + x(2)*(3.0*c(39))) + x(1)*(2.0*c(41) + x(2)*(4.0*c(42) + x(2)*(6.0*c(43))) + x(1)*(3.0*c(45) + x(2)*(6.0*c(46) + x(2)*(9.0*c(47))))) + x(0)*(c(53) + x(2)*(2.0*c(54) + x(2)*(3.0*c(55))) + x(1)*(2.0*c(57) + x(2)*(4.0*c(58) + x(2)*(6.0*c(59))) + x(1)*(3.0*c(61) + x(2)*(6.0*c(62) + x(2)*(9.0*c(63)))))))),
                    2.0*c(2) + x(2)*(6.0*c(3)) + x(1)*(2.0*c(6) + x(2)*(6.0*c(7)) + x(1)*(2.0*c(10) + x(2)*(6.0*c(11)) + x(1)*(2.0*c(14) + x(2)*(6.0*c(15))))) + x(0)*(2.0*c(18) + x(2)*(6.0*c(19)) + x(1)*(2.0*c(22) + x(2)*(6.0*c(23)) + x(1)*(2.0*c(26) + x(2)*(6.0*c(27)) + x(1)*(2.0*c(30) + x(2)*(6.0*c(31))))) + x(0)*(2.0*c(34) + x(2)*(6.0*c(35)) + x(1)*(2.0*c(38) + x(2)*(6.0*c(39)) + x(1)*(2.0*c(42) + x(2)*(6.0*c(43)) + x(1)*(2.0*c(46) + x(2)*(6.0*c(47))))) + x(0)*(2.0*c(50) + x(2)*(6.0*c(51)) + x(1)*(2.0*c(54) + x(2)*(6.0*c(55)) + x(1)*(2.0*c(58) + x(2)*(6.0*c(59)) + x(1)*(2.0*c(62) + x(2)*(6.0*c(63)))))))));
            }
        };
    } // namespace detail

    // The following templated typedefs allow a particular dimension N and particular polynomial
    // specified by an integer parameter Degree to be translated to the specific polynomial
    // class implemented above. Degree = -1 means use the bicubic/tricubic. If one tries to
    // use an underfined polynomial, a compiler error relating to "InvalidPoly" will be generated.

    struct InvalidPoly {};

    template<int N, int Degree>
    struct StencilPoly
    {
        typedef InvalidPoly T_Poly;
    };

    template<>
    struct StencilPoly<2,2>
    {
        typedef detail::N2_PolyDegree2 T_Poly;
    };

    template<>
    struct StencilPoly<2,3>
    {
        typedef detail::N2_PolyDegree3 T_Poly;
    };

    template<>
    struct StencilPoly<2,4>
    {
        typedef detail::N2_PolyDegree4 T_Poly;
    };

    template<>
    struct StencilPoly<2,5>
    {
        typedef detail::N2_PolyDegree5 T_Poly;
    };

    template<>
    struct StencilPoly<2,-1>
    {
        typedef detail::N2_Bicubic T_Poly;
    };

    template<>
    struct StencilPoly<3,2>
    {
        typedef detail::N3_PolyDegree2 T_Poly;
    };

    template<>
    struct StencilPoly<3,3>
    {
        typedef detail::N3_PolyDegree3 T_Poly;
    };

    template<>
    struct StencilPoly<3,4>
    {
        typedef detail::N3_PolyDegree4 T_Poly;
    };

    template<>
    struct StencilPoly<3,5>
    {
        typedef detail::N3_PolyDegree5 T_Poly;
    };

    template<>
    struct StencilPoly<3,-1>
    {
        typedef detail::N3_Tricubic T_Poly;
    };
} // namespace algoim

#endif
